/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2007,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#include <stdio.h>
#include <spu_intrinsics.h>
#include "abi_util.h"
#include "libtest.h"

/* This testcase attempts to verify compliance of a C compiler to the ABI.
 * Some ABI compliance is difficult to test and problems may not be readily
 * identifiable. 
 *
 * Compliance testing is accomplised through successful compilation of 
 * the testcase and successful execution of the test.
 */


/* BYTE ORDERING TEST
 * This test verifies byte order of each of the data types. In general
 * this test accomplishes very little since byte ordering is inherant to
 * the processor architecture.
 */
unsigned short bo_halfword = 0x1011;
unsigned int bo_word = 0x20212223;
unsigned long long bo_doubleword = 0x3031323334353637ull;
vector unsigned char bo_quadword = ((vector unsigned char) {
					       0x40, 0x41, 0x42, 0x43,
					       0x44, 0x45, 0x46, 0x47,
					       0x48, 0x49, 0x4A, 0x4B,
					       0x4C, 0x4D, 0x4E, 0x4F});

#define VERIFY_BO(_label, _data, _size, _base) {					\
  int i;										\
  unsigned char *ptr;									\
  unsigned char exp;									\
  if (!failed) {									\
    for (i=0, exp=_base, ptr=(unsigned char *)(&_data);					\
         i<(int)_size;									\
         i++, exp++) {									\
      if (ptr[i] != exp) {								\
        failed = 1;									\
        DEBUG(1, "FAILED\n\t %s expected = 0x%x  got = 0x%x\n", _label, exp, ptr[i]);	\
        break;										\
      }											\
    }											\
  }											\
}

int testByteOrdering()
{
  int failed = 0;

  DEBUG(1, "Verify byte ordering - ");
  VERIFY_BO("halfword", bo_halfword, sizeof(unsigned short), 0x10);
  VERIFY_BO("word", bo_word, sizeof(unsigned int), 0x20);
  VERIFY_BO("doubleword", bo_doubleword, sizeof(unsigned long long), 0x30);
  VERIFY_BO("quadword", bo_quadword, sizeof(vector unsigned char), 0x40);
  if (!failed) DEBUG(1, "PASSED\n");

  return (failed);
}


/* DATA TYPE TEST
 * This test assures that each of the allowable data types is properly sized
 * and aligned according to the documented alignemt. Alignment is verified
 * by placing an instance of a data type on an non-natural alignment and assure
 * that is was correctly aligned.
 */
enum elist { ENUM1, ENUM2, ENUM3 };

struct {
  char a;
  char b;
  char c;
} c;

struct {
  char a;
  unsigned char b;
  char c;
} uc;

struct {
  char a;
  signed char b;
  char c;
} sc;

struct {
  char a;
  short b;
  char c;
} s;

struct {
  char a;
  unsigned short b;
  char c;
} us;

struct {
  char a;
  signed short b;
  char c;
} ss;

struct {
  char a;
  enum elist b;
  char c;
} e;


struct {
  char a;
  int b;
  char c;
} i;

struct {
  char a;
  unsigned int b;
  char c;
} ui;

struct {
  char a;
  signed int b;
  char c;
} si;

struct {
  char a;
  long int b;
  char c;
} li;

struct {
  char a;
  unsigned long int b;
  char c;
} uli;

struct {
  char a;
  signed long int b;
  char c;
} sli;

struct {
  char a;
  long long b;
  char c;
} ll;

struct {
  char a;
  unsigned long long b;
  char c;
} ull;

struct {
  char a;
  signed long long b;
  char c;
} sll;

struct {
  char a;
  char *b;
  char c;
} ptr;

struct {
  char a;
  void (*b)(void);
  char c;
} fptr;

struct {
  char a;
  float b;
  char c;
} f;

struct {
  char a;
  double b;
  char c;
} d;

struct {
  char a;
  vector unsigned char b;
  char c;
} vuc;

struct {
  char a;
  vector signed char b;
  char c;
} vsc;

struct {
  char a;
  vector unsigned short b;
  char c;
} vus;

struct {
  char a;
  vector signed short b;
  char c;
} vss;

struct {
  char a;
  vector unsigned int b;
  char c;
} vui;

struct {
  char a;
  vector signed int b;
  char c;
} vsi;

struct {
  char a;
  vector unsigned long long b;
  char c;
} vull;

struct {
  char a;
  vector signed long long b;
  char c;
} vsll;

struct {
  char a;
  vector float b;
  char c;
} vf;

struct {
  char a;
  vector double b;
  char c;
} vd;


struct {
  char a;
  _Bool b;
  char c;
} B;


struct {
  char a;
  long double b;
  char c;
} ld;






int testDataTypeSizes()
{
  int failed;

  DEBUG(1, "Verify data type sizes - ");

  failed = 0;
  if (sizeof(char) != 1) failed = 1;
  else if (sizeof(unsigned char) != 1) failed = 2;
  else if (sizeof(signed char) != 1) failed = 3;
  else if (sizeof(short) != 2) failed = 4;
  else if (sizeof(unsigned short) != 2) failed = 5;
  else if (sizeof(signed short) != 2) failed = 6;
  else if (sizeof(int) != 4) failed = 7;
  else if (sizeof(signed int) != 4) failed = 8;
  else if (sizeof(long int) != 4) failed = 9;
  else if (sizeof(signed long) != 4) failed = 10;
  else if (sizeof(enum elist) != 4) failed = 11;
  else if (sizeof(unsigned int) != 4) failed = 12;
  else if (sizeof(unsigned long) != 4) failed = 13;
  else if (sizeof(long long) != 8) failed = 14;
  else if (sizeof(signed long long) != 8) failed = 15;
  else if (sizeof(unsigned long long) != 8) failed = 16;
  else if (sizeof(&failed) != 4) failed = 17;
  else if (sizeof(&testDataTypeSizes) != 4) failed = 18;
  else if (sizeof(float) != 4) failed = 19;
  else if (sizeof(double) != 8) failed = 20;
  else if (sizeof(qword) != 16) failed = 21;
  else if (sizeof(vector unsigned char) != 16) failed = 22;
  else if (sizeof(vector signed char) != 16) failed = 23;
  else if (sizeof(vector unsigned short) != 16) failed = 24;
  else if (sizeof(vector signed short) != 16) failed = 25;
  else if (sizeof(vector unsigned int) != 16) failed = 26;
  else if (sizeof(vector signed int) != 16) failed = 27;
  else if (sizeof(vector unsigned long long) != 16) failed = 28;
  else if (sizeof(vector signed long long) != 16) failed = 29;
  else if (sizeof(vector float) != 16) failed = 30;
  else if (sizeof(vector double) != 16) failed = 31;
  else if (sizeof(_Bool) != 1) failed = 32;
  else if (sizeof(long double) != 8) failed = 33;

  if (failed) {
    DEBUG(1, "FAILED (%d)\n", failed);
  } else {
    DEBUG(1, "PASSED\n");
  }

  return (failed);
}


int testDataTypeAlignments()
{
  int failed;

  DEBUG(1, "Verify data type alignments - ");
  /* For each of the structure, verify the size, its alignment,
   * and the alignment of the field b.
   */
  failed = 0;
  if      ((sizeof(c)   != 3)  || ((unsigned int)(&c)   & 0)  || ((unsigned int)(&c.b)   & 0))  failed = 1;
  else if ((sizeof(uc)  != 3)  || ((unsigned int)(&uc)  & 0)  || ((unsigned int)(&uc.b)  & 0))  failed = 2;
  else if ((sizeof(sc)  != 3)  || ((unsigned int)(&sc)  & 0)  || ((unsigned int)(&sc.b)  & 0))  failed = 3;
  else if ((sizeof(s)   != 6)  || ((unsigned int)(&s)   & 1)  || ((unsigned int)(&s.b)   & 1))  failed = 4;
  else if ((sizeof(us)  != 6)  || ((unsigned int)(&us)  & 1)  || ((unsigned int)(&us.b)  & 1))  failed = 5;
  else if ((sizeof(ss)  != 6)  || ((unsigned int)(&ss)  & 1)  || ((unsigned int)(&ss.b)  & 1))  failed = 6;
  else if ((sizeof(i)   != 12) || ((unsigned int)(&i)   & 3)  || ((unsigned int)(&i.b)   & 3))  failed = 7;
  else if ((sizeof(ui)  != 12) || ((unsigned int)(&ui)  & 3)  || ((unsigned int)(&ui.b)  & 3))  failed = 8;
  else if ((sizeof(si)  != 12) || ((unsigned int)(&si)  & 3)  || ((unsigned int)(&si.b)  & 3))  failed = 9;
  else if ((sizeof(li)  != 12) || ((unsigned int)(&li)  & 3)  || ((unsigned int)(&li.b)  & 3))  failed = 10;
  else if ((sizeof(uli) != 12) || ((unsigned int)(&uli) & 3)  || ((unsigned int)(&uli.b) & 3))  failed = 11;
  else if ((sizeof(sli) != 12) || ((unsigned int)(&sli) & 3)  || ((unsigned int)(&sli.b) & 3))  failed = 12;
  else if ((sizeof(e)   != 12) || ((unsigned int)(&e)   & 3)  || ((unsigned int)(&e.b)   & 3))  failed = 13;
  else if ((sizeof(f)   != 12) || ((unsigned int)(&f)   & 3)  || ((unsigned int)(&f.b)   & 3))  failed = 14;
  else if ((sizeof(ptr) != 12) || ((unsigned int)(&ptr) & 3)  || ((unsigned int)(&ptr.b) & 3))  failed = 15;
  else if ((sizeof(fptr)!= 12) || ((unsigned int)(&fptr)& 3)  || ((unsigned int)(&fptr.b)& 3))  failed = 16;
  else if ((sizeof(ll)  != 24) || ((unsigned int)(&ll)  & 7)  || ((unsigned int)(&ll.b)  & 7))  failed = 17;
  else if ((sizeof(ull) != 24) || ((unsigned int)(&ull) & 7)  || ((unsigned int)(&ull.b) & 7))  failed = 18;
  else if ((sizeof(sll) != 24) || ((unsigned int)(&sll) & 7)  || ((unsigned int)(&sll.b) & 7))  failed = 19;
  else if ((sizeof(d)   != 24) || ((unsigned int)(&d)   & 7)  || ((unsigned int)(&d.b)   & 7))  failed = 20;
  else if ((sizeof(vuc) != 48) || ((unsigned int)(&vuc) & 15) || ((unsigned int)(&vuc.b) & 15)) failed = 21;
  else if ((sizeof(vsc) != 48) || ((unsigned int)(&vsc) & 15) || ((unsigned int)(&vsc.b) & 15)) failed = 22;
  else if ((sizeof(vus) != 48) || ((unsigned int)(&vus) & 15) || ((unsigned int)(&vus.b) & 15)) failed = 23;
  else if ((sizeof(vss) != 48) || ((unsigned int)(&vss) & 15) || ((unsigned int)(&vss.b) & 15)) failed = 24;
  else if ((sizeof(vui) != 48) || ((unsigned int)(&vui) & 15) || ((unsigned int)(&vui.b) & 15)) failed = 25;
  else if ((sizeof(vsi) != 48) || ((unsigned int)(&vsi) & 15) || ((unsigned int)(&vsi.b) & 15)) failed = 26;
  else if ((sizeof(vull)!= 48) || ((unsigned int)(&vull)& 15) || ((unsigned int)(&vull.b)& 15)) failed = 27;
  else if ((sizeof(vsll)!= 48) || ((unsigned int)(&vsll)& 15) || ((unsigned int)(&vsll.b)& 15)) failed = 28;
  else if ((sizeof(vf)  != 48) || ((unsigned int)(&vf)  & 15) || ((unsigned int)(&vf.b)  & 15)) failed = 29;
  else if ((sizeof(vd)  != 48) || ((unsigned int)(&vd)  & 15) || ((unsigned int)(&vd.b)  & 15)) failed = 30;
  else if ((sizeof(B)   !=  3) || ((unsigned int)(&B)   &  0) || ((unsigned int)(&B.b)   &  0)) failed = 31;
  else if ((sizeof(ld)  != 24) || ((unsigned int)(&ld)  &  7) || ((unsigned int)(&ld.b)  &  7)) failed = 32;

  if (failed) {
    DEBUG(1, "FAILED (%d)\n", failed);
  } else {
    DEBUG(1, "PASSED\n");
  }
  return (failed);
}


unsigned char c1 __attribute__ ((aligned (16)));
unsigned char g_uc;

unsigned char c2 __attribute__ ((aligned (16)));
signed char g_c;

unsigned char c3 __attribute__ ((aligned (16)));
unsigned short g_us;

unsigned char c4 __attribute__ ((aligned (16)));
signed short g_s;

unsigned char c5 __attribute__ ((aligned (16)));
unsigned int g_ui;

unsigned char c6 __attribute__ ((aligned (16)));
signed int g_i;

unsigned char c7 __attribute__ ((aligned (16)));
float g_f;

unsigned char c8 __attribute__ ((aligned (16)));
double g_d;

unsigned char c9 __attribute__ ((aligned (16)));
unsigned long long g_ull;

unsigned char c10 __attribute__ ((aligned (16)));
signed long long g_ll;

unsigned char c11 __attribute__ ((aligned (16)));
qword  g_qw;

unsigned char c12 __attribute__ ((aligned (16)));
vector unsigned int  g_vui;


int testGlobalAlignments()
{
  int failed = 0;

  DEBUG(1, "Verify global data alignments - ");

  /* Verify that all types of globals are quadword aligned.
   */
  if ((unsigned int)(&g_uc) & 0xF) failed = 1;
  else if ((unsigned int)(&g_c) & 0xF) failed = 2;
  else if ((unsigned int)(&g_us) & 0xF) failed = 3;
  else if ((unsigned int)(&g_s) & 0xF) failed = 4;
  else if ((unsigned int)(&g_ui) & 0xF) failed = 5;
  else if ((unsigned int)(&g_i) & 0xF) failed = 6;
  else if ((unsigned int)(&g_f) & 0xF) failed = 7;
  else if ((unsigned int)(&g_d) & 0xF) failed = 8;
  else if ((unsigned int)(&g_ull) & 0xF) failed = 9;
  else if ((unsigned int)(&g_ll) & 0xF) failed = 10;
  else if ((unsigned int)(&g_qw) & 0xF) failed = 11;
  else if ((unsigned int)(&g_vui) & 0xF) failed = 12;

  if (failed) {
    DEBUG(1, "FAILED (%d)\n", failed);
  } else {
    DEBUG(1, "PASSED\n");
  }
  return (failed);
}


int testBitFields()
{
  unsigned int w0, w1;
  union {
    struct {
      unsigned int a : 5;
      unsigned int b : 9;
      unsigned int c : 6;
      unsigned int d : 12;
    } b1;
    unsigned int i1;
  } u1;

  union {
    struct {
      unsigned int  x;		/* used to force alignment of the structure */
      unsigned char a : 5;
      unsigned char b : 4;
      unsigned char c : 3;
      unsigned char d : 2;
      unsigned char e : 8;
    } b2;
    unsigned int i2[2];
  } u2;

  union {
    struct {
      unsigned int a: 6;
      unsigned int  : 5;
      unsigned int b: 4;
      unsigned int  : 0;
      unsigned int c: 3;
      unsigned int d: 2;
      unsigned int e: 1;
    } b3;
    unsigned int i3[2];
  } u3;

  struct {
    signed char 	a: 8;
    char 		b: 8;
    unsigned char	c: 8;
    signed short	d: 16;
    short		e: 16;
    unsigned short	f: 16;
    signed int		g: 32;
    int			h: 32;
    unsigned int	i: 32;
    enum elist		j: 32;
    signed long		k: 32;
    long		l: 32;
    unsigned long	m: 32;
    signed long	long	n: 64;
    long long		o: 64;
    unsigned long long	p: 64;
  } bmax;

  DEBUG(1, "Verify bitfields - ");

  bmax.l = 0xfeedface;

  /* Verify that bitfields are allocated left to right.
   */
  u1.i1 = 0;
  u1.b1.b = 0x1FF;
  w0 = u1.i1;
  if (w0 != 0x07FC0000) {
    DEBUG(1, "FAILED\n\tleft to right allocation not observed\n");
    DEBUG(1, "\texpected 0x07FC0000  got 0x%x\n", w0);
    return (1);
  }

  /* Verify that a bit-field doesn't cross its unit boundary.
   */
  u2.b2.x = 0;
  u2.b2.a = 0x11;
  u2.b2.b = 0xA;
  u2.b2.c = 0x5;
  u2.b2.d = 0x3;
  u2.b2.e = 0x81;

  w0 = u2.i2[1];
  if ((w0 & 0xF8FEC0FF) != 0x88AAC081) {
    DEBUG(1, "FAILED\n\tbitfields likely crossed unit boundaries\n");
    DEBUG(1, "\texpected 0x88AAC000  got 0x%x\n", w0);
    return (1);
  }

  /* Verify unnamed bits bitfields enforcing other members from 
   * residing in the storage unit.
   */
  u3.i3[0] = 0;
  u3.i3[1] = 0;

  u3.b3.a = 0x21;
  u3.b3.b = 0xA;
  u3.b3.c = 0x5;
  u3.b3.d = 0x3;

  w0 = u3.i3[0];
  w1 = u3.i3[1];

  if ((w0 != 0x84140000) || (w1 != 0xB8000000)) {
    DEBUG(1, "FAILED\n\tzero width bitfields not correctly handled\n");
    DEBUG(1, "\texpected 0x84120000 0xB8000000  got 0x%x 0x%x\n", w0, w1);
    return (1);
  }
  DEBUG(1, "PASSED\n");

  return (0);
}

/* The nonVolatileFunc is used to determine if the non-volatile registers
 * are properly saved. The main purpose of this function is to use as many 
 * registers as possible. The caller can then test to make sure that none
 * of the non-volatile registers are corrupted by the routine.
 */
float m[80];

float nonVolatileFunc(float in0, float in1, float in2, float in3, float in4, float in5, float in6, float in7)
{
  int i;
  int count;
  float out0, out1, out2, out3, out4, out5, out6, out7;
  float m00, m01, m02, m03, m04, m05, m06, m07, m08, m09;
  float m10, m11, m12, m13, m14, m15, m16, m17, m18, m19;
  float m20, m21, m22, m23, m24, m25, m26, m27, m28, m29;
  float m30, m31, m32, m33, m34, m35, m36, m37, m38, m39;
  float m40, m41, m42, m43, m44, m45, m46, m47, m48, m49;
  float m50, m51, m52, m53, m54, m55, m56, m57, m58, m59;
  float m60, m61, m62, m63, m64, m65, m66, m67, m68, m69;
  float m70, m71, m72, m73, m74, m75, m76, m77, m78, m79;

  count = get5();
  out0 = in0;
  out1 = in1;
  out2 = in2;
  out3 = in3;
  out4 = in4;
  out5 = in5;
  out6 = in6;
  out7 = in7;

  m00 = m[0];  m01 = m[1];  m02 = m[2];  m03 = m[3]; m04 = m[4]; 
  m05 = m[5];  m06 = m[6];  m07 = m[7];  m08 = m[8]; m09 = m[9];
  m10 = m[10]; m11 = m[11]; m12 = m[12]; m13 = m[13]; m14 = m[14];
  m15 = m[15]; m16 = m[16]; m17 = m[17]; m18 = m[18]; m19 = m[19];
  m20 = m[20]; m21 = m[21]; m22 = m[22]; m23 = m[23]; m24 = m[24]; 
  m25 = m[25]; m26 = m[26]; m27 = m[27]; m28 = m[28]; m29 = m[29];
  m30 = m[30]; m31 = m[31]; m32 = m[32]; m33 = m[33]; m34 = m[34];
  m35 = m[35]; m36 = m[36]; m37 = m[37]; m38 = m[38]; m39 = m[39];
  m40 = m[40]; m41 = m[41]; m42 = m[42]; m43 = m[43]; m44 = m[44]; 
  m45 = m[45]; m46 = m[46]; m47 = m[47]; m48 = m[48]; m49 = m[49];
  m50 = m[50]; m51 = m[51]; m52 = m[52]; m53 = m[53]; m54 = m[54];
  m55 = m[55]; m56 = m[56]; m57 = m[57]; m58 = m[58]; m59 = m[59];
  m60 = m[60]; m61 = m[61]; m62 = m[62]; m63 = m[63]; m64 = m[64];
  m65 = m[65]; m66 = m[66]; m67 = m[67]; m68 = m[68]; m69 = m[69];
  m70 = m[70]; m71 = m[71]; m72 = m[72]; m73 = m[73]; m74 = m[74];
  m75 = m[75]; m76 = m[76]; m77 = m[77]; m78 = m[78]; m79 = m[79];

  for (i=0; i<count; i++) {
    out0 = m00 * out0 + m01;
    out1 = m10 * out1 + m11;
    out2 = m20 * out2 + m21;
    out3 = m30 * out3 + m31;
    out4 = m40 * out4 + m41;
    out5 = m50 * out5 + m51;
    out6 = m60 * out6 + m61;
    out7 = m70 * out7 + m71;

    out0 = m02 * out0 + m03;
    out1 = m12 * out1 + m13;
    out2 = m22 * out2 + m23;
    out3 = m32 * out3 + m33;
    out4 = m42 * out4 + m43;
    out5 = m52 * out5 + m53;
    out6 = m62 * out6 + m63;
    out7 = m72 * out7 + m73;

    out0 = m04 * out0 + m05;
    out1 = m14 * out1 + m15;
    out2 = m24 * out2 + m25;
    out3 = m34 * out3 + m35;
    out4 = m44 * out4 + m45;
    out5 = m54 * out5 + m55;
    out6 = m64 * out6 + m65;
    out7 = m74 * out7 + m75;

    out0 = m06 * out0 + m07;
    out1 = m16 * out1 + m17;
    out2 = m26 * out2 + m27;
    out3 = m36 * out3 + m37;
    out4 = m46 * out4 + m47;
    out5 = m56 * out5 + m57;
    out6 = m66 * out6 + m67;
    out7 = m76 * out7 + m77;

    out0 = m08 * out0 + m09;
    out1 = m18 * out1 + m19;
    out2 = m28 * out2 + m29;
    out3 = m38 * out3 + m39;
    out4 = m48 * out4 + m49;
    out5 = m58 * out5 + m59;
    out6 = m68 * out6 + m69;
    out7 = m78 * out7 + m79;
  }

  return (out0 + out1 + out2 + out3 + out4 + out5 + out6 + out7);
}


int testRegisters()
{
  int i, j, k, count;
  float f;
  DEBUG(1, "Verify register usage - ");

  for (i=0; i<100; i++) m[i] = 1.0f;

  /* Verify the volatility of the register file.
   */
  /* Verify volatiles. */
  f = 1.0f;
  j = 1;
  k = 0;
  count = get5();
  for (i=0; i<count; i++) {
    f *= 1.5f;
    j <<= 1;
    k += i;
    setVolatileRegisters();
  }
  if ((f != 7.59375f) || (j != 32) || (k != 10)) {
    DEBUG(1, "FAILED\n\tvolatile register setting affected calling function\n");
    DEBUG(1, "\texpected 7.59375f 32 10   got = %f %d %d\n", f, j, k);
    return (1);
  }

  /* Verify non-volatiles */
  if (testNonVolatiles(nonVolatileFunc)) {
    DEBUG(1, "FAILED\n\tnon-volatile register corrupted\n");
    return (1);
  }

  DEBUG(1, "PASSED\n");
  return (0);
}


int verifyStackFrame() __attribute__ ((noinline));

int verifyStackFrame()
{
  int count;
  int instructions;
  unsigned int *ptr;
  unsigned int initialSP, currentSP;
  unsigned int initialLR, currentLR;

  /* Fetch previously saved stack data
   */
  initialSP  = spu_extract(currentSD.sp, 0);
  initialLR  = spu_extract(currentSD.lr, 0);

  (void)saveStackData();

  /* Fetch latest save stack data
   */
  currentSP  = spu_extract(currentSD.sp, 0);
  currentLR  = spu_extract(currentSD.lr, 0);

  DEBUG(1, "Verify stack - ");

  /* Verify that the stack grew downward 
   */
  if (currentSP >= initialSP) {
    DEBUG(1, "FAILED\n\tstack grew in the wrong direction or not at all\n");
    DEBUG(1, "\tinitial SP 0x%x  current SP 0x%x\n", initialSP, currentSP);
    return (1);
  }

  /* Verify the back chain pointer 
   */
  ptr = (unsigned int *)(currentSP);
  if (initialSP != *ptr) {
    DEBUG(1, "FAILED\n\tback chain pointer doesn't point to initial stack pointer\n");
    DEBUG(1, "\texpected 0x%x  got 0x%x\n", initialSP, *ptr);
    return (1);
  }

  /* Verify there is a reasonable back chain pointer sequence
   */
  count = 0;
  ptr = (unsigned int *)(currentSP);
  while (ptr) {
    if (++count > 5) {
      DEBUG(1, "FAILED\n\tstack back chain length too large\n");
      return (1);
    }
    ptr = (unsigned int *)(*ptr);
  }

  
  /* Verify that the link register was saved in the callers stack frame
   * We assume that the calls to saveStackData and verifyStackFrame are
   * within 20 instructions of each other.
   */
  ptr = (unsigned int *)(initialSP);
  instructions = (*(ptr+4) - initialLR) / 4;
  if ((instructions < 1) || (instructions > 20)) {
    DEBUG(1, "FAILED\n\tthe link register was not correctly stored in the callers stack frame\n");
    DEBUG(1, "\texpected 0x%x  got 0x%x\n", initialLR, *(ptr+4));
    return (1);
  }

  DEBUG(1, "PASSED\n");
  return (0);	  
}




int testStackFrame()
{
  /* This test attempts to verify that the stack frame is compliance
   * to the ABI by:
   * 1) verifying the link register is properly placed into a stack frame
   * 2) verifying the stack grow in the correct direction and points to
   *	callers stack frame.
   */
  (void)saveStackData();
  return (verifyStackFrame());
}



static struct skipStruct skip;


int testParameterPassing()
{
  int i, *ptr; 
  int result;
  struct parmPassing s1 = {
    41, 42, 43, 44.0f, 45, 46.0, ((vector unsigned int) { 47,48,49,50}), 51
  };
  struct parmPassing s2 = {
    53, 54, 55, 56.0f, 57, 58.0, ((vector unsigned int) { 59,60,61,62}), 63
  };


  DEBUG(1, "Verify parameter passing - ");

  for (i=0, ptr=(int *)(skip.data); i<(int)(sizeof(struct skipStruct)/sizeof(int)); i++) {
    ptr[i] = i;
  }

  /* Verify simple parameter passing */
  result = testSimpleParmPassing1((unsigned char)11, 				/* 3 */
				  (unsigned short)12,				/* 4 */
				  (unsigned int)13,				/* 5 */
				  (float)14.0f,					/* 6 */
				  (unsigned long long)15, 			/* 7 */
				  (double)16.0,					/* 8 */
				  ((vector unsigned int) { 17,18,19,20}),	/* 9 */
				  (long int)21,					/* 10 */
				  (unsigned char)22,				/* 11 */
				  (unsigned short)23,				/* 12 */
				  (unsigned int)24,				/* 13 */
				  (float)25.0f,					/* 14 */
				  (unsigned long long)26,			/* 15 */
				  (double)27.0,					/* 16 */
				  ((vector unsigned int) { 28,29,30,31}),	/* 17 */
				  (long int)32);				/* 18 */
  if (result) {
    DEBUG(1, "FAILED\n\tsimple register parameter passing failed on parameter %d\n", result);
    return (1);
  }

  /* Verify stucture register passing and both simple register and stack parameter passing
   */
  result = testSimpleParmPassing2(skip,						/* 3 - 71 */
				  (unsigned char)11, 				/* 72 */
				  (unsigned short)12,				/* 73 */
				  (unsigned int)13,				/* 74 */
				  (float)14.0f,					/* stack */
				  (unsigned long long)15, 			/* stack */
				  (double)16.0,					/* stack */
				  ((vector unsigned int) { 17,18,19,20}),	/* stack */
				  (long int)21,					/* stack */
				  (unsigned char)22,				/* stack */
				  (unsigned short)23,				/* stack */
				  (unsigned int)24,				/* stack */
				  (float)25.0f,					/* stack */
				  (unsigned long long)26,			/* stack */
				  (double)27.0,					/* stack */
				  ((vector unsigned int) { 28,29,30,31}),	/* stack */
				  (long int)32);				/* stack */
  if (result) {
    DEBUG(1, "FAILED\n\tsimple structure register, simple stack parameter passing failed on parameter %d\n", result);
    return (1);
  }

  /* Verify struct parameter passsing */
  result = testStructParmPassing1(40, s1, 52, s2, 64);
  if (result) {
    DEBUG(1, "FAILED\n\tstructure parameter passing (test 1) failed on parameter %d\n", result);
    return (1);
  }

 result = testStructParmPassing2(skip, 40, s1, 52, s2, 64);
  if (result) {
    DEBUG(1, "FAILED\n\tstructure parameter passing (test 2) failed on parameter %d\n", result);
    return (1);
  }

 result = testStructParmPassing3(skip, skip, 40, s1, 52, s2, 64);
  if (result) {
    DEBUG(1, "FAILED\n\tstructure parameter passing (test 3) failed on parameter %d\n", result);
    return (1);
  }

  DEBUG(1, "PASSED\n");
  return (0);
}


unsigned char retUC = 1;
unsigned short retUS = 0x1234;
unsigned int retUI = 0x30313233;
unsigned long long retULL = 0x4000000120000004ULL;
float retF = 5.0f;
double retD = 6.0;
struct retSmall retSS = {0x70000007, 0x8, 0x9009, 0xA, 11.0f};
vector unsigned int retV = ((vector unsigned int) { 0xC0000001, 0xC0000002, 0xC0000003, 0xC0000004});
struct retUChar retUCS = { 0x13 };


struct retLarge1 retLS1;
struct retLarge2 retLS2;

unsigned char returnUChar(void)
{
  return (retUC);
}

unsigned short returnUShort(void)
{
  return (retUS);
}

unsigned int returnUInt(void)
{
  return (retUI);
}

unsigned long long returnULongLong(void)
{
  return (retULL);
}

float returnFloat(void)
{
  return (retF);
}

double returnDouble(void)
{
  return (retD);
}


struct retSmall returnStruct(void)
{
  return (retSS);
}

vector unsigned int  returnVector(void)
{
  return (retV);
}

struct retLarge1 returnLargeStruct1(vector unsigned int start)
{
  int i;

  for (i=0; i<70; i++) {
    retLS1.b[i] = start;
    start = spu_add(start, 4);
  }
  retLS1.c = spu_extract(start, 3);
  
  return (retLS1);
}


struct retLarge2 returnLargeStruct2(vector unsigned int start)
{
  int i;

  for (i=0; i<71; i++) {
    retLS2.b[i] = start;
    start = spu_add(start, 4);
  }
  retLS2.c = spu_extract(start, 3);
  
  return (retLS2);
}




struct retUChar returnUCharStruct(void)
{
  return (retUCS);
}


int testReturnValues(void)
{

  DEBUG(1, "Verify return values - ");

  /* Verify scalar, vectors and small return aggregates */
  if (testReturnUChar(returnUChar)) {
    DEBUG(1, "FAILED\n\tverifying unsigned char return value\n");
    return (1);
  }
  if (testReturnUShort(returnUShort)) {
    DEBUG(1, "FAILED\n\tverifying unsigned short return value\n");
    return (1);
  }
  if (testReturnUInt(returnUInt)) {
    DEBUG(1, "FAILED\n\tverifying unsigned int return value\n");
    return (1);
  }
  if (testReturnULongLong(returnULongLong)) {
    DEBUG(1, "FAILED\n\tverifying unsigned long long return value\n");
    return (1);
  }
  if (testReturnFloat(returnFloat)) {
    DEBUG(1, "FAILED\n\tverifying float return value\n");
    return (1);
  }
  if (testReturnDouble(returnDouble)) {
    DEBUG(1, "FAILED\n\tverifying double return value\n");
    return (1);
  }
  if (testReturnStruct(returnStruct)) {
    DEBUG(1, "FAILED\n\tverifying small structure return value\n");
    return (1);
  }
  if (testReturnVector(returnVector)) {
    DEBUG(1, "FAILED\n\tverifying vector return value\n");
    return (1);
  }

  if (testReturnUCharStruct(returnUCharStruct)) {
    DEBUG(1, "FAILED\n\tverifying uchar structure return value\n");
    return (1);
  }

#if 0
  /* Verify large return aggregates */
  if (testReturnLargeStruct1(returnLargeStruct1)) {
    DEBUG(1, "FAILED\n\tverifying large register structure return value\n");
    return (1);
  }
#endif
  if (testReturnLargeStruct2(returnLargeStruct2)) {
    DEBUG(1, "FAILED\n\tverifying large stack structure return value\n");
    return (1);
  }

  DEBUG(1, "PASSED\n");
  return (0);
}


int testVarArgs(void)
{
  int i, *ptr; 
  int result;
  struct {
    float a;
    unsigned char b;
    unsigned short c;
    vector unsigned int d;
  } x = { 9.0f, 10, 11, ((vector unsigned int) { 17,18,19,20})};

  DEBUG(1, "Verify variable arguments - ");

  /* Verify function with all variable arguments 
   */
  result = testVarArgsAll(1.0f, (unsigned char)2, (unsigned short)0x3003, ((vector unsigned int) { 4, 5, 6, 7}), 8.0, 0x8000000000000009ULL);
  if (result) {
    DEBUG(1, "FAILED\n\tverifying variable arguments failed on parameter %d\n", result);
    return (1);
  }

  result = testVarArgsStruct((unsigned int)21, x, 21.0f);
  if (result) {
    DEBUG(1, "FAILED\n\tverifying variable structure arguments failed on parameter %d\n", result);
    return (1);
  }

  /* Verify function with some fixed parameters
   */
  result = testVarArgsMost1(22, 23.0f, 24.0f, 25, (unsigned char)26, 27.0f, 
			   (unsigned short)28, ((vector float) { 14.0f, 15.0f, 16.0f, 17.0f}), 
			   (unsigned int)29, 30ULL);
  if (result) {
    DEBUG(1, "FAILED\n\tverifying variable arguments with fixed parameters failed on parameter %d\n", result);
    return (1);
  }

  /* Verify function with some fixed parameters that spills into the stack
   */
  for (i=0, ptr=(int *)(skip.data); i<(int)(sizeof(struct skipStruct)/sizeof(int)); i++) {
    ptr[i] = i;
  }

  result = testVarArgsMost2(skip, 22, 23.0f, 24.0f, 25, (unsigned char)26, 27.0f, 
			   (unsigned short)28, ((vector float) { 14.0f, 15.0f, 16.0f, 17.0f}), 
			   (unsigned int)29, 30ULL);
  if (result) {
    DEBUG(1, "FAILED\n\tverifying variable arguments spilling into stack with fixed parameters failed on parameter %d\n", result);
    return (1);
  }

  DEBUG(1, "PASSED\n");
  return (0);
}


int main()
{
  DECLARE_TEST_VARS()

  retLS1.a = 1;
  retLS2.a = 2;


  INVOKE_TEST(testByteOrdering);
  INVOKE_TEST(testDataTypeSizes);
  INVOKE_TEST(testDataTypeAlignments);
  INVOKE_TEST(testGlobalAlignments);
  INVOKE_TEST(testBitFields);

  INVOKE_TEST(testRegisters);
  INVOKE_TEST(testStackFrame);
  INVOKE_TEST(testParameterPassing);
  INVOKE_TEST(testReturnValues);
  INVOKE_TEST(testVarArgs);

  DISPLAY_RESULTS();
}
