/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2007,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ppu_intrinsics.h>
#include <math.h>
#include "libtest.h"

#define SAMPLES	20

static char *passed = "passed\n";
static char *failed = "FAILED\n";

unsigned int seed = 0;

unsigned int rand_local()
{
  seed = seed * 69607 + 54329;
  return (seed);
}

unsigned long long rand_ullong()
{
  return ((unsigned long long)(rand_local()) | ((unsigned long long)(rand_local()) << 32));
}

float rand_float(float min, float max)
{
  union {
    float f;
    unsigned int ui;
  } x;

  x.ui = (rand_local() & 0x7FFFFF) | 0x3F800000;
  x.f -= 1.0f;
  x.f *= (max - min);
  x.f += min;
  return (x.f);
}

double rand_double(double min, double max)
{
  union {
    unsigned int ui[2];
    double d;
  } x;

  x.ui[0] = (rand_local() & 0x000FFFFF) | 0x3FF00000;
  x.ui[1] = rand_local();
  x.d -= 1.0;
  x.d *= max - min;
  x.d += min;
  return (x.d);
}

int test_cctph()
{
    DEBUG(1, "__cctph............");

    __cctph();
  
    DEBUG(1, passed);
    return (0);
}

int test_cctpl()
{
    DEBUG(1, "__cctpl............");

    __cctpl();
  
    DEBUG(1, passed);
    return (0);
}

int test_cctpm()
{
    DEBUG(1, "__cctpm............");

    __cctpm();
  
    DEBUG(1, passed);
    return (0);
}

int test_cntlzd()
{
    unsigned long long a, savea;
    unsigned int d, exp;
    
    DEBUG(1, "__cntlzd...........");
    
    a = rand_ullong();
    d = __cntlzd(a);
    
    exp = 64;
    savea = a;
    while (a != 0) {
        a >>= 1;
        exp--;
    }
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%08x%08x d=%u exp=%u\n", (unsigned int)(savea >> 32),
                                                 (unsigned int)savea,
                                                 d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_cntlzw()
{
    unsigned int a, savea, d, exp;
    DEBUG(1, "__cntlzw...........");

    a = rand_local();
    d = __cntlzw(a);
    
    exp = 32;
    savea = a;
    while (a != 0) {
        a >>= 1;
        exp--;
    }
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%08x d=%u exp=%u\n", savea, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_db10cyc()
{
    DEBUG(1, "__db10cyc..........");

    __db10cyc();
  
    DEBUG(1, passed);
    return (0);
}

int test_db12cyc()
{
    DEBUG(1, "__db12cyc..........");

    __db12cyc();
  
    DEBUG(1, passed);
    return (0);
}

int test_db16cyc()
{
    DEBUG(1, "__db16cyc..........");

    __db16cyc();
  
    DEBUG(1, passed);
    return (0);
}

int test_db8cyc()
{
    DEBUG(1, "__db8cyc...........");

    __db8cyc();
  
    DEBUG(1, passed);
    return (0);
}

int test_dcbf()
{
    unsigned long long *ptr;
    DEBUG(1, "__dcbf.............");

    ptr = (unsigned long long *)malloc(sizeof(unsigned long long));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    *ptr = rand_ullong();
    __dcbf(ptr);
  
    free((void *)ptr);

    DEBUG(1, passed);
    return (0);
}

int test_dcbst()
{
    unsigned long long *ptr;
    DEBUG(1, "__dcbst............");

    ptr = (unsigned long long *)malloc(sizeof(unsigned long long));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }

    *ptr = rand_ullong();
    __dcbst(ptr);
  
    free((void *)ptr);
    DEBUG(1, passed);
    return (0);
}

int test_dcbt()
{
    unsigned long long *ptr;
    DEBUG(1, "__dcbt.............");

    ptr = (unsigned long long *)malloc(sizeof(unsigned long long));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    *ptr = rand_ullong();
    __dcbt(ptr);
  
    free((void *)ptr);
    DEBUG(1, passed);
    return (0);
}

int test_dcbt_TH1000()
{
    void *ptr;
    DEBUG(1, "__dcbt_TH1000......");

    ptr = (void *)malloc(sizeof(128));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    __dcbt_TH1000(ptr, 1, 0, 15);
  
    /* Special aliases */
    __protected_stream_set(3, ptr, 14);
    __protected_unlimited_stream_set(3, ptr, 13);

    free((void *)ptr);
    DEBUG(1, passed);
    return (0);
}

int test_dcbt_TH1010()
{
    DEBUG(1, "__dcbt_TH1010......");

    __dcbt_TH1010(0, 10, 20, 1, 0, 15);
  
    /* Special aliases */
    __protected_stream_count(3, 14);
    __protected_stream_go();
    __protected_stream_stop(14);
    __protected_stream_stop_all();

    DEBUG(1, passed);
    return (0);
}

int test_dcbtst()
{
    unsigned long long *ptr;
    DEBUG(1, "__dcbtst...........");

    ptr = (unsigned long long *)malloc(2*128);
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    *ptr = rand_ullong();
    __dcbtst(ptr);
  
    free((void *)ptr);
    DEBUG(1, passed);
    return (0);
}


char buf[3*128];

int test_dcbz()
{
    int i;
    void *ptr;

    DEBUG(1, "__dcbz.............");

    /* Fill buffer non-zero data. */

    (void)memset(buf, 0xFF, sizeof(buf));

    /* Verify memset worked */
    for (i=0; i<(int)sizeof(buf); i++) {
      if (buf[i] != (char)0xFF) {
	DEBUG(1, failed);
	DEBUG(2, "\t\tmemset error\n");
      }
    }

    ptr = (void *)&buf[128];

    __dcbz(ptr);

    /* Verify that a portion of buf got zeroed
     */
    for (i=0; i<(int)sizeof(buf); i++) {
      if (buf[i] == (char)0) break;
    }
    if (i >= (int)sizeof(buf)) {
      DEBUG(1, failed);
      return (1);
    }
    DEBUG(1, passed);
    return (0);
}

int test_eieio()
{
    DEBUG(1, "__eieio............");

    __eieio();
  
    DEBUG(1, passed);
    return (0);
}

int test_fabs()
{
    double a, d, exp;
    DEBUG(1, "__fabs.............");

    a = rand_double(-4.0, 4.0);
    d = __fabs(a);
    
    if (a < 0.0) {
        exp = -a;
    }
    else {
        exp = a;
    }
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%f exp=%f\n", a, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fabsf()
{
    float a, d, exp;
    DEBUG(1, "__fabsf............");

    a = rand_float(-4.0, 4.0);
    d = __fabsf(a);
  
    if (a < 0.0f) {
        exp = -a;
    }
    else {
        exp = a;
    }
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%f exp=%f\n", a, d, exp);
        return (1);
    }
    
    DEBUG(1, passed);
    return (0);
}

int test_fcfid()
{
    long long a;
    double d, exp;
    DEBUG(1, "__fcfid............");

    a = rand_ullong();
    d = __fcfid(a);
    exp = (double)a;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%lld d=%f exp=%f\n", a, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fctid()
{
    double a, frac;
    long d, exp;
    DEBUG(1, "__fctid............");

    a = rand_double(-4.0, 4.0);
    d = __fctid(a);
    exp = (long)a;
    frac = a - (double)exp;
    if (a >= 0 && frac >= 0.5f) {
        exp++;
    }
    else if (a < 0 && frac <= -0.5f) {
        exp--;
    }
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%ld exp=%ld\n", a, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fctidz()
{
    double a;
    long d, exp;
    DEBUG(1, "__fctidz...........");

    a = rand_double(-4.0, 4.0);
    d = __fctidz(a);
    exp = (long)a;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%ld exp=%ld\n", a, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fctiw()
{
    double a, frac;
    int d, exp, i;
    DEBUG(1, "__fctiw............");

    for (i=0; i<SAMPLES; i++) {
      a = rand_double(-4.0, 4.0);
      d = __fctiw(a);
      exp = (int)a;
      frac = a - (double)exp;
      if (a >= 0.0) {
	if (frac >= 0.5f) {
	  exp++;   
	}
      } else {
	if (frac <= -0.5f) {
	  exp--;
	}
      }
    
      if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%d exp=%d\n", a, d, exp);
        return (1);
      }
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fctiwz()
{
    double a;
    int d, exp, i;
    DEBUG(1, "__fctiwz...........");

    for (i=0; i<SAMPLES; i++) {
      a = rand_double(-4.0, 4.0);
      d = __fctiwz(a);
      exp = (int)a;
    
      if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%d exp=%d\n", a, d, exp);
        return (1);
      }
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fmadd()
{
    double a, b, c, d, exp;
    DEBUG(1, "__fmadd............");

    a = rand_double(-4.0, 4.0);
    b = rand_double(-4.0, 4.0);
    c = rand_double(-4.0, 4.0);
    d = __fmadd(a, b, c);
    exp = a*b + c;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f b=%f c=%f d=%f exp=%f\n", a, b, c, d, exp);
        return (1);
    }
    
    DEBUG(1, passed);
    return (0);
}

int test_fmadds()
{
    float a, b, c, d, exp;
    DEBUG(1, "__fmadds...........");

    a = rand_float(-4.0, 4.0);
    b = rand_float(-4.0, 4.0);
    c = rand_float(-4.0, 4.0);
    d = __fmadds(a, b, c);
    exp = a*b + c;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f b=%f c=%f d=%f exp=%f\n", a, b, c, d, exp);
        return (1);
    }
    
    DEBUG(1, passed);
    return (0);
}

int test_fmsub()
{
    double a, b, c, d, exp;
    DEBUG(1, "__fmsub............");

    a = rand_double(-4.0, 4.0);
    b = rand_double(-4.0, 4.0);
    c = rand_double(-4.0, 4.0);
    d = __fmsub(a, b, c);
    exp = a*b - c;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f b=%f c=%f d=%f exp=%f\n", a, b, c, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fmsubs()
{
    float a, b, c, d, exp;
    DEBUG(1, "__fmsubs...........");

    a = rand_float(-4.0, 4.0);
    b = rand_float(-4.0, 4.0);
    c = rand_float(-4.0, 4.0);
    d = __fmsubs(a, b, c);
    exp = a*b - c;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f b=%f c=%f d=%f exp=%f\n", a, b, c, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fmul()
{
    double a, b, d, exp;
    DEBUG(1, "__fmul.............");

    a = rand_double(-4.0, 4.0);
    b = rand_double(-4.0, 4.0);
    d = __fmul(a, b);
    exp = a*b;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f b=%f d=%f exp=%f\n", a, b, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fmuls()
{
    float a, b, d, exp;
    DEBUG(1, "__fmuls............");

    a = rand_float(-4.0, 4.0);
    b = rand_float(-4.0, 4.0);
    d = __fmuls(a, b);
    exp = a*b;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f b=%f d=%f exp=%f\n", a, b, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fnabs()
{
    double a, d, exp;
    DEBUG(1, "__fnabs............");

    a = rand_double(-4.0, 4.0);
    d = __fnabs(a);
    if (a < 0.0) {
        exp = a;
    }
    else {
        exp = -a;
    }
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%f exp=%f\n", a, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fnabsf()
{
    float a, d, exp;
    DEBUG(1, "__fnabsf...........");

    a = rand_float(-4.0, 4.0);
    d = __fnabsf(a);
    if (a < 0.0f) {
        exp = a;
    }
    else {
        exp = -a;
    }
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%f exp=%f\n", a, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fnmadd()
{
    double a, b, c, d, exp;
    DEBUG(1, "__fnmadd...........");

    a = rand_double(-4.0, 4.0);
    b = rand_double(-4.0, 4.0);
    c = rand_double(-4.0, 4.0);
    d = __fnmadd(a, b, c);
    exp = -(a*b + c);
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f b=%f c=%f d=%f exp=%f\n", a, b, c, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fnmadds()
{
    float a, b, c, d, exp;
    DEBUG(1, "__fnmadds..........");

    a = rand_float(-4.0, 4.0);
    b = rand_float(-4.0, 4.0);
    c = rand_float(-4.0, 4.0);
    d = __fnmadds(a, b, c);
    exp = -(a*b + c);
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f b=%f c=%f d=%f exp=%f\n", a, b, c, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fnmsub()
{
    double a, b, c, d, exp;
    DEBUG(1, "__fnmsub...........");

    a = rand_double(-4.0, 4.0);
    b = rand_double(-4.0, 4.0);
    c = rand_double(-4.0, 4.0);
    d = __fnmsub(a, b, c);
    exp = -(a*b - c);
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f b=%f c=%f d=%f exp=%f\n", a, b, c, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fnmsubs()
{
    float a, b, c, d, exp;
    DEBUG(1, "__fnmsubs..........");

    a = rand_float(-4.0, 4.0);
    b = rand_float(-4.0, 4.0);
    c = rand_float(-4.0, 4.0);
    d = __fnmsubs(a, b, c);
    exp = -(a*b - c);
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f b=%f c=%f d=%f exp=%f\n", a, b, c, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fres()
{
    float a, d, exp, error;
    DEBUG(1, "__fres.............");

    a = rand_float(-4.0, 4.0);
    d = __fres(a);
    exp = 1.0f/a;
    error = abs((d - exp)/exp);
    
    if (error > 1.0f/256.0f) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%f exp=%f error=%f\n", a, d, exp, error);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_frsp()
{
    double a;
    float d, exp;
    DEBUG(1, "__frsp.............");
    
    a = rand_double(-4.0, 4.0);
    d = __frsp(a);
    /* TODO: not sure if this is equivalent */
    exp = (float)a;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%f exp=%f\n", a, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_frsqrte()
{
    double a, d, exp, error;
    DEBUG(1, "__frsqrte..........");

    a = rand_double(0, 4.0);
    d = __frsqrte(a);
    exp = 1.0/sqrt(a);
    error = abs((d - exp)/(exp));
    
    if (error > 1.0/32.0f) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%f exp=%f error=%f\n", a, d, exp, error);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fsel()
{
    double a, b, c, d, exp;
    DEBUG(1, "__fsel.............");

    a = rand_double(-1.0, 1.0);
    b = rand_double(-4.0, 4.0);
    c = rand_double(-4.0, 4.0);
    d = __fsel(a, b, c);
    if (a < 0.0) {
        exp = c;
    }
    else {
        exp = b;
    }
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f b=%f c=%f d=%f exp=%f\n", a, b, c, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fsels()
{
    float a, b, c, d, exp;
    DEBUG(1, "__fsels............");

    a = rand_float(-1.0, 1.0);
    b = rand_float(-4.0, 4.0);
    c = rand_float(-4.0, 4.0);
    d = __fsels(a, b, c);
    if (a < 0.0f) {
        exp = c;
    }
    else {
        exp = b;
    }
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f b=%f c=%f d=%f exp=%f\n", a, b, c, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_fsqrt()
{
    double a, d, exp;
    DEBUG(1, "__fsqrt............");

    a = rand_double(0.0, 4.0);
    d = __fsqrt(a);
    exp = sqrt(a);
  
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%f exp=%f\n", a, d, exp);
        return (1);
    }
    
    DEBUG(1, passed);
    return (0);
}

int test_fsqrts()
{
    float a, d, exp;
    DEBUG(1, "__fsqrts...........");

    a = rand_float(0.0, 4.0);
    d = __fsqrts(a);
    exp = sqrtf(a);
  
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%f d=%f exp=%f\n", a, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_icbi()
{
    unsigned long long *ptr;
    DEBUG(1, "__icbi.............");

    ptr = (unsigned long long *)malloc(sizeof(unsigned long long));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    *ptr = rand_ullong();
    __icbi(ptr);
  
    free((void *)ptr);
    DEBUG(1, passed);
    return (0);
}

int test_isync()
{
    DEBUG(1, "__isync............");

    __isync();
  
    DEBUG(1, passed);
    return (0);
}

/* ldarx is only supported in 64-bit */
#if defined (__powerpc64__)
int test_ldarx()
{
    unsigned long long *ptr, d, exp;
    DEBUG(1, "__ldarx............");

    ptr = (unsigned long long *)malloc(sizeof(unsigned long long));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    *ptr = rand_ullong();
    d = __ldarx((void *)ptr);
    exp = *ptr;
    
    free((void *)ptr);
    if (d != exp) {
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%08x %08x d=%08x %08x \n", (unsigned int)(exp >> 32),
                                                     (unsigned int)exp,
                                                     (unsigned int)(d >> 32),
                                                     (unsigned int)d);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}
#endif

int test_ldbrx()
{
    unsigned long long *ptr, d, exp, mask0, mask1, mask2, mask3;
    DEBUG(1, "__ldbrx............");

    ptr = (unsigned long long *)malloc(sizeof(unsigned long long));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    *ptr = rand_ullong();
    d = __ldbrx(ptr);
    /* Reverse the endian order */
    mask0 = 0xFF000000ULL;
    mask1 = 0x00FF0000ULL;
    mask2 = 0x0000FF00ULL;
    mask3 = 0x000000FFULL;
    exp = ((*ptr & (mask0 << 32)) >> 56) | 
          ((*ptr & (mask1 << 32)) >> 40) |
          ((*ptr & (mask2 << 32)) >> 24) |
          ((*ptr & (mask3 << 32)) >> 8) |
          ((*ptr & mask0) << 8) |
          ((*ptr & mask1) << 24) |
          ((*ptr & mask2) << 40) |
          ((*ptr & mask3) << 56);
    
    free((void *)ptr);
    if (d != exp) {
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%08x %08x d=%08x %08x \n", (unsigned int)(exp >> 32),
                                                     (unsigned int)exp,
                                                     (unsigned int)(d >> 32),
                                                     (unsigned int)d);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_lhbrx()
{
    unsigned short *ptr, d, exp;
    DEBUG(1, "__lhbrx............");

    ptr = (unsigned short *)malloc(sizeof(unsigned short));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    *ptr = rand_local();
    d = __lhbrx(ptr);
    /* Reverse the endian order */
    exp = ((*ptr & 0xFF00) >> 8) |
          ((*ptr & 0x00FF) << 8);
    
    free((void *)ptr);
    if (d != exp) {
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%u d=%u \n", exp, d);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_lwarx()
{
    unsigned int *ptr, d, exp;
    DEBUG(1, "__lwarx............");

    ptr = (unsigned int *)malloc(sizeof(unsigned int));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    *ptr = rand_local();
    d = __lwarx((void *)ptr);
    exp = *ptr;
    
    free((void *)ptr);
    if (d != exp) {
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%u d=%u \n", exp, d);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_lwbrx()
{
    unsigned int *ptr, d, exp;
    DEBUG(1, "__lwbrx............");

    ptr = (unsigned int *)malloc(sizeof(unsigned int));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    *ptr = rand_local();
    d = __lwbrx(ptr);
    /* Reverse the endian order */
    exp = ((*ptr & 0xFF000000) >> 24) |
          ((*ptr & 0x00FF0000) >> 8) |
          ((*ptr & 0x0000FF00) << 8) |
          ((*ptr & 0x000000FF) << 24);
    
    free((void *)ptr);
    if (d != exp) {
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%u d=%u \n", exp, d);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_lwsync()
{
    DEBUG(1, "__lwsync...........");

    __lwsync();
  
    DEBUG(1, passed);
    return (0);
}

int test_mffs_setflm()
{
    union {
      double d;
      unsigned long long ull;
    } a, b, c, d, e;

    DEBUG(1, "__mffs,__setflm....");

    a.d = __mffs();
    b.ull = a.ull & 0xFFFFFFFFFDFFFFFFULL;
    c.d = __setflm(b.d);
    d.d = __mffs();

    /* Restore the original value */
    __mtfsf(0xFF, a.d);
    
    e.d = __mffs();
    
    if (b.ull != d.ull) {
        DEBUG(1, failed);
	DEBUG(2, "\t\tsetflm failed expected=0x%016llx got=0x%016llx\n",
	      b.ull, d.ull);
	return (1);
    }

    if (a.ull != c.ull) {
        DEBUG(1, failed);
	DEBUG(2, "\t\tsetflm return value failed expected=0x%016llx got=0x%016llx\n",
	      a.ull, c.ull);
	return (1);
    }

    if (a.ull != e.ull) {
        DEBUG(1, failed);
	DEBUG(2, "\t\tmtfsf failed expected=0x%016llx got=0x%016llx\n",
	      a.ull, e.ull);
	return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

/* mfspr,mtspr are only supported in 64-bit */
#if defined (__powerpc64__)
int test_mfspr_mtspr()
{
    unsigned long long a, d, exp;
    DEBUG(1, "__mfspr,__mtspr....");

    a = rand_ullong();
    __mtspr(9, a);
    d = __mfspr(9);
    exp = a;
    
    if (d != exp) {
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%08x %08x d=%08x %08x\n", (unsigned int)(exp >> 32),
                                                    (unsigned int)exp,
                                                    (unsigned int)(d >> 32),
                                                    (unsigned int)d);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}
#endif

int test_mftb()
{
    unsigned long long d;
    DEBUG(1, "__mftb.............");

    d = __mftb();
  
    DEBUG(1, passed);
    return (0);
}

int test_mtfsb0()
{
    volatile unsigned long long *aptr, *dptr, exp, mask;
    volatile double a, d, *dp1, *dp2;
    DEBUG(1, "__mtfsb0...........");

    dp1 = (double*)malloc(sizeof(double));
    dp2 = (double*)malloc(sizeof(double));
    if ((dp2 == NULL) || (dp1 == NULL)) {
      if (dp1) free((void *)dp1);
      if (dp2) free((void *)dp2);
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    
    a = __mffs();
    *dp1 = a;
    aptr = (unsigned long long*)dp1;
    __mtfsb0(6);
    d = __mffs();
    *dp2 = d;
    dptr = (unsigned long long*)dp2;
    mask = 0xFFFFFFFF;
    mask <<= 32;
    mask |= 0xFDFFFFFF;
    exp = (*aptr) & mask;
    /* Restore the original value */
    __mtfsf(0xFF, a);

    if (exp != *dptr) {
        free((void *)dp1);
        free((void *)dp2);
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%08x %08x d=%08x %08x\n", (unsigned int)(exp >> 32), 
                                                    (unsigned int)exp,
                                                    (unsigned int)(*dptr >> 32),
                                                    (unsigned int)*dptr);
        return (1);
    }
  
    free((void *)dp1);
    free((void *)dp2);
    DEBUG(1, passed);
    return (0);
}

int test_mtfsb1()
{
    volatile unsigned long long *aptr, *dptr, exp;
    volatile double a, d, *dp1, *dp2;
    DEBUG(1, "__mtfsb1...........");

    dp1 = (double*)malloc(sizeof(double));
    dp2 = (double*)malloc(sizeof(double));
    if ((dp2 == NULL) || (dp1 == NULL)) {
      if (dp1) free((void *)dp1);
      if (dp2) free((void *)dp2);
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    
    a = __mffs();
    *dp1 = a;
    aptr = (unsigned long long *)dp1;
    __mtfsb1(3);
    d = __mffs();
    *dp2 = d;
    dptr = (unsigned long long*)dp2;
    exp = (*aptr) | 0x10000000;
    /* Restore the original value */
    __mtfsf(0xFF, a);
    
    if (exp != *dptr) {
        free((void *)dp1);
        free((void *)dp2);
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%08x %08x d=%08x %08x\n", (unsigned int)(exp >> 32), 
                                                    (unsigned int)exp,
                                                    (unsigned int)(*dptr >> 32),
                                                    (unsigned int)*dptr);
        return (1);
    }
    
    free((void *)dp1);
    free((void *)dp2);
    DEBUG(1, passed);
    return (0);
}

int test_mtfsf()
{
    volatile double b, d, save, *dp1, *dp2, *dp3;
    volatile unsigned long long exp, *saveptr, *bptr, *dptr, mask;
    DEBUG(1, "__mtfsf............");

    dp1 = (double*)malloc(sizeof(double));
    dp2 = (double*)malloc(sizeof(double));
    dp3 = (double*)malloc(sizeof(double));
    if ((dp2 == NULL) || (dp1 == NULL) || (dp3 == NULL)) {
      if (dp1) free((void *)dp1);
      if (dp2) free((void *)dp2);
      if (dp3) free((void *)dp3);
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    
    save = __mffs();
    *dp1 = save;
    saveptr = (unsigned long long*)dp1;
    b = save;
    *dp2 = b;
    bptr = (unsigned long long *)dp2;
    mask = 0xFFFFFFFF;
    mask <<= 32;
    mask |= 0xFDFFFFFF;
    *bptr &= mask;
    b = *dp2;
    __mtfsf(0x60, b);
    d = __mffs();
    *dp3 = d;
    dptr = (unsigned long long*)dp3;
    exp = *bptr;
    /* Restore the original value */
    __mtfsf(0xFF, save);
          
    if (exp != *dptr) {
        free((void *)dp1);
        free((void *)dp2);
        free((void *)dp3);
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%08x %08x d=%08x %08x b=%08x %08x\n", 
                                                    (unsigned int)(exp >> 32), 
                                                    (unsigned int)exp,
                                                    (unsigned int)(*dptr >> 32),
                                                    (unsigned int)*dptr,
                                                    (unsigned int)(*bptr >> 32),
                                                    (unsigned int)*bptr);
        return (1);
    }
    
    free((void *)dp1);
    free((void *)dp2);
    free((void *)dp3);
    DEBUG(1, passed);
    return (0);
}

int test_mtfsfi()
{
    volatile double save, d, *dp1, *dp2;
    volatile unsigned long long exp, *saveptr, *dptr, mask;
    DEBUG(1, "__mtfsfi...........");

    dp1 = (double *)malloc(sizeof(double));
    dp2 = (double *)malloc(sizeof(double));
    if ((dp2 == NULL) || (dp1 == NULL)) {
      if (dp1) free((void *)dp1);
      if (dp2) free((void *)dp2);
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }

    save = __mffs();
    *dp1 = save;
    saveptr = (unsigned long long *)dp1;
    __mtfsfi(2, 4);
    d = __mffs();
    *dp2 = d;
    dptr = (unsigned long long*)dp2;
    mask = 0xFFFFFFFF;
    mask <<= 32;
    mask |= 0xFF0FFFFF;
    exp = (*saveptr & mask) | 
          0x20400000; /* VX bit (2) should be set also */
    /* Restore the original value */
    __mtfsf(0xFF, save);
    
    if (exp != *dptr) {
        free((void *)dp1);
        free((void *)dp2);
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%08x %08x d=%08x %08x save=%08x %08x\n", 
                                          (unsigned int)(exp >> 32), 
                                          (unsigned int)exp,
                                          (unsigned int)(*dptr >> 32),
                                          (unsigned int)*dptr,
                                          (unsigned int)(*saveptr >> 32),
                                          (unsigned int)*saveptr);
        return (1);
    }
    
    free((void *)dp1);
    free((void *)dp2);
    DEBUG(1, passed);
    return (0);
}

/* mulhdu and mulhd are only supported in 64-bit */
#if defined (__powerpc64__)
int test_mulhdu()
{
    int i;
    unsigned long long a, b, d, exp, ah, al, bh, bl;
    DEBUG(1, "__mulhdu...........");

    for (i=0; i<SAMPLES; i++) {
      a = rand_ullong();
      b = rand_ullong();
      d = __mulhdu(a, b);
      ah = a >> 32;
      al = a & 0xFFFFFFFFULL;
      bh = b >> 32;
      bl = b & 0xFFFFFFFFULL;
      exp = ah*bh + (al*bh >> 32) + (ah*bl >> 32) + 
	(((al*bl >> 32) + ((ah*bl & 0xFFFFFFFFULL) +
			   (al*bh & 0xFFFFFFFFULL))) >> 32);
			  
      if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%llu b=%llu d=%llu exp=%llu\n", a, b, d, exp);
        return (1);
      }
    }
    
    DEBUG(1, passed);
    return (0);
}

int test_mulhd()
{
    int i;
    unsigned long long ah, al, bh, bl, abs_a, abs_b, eh, el;
    long long a, b, d, exp;
    DEBUG(1, "__mulhd............");

    for (i=0; i<SAMPLES; i++) {
      a = (long long)rand_ullong();
      b = (long long)rand_ullong();
      d = __mulhd(a, b);

      abs_a = llabs(a);
      abs_b = llabs(b);
      ah = abs_a >> 32;
      al = abs_a & 0xFFFFFFFFULL;
      bh = abs_b >> 32;
      bl = abs_b & 0xFFFFFFFFULL;

      eh = ah*bh + (al*bh >> 32) + (ah*bl >> 32) + 
	(((al*bl >> 32) + ((ah*bl & 0xFFFFFFFFULL) +
			   (al*bh & 0xFFFFFFFFULL))) >> 32);
      el = abs_a * abs_b;

      exp = eh ^ ((a ^ b) >> 63);
      if (el == 0) exp++;

      if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%lld b=%lld d=%lld exp=%lld\n", a, b, d, exp);
        return (1);
      }
    }
    DEBUG(1, passed);
    return (0);
}
#endif

int test_mulhwu()
{
    int i;
    unsigned int a, b, d;
    unsigned long long exp;
    DEBUG(1, "__mulhwu...........");

    for (i=0; i<SAMPLES; i++) {
      a = rand_local();
      b = rand_local();
      d = __mulhwu(a, b);
      exp = ((unsigned long long)a*b) >> 32;
  
      if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%u b=%u d=%u exp=%u\n", a, b, d, (unsigned int)exp);
        return (1);
      }
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_mulhw()
{
    int i;
    int a, b, d;
    long long exp;
    DEBUG(1, "__mulhw............");

    for (i=0; i<SAMPLES; i++) {
      a = rand_local();
      b = rand_local();
      d = __mulhw(a, b);
      exp = ((long long)a*b) >> 32;
  
      if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%d b=%d d=%d exp=%d\n", a, b, d, (int)exp);
        return (1);
      }
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_nop()
{
    DEBUG(1, "__nop..............");

    __nop();
  
    DEBUG(1, passed);
    return (0);
}

/* rldcl, rldcr, rldic, rldicl, rldicr, rldimi intrinsics are 
 * only supported in 64-bit */
#if defined (__powerpc64__)
int test_rldcl()
{
    unsigned long long a, b, d, exp, c, mask;
    unsigned int count;
    DEBUG(1, "__rldcl............");

    a = rand_ullong();
    b = rand_ullong();
    d = __rldcl(a, b, 37);
    
    /* Rotate left by b bits */
    count = b & 0x3F; /* b mod 64 */
    if (count != 0) {
        exp = (a << count) | (a >> (64 - count));
    }
    else {
        exp = a;
    }
    
    /* Clear left */
    c = 0xFFFFFFFF;
    c <<= 32;
    c |= 0xFFFFFFFF;
    mask = c >> 37;
    exp &= mask;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%08x %08x count=%u d=%08x %08x exp=%08x %08x\n", 
                                  (unsigned int)(a >> 32),
                                  (unsigned int)a, 
                                  count, 
                                  (unsigned int)(d >> 32),
                                  (unsigned int)d, 
                                  (unsigned int)(exp >> 32),
                                  (unsigned int)exp);
        return (1);
    }
        
    DEBUG(1, passed);
    return (0);
}

int test_rldcr()
{
    unsigned long long a, b, d, exp, c, mask;
    unsigned int count;
    DEBUG(1, "__rldcr............");

    a = rand_ullong();
    b = rand_ullong();
    d = __rldcr(a, b, 37);
    
    /* Rotate left by b bits */
    count = b & 0x3F; /* b mod 64 */
    if (count != 0) {
        exp = (a << count) | (a >> (64 - count));
    }
    else {
        exp = a;
    }
    
    /* Clear right */
    c = 0xFFFFFFFF;
    c <<= 32;
    c |= 0xFFFFFFFF;
    mask = c << (64-(37+1));
    exp &= mask;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%08x %08x count=%u d=%08x %08x exp=%08x %08x\n", 
                                  (unsigned int)(a >> 32),
                                  (unsigned int)a, 
                                  count, 
                                  (unsigned int)(d >> 32),
                                  (unsigned int)d, 
                                  (unsigned int)(exp >> 32),
                                  (unsigned int)exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_rldic()
{
    unsigned long long a, d, exp, c, mask1, mask2;
    unsigned int count;
    DEBUG(1, "__rldic............");

    a = rand_ullong();
    d = __rldic(a, 27, 17);
    
    /* Rotate left */
    count = 27;
    exp = (a << count) | (a >> (64 - count));
    
    /* Clear left */
    c = 0xFFFFFFFF;
    c <<= 32;
    c |= 0xFFFFFFFF;
    mask1 = c >> 17;
    mask2 = c << 27;
    exp &= mask1 & mask2;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%08x %08x d=%08x %08x exp=%08x %08x\n", 
                                  (unsigned int)(a >> 32),
                                  (unsigned int)a, 
                                  (unsigned int)(d >> 32),
                                  (unsigned int)d, 
                                  (unsigned int)(exp >> 32),
                                  (unsigned int)exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_rldicl()
{
    unsigned long long a, d, exp, c, mask;
    unsigned int count;
    DEBUG(1, "__rldicl...........");

    a = rand_ullong();
    d = __rldicl(a, 27, 17);
    
    /* Rotate left */
    count = 27;
    exp = (a << count) | (a >> (64 - count));
    
    /* Clear left */
    c = 0xFFFFFFFF;
    c <<= 32;
    c |= 0xFFFFFFFF;
    mask = c >> 17;
    exp &= mask;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%08x %08x d=%08x %08x exp=%08x %08x\n", 
                                  (unsigned int)(a >> 32),
                                  (unsigned int)a, 
                                  (unsigned int)(d >> 32),
                                  (unsigned int)d, 
                                  (unsigned int)(exp >> 32),
                                  (unsigned int)exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_rldicr()
{
    unsigned long long a, d, exp, c, mask;
    unsigned int count;
    DEBUG(1, "__rldicr...........");

    a = rand_ullong();
    d = __rldicr(a, 27, 17);
    
    /* Rotate left */
    count = 27;
    exp = (a << count) | (a >> (64 - count));
    
    /* Clear right */
    c = 0xFFFFFFFF;
    c <<= 32;
    c |= 0xFFFFFFFF;
    mask = c << (64-(17+1));
    exp &= mask;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%08x %08x d=%08x %08x exp=%08x %08x\n", 
                                  (unsigned int)(a >> 32),
                                  (unsigned int)a, 
                                  (unsigned int)(d >> 32),
                                  (unsigned int)d, 
                                  (unsigned int)(exp >> 32),
                                  (unsigned int)exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_rldimi()
{
    unsigned long long a, b, d, exp, exp1, exp2, c, mask1, mask2, mask;
    unsigned int count;
    DEBUG(1, "__rldimi...........");

    a = rand_ullong();
    b = rand_ullong();
    d = __rldimi(a, b, 27, 17);
    
    /* Generate mask where bits 17 through 63-27 are 1s*/
    c = 0xFFFFFFFF;
    c <<= 32;
    c |= 0xFFFFFFFF;
    mask1 = c >> 17;
    mask2 = c << 27;
    mask = mask1 & mask2;
    
    /* Mask a with the complement of the mask */
    exp1 = a & (~mask);
    
    /* Rotate b left and mask */
    count = 27;
    exp2 = (b << count) | (b >> (64 - count));
    exp2 &= mask;
    
    /* Combine the results */
    exp = exp1 | exp2;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%08x %08x b=%08x %08x d=%08x %08x exp=%08x %08x\n", 
                                  (unsigned int)(a >> 32),
                                  (unsigned int)a, 
                                  (unsigned int)(b >> 32),
                                  (unsigned int)b,
                                  (unsigned int)(d >> 32),
                                  (unsigned int)d, 
                                  (unsigned int)(exp >> 32),
                                  (unsigned int)exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}
#endif

int test_rlwimi()
{
    unsigned int a, b, d, exp, exp1, exp2, c, mask1, mask2, mask, count;
    DEBUG(1, "__rlwimi...........");

    a = rand_ullong();
    b = rand_ullong();
    d = __rlwimi(a, b, 27, 17, 11);
    
    /* Generate mask where bits 17 through 11 are 1s*/
    c = 0xFFFFFFFF;
    mask1 = c >> 11;
    mask2 = c << (32-(17+1));
    mask = mask1 & mask2;
    
    /* Mask a with the complement of the mask */
    exp1 = a & (~mask);
    
    /* Rotate b left and mask */
    count = 27;
    exp2 = (b << count) | (b >> (32 - count));
    exp2 &= mask;
    
    /* Combine the results */
    exp = exp1 | exp2;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%08x b=%08x d=%08x exp=%08x\n", a, b, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_rlwinm()
{
    unsigned int a, d, exp, c, mask1, mask2, mask, count;
    DEBUG(1, "__rlwinm...........");

    a = rand_ullong();
    d = __rlwinm(a, 27, 11, 17);
    
    /* Generate mask where bits 17 through 11 are 1s*/
    c = 0xFFFFFFFF;
    mask1 = c >> 11;
    mask2 = c << (32-(17+1));
    mask = mask1 & mask2;
    
    /* Rotate a left */
    count = 27;
    exp = (a << count) | (a >> (32 - count));
    exp &= mask;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%08x d=%08x exp=%08x\n", a, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_rlwnm()
{
    unsigned int a, b, d, exp, c, mask1, mask2, mask, count;
    DEBUG(1, "__rlwnm............");

    a = rand_ullong();
    b = rand_ullong();
    d = __rlwnm(a, b, 11, 17);
    
    /* Generate mask where bits 17 through 11 are 1s*/
    c = 0xFFFFFFFF;
    mask1 = c >> 11;
    mask2 = c << (32-(17+1));
    mask = mask1 & mask2;
    
    /* Rotate a left */
    count = b & 0x1F; /* b mod 32 */
    if (count != 0) {
        exp = (a << count) | (a >> (32 - count));
    }
    else {
        exp = a;
    }
    exp &= mask;
    
    if (exp != d) {
        DEBUG(1, failed);
        DEBUG(2, "\t\ta=%08x b=%08x d=%08x exp=%08x\n", a, b, d, exp);
        return (1);
    }
  
    DEBUG(1, passed);
    return (0);
}

int test_stdbrx()
{
    unsigned long long *ptr, b, exp, mask0, mask1, mask2, mask3;
    DEBUG(1, "__stdbrx...........");

    ptr = (unsigned long long *)malloc(sizeof(unsigned long long));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    b = rand_ullong();
    __stdbrx(ptr, b);
    /* Reverse the endian order */
    mask0 = 0xFF000000ULL;
    mask1 = 0x00FF0000ULL;
    mask2 = 0x0000FF00ULL;
    mask3 = 0x000000FFULL;
    exp = ((b & (mask0 << 32)) >> 56) | 
          ((b & (mask1 << 32)) >> 40) |
          ((b & (mask2 << 32)) >> 24) |
          ((b & (mask3 << 32)) >> 8) |
          ((b & mask0) << 8) |
          ((b & mask1) << 24) |
          ((b & mask2) << 40) |
          ((b & mask3) << 56);
    
    if (*ptr != exp) {
        free((void *)ptr);
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%08x %08x *ptr=%08x %08x \n", (unsigned int)(exp >> 32),
                                                        (unsigned int)exp, 
                                                        (unsigned int)(*ptr >> 32),
                                                        (unsigned int)*ptr);
        return (1);
    }
  
    free((void *)ptr);
    DEBUG(1, passed);
    return (0);
}

/* stdcx is only supported in 64-bit */
#if defined (__powerpc64__)
int test_stdcx()
{
    unsigned long long *ptr, b, c;
    unsigned int d;
    DEBUG(1, "__stdcx............");
    
    ptr = (unsigned long long *)malloc(sizeof(unsigned long long));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    *ptr = rand_ullong();
    c = __ldarx((void *)ptr);
    b = rand_ullong();
    d = __stdcx((void *)ptr, b);
    
    if (d == 0 || *ptr != b) {
        free((void *)ptr);
        DEBUG(1, failed);
        DEBUG(2, "\t\td=%d *ptr=%llu b=%llu c=%llu\n", d, *ptr, b, c);
        return (1);
    }
  
    free((void *)ptr);
    DEBUG(1, passed);
    return (0);
}
#endif


int test_sthbrx()
{
    unsigned short *ptr, b, exp;
    DEBUG(1, "__sthbrx...........");

    ptr = (unsigned short *)malloc(sizeof(unsigned short));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    b = rand_local();
    __sthbrx(ptr, b);
    /* Reverse the endian order */
    exp = ((b & 0xFF00) >> 8) |
          ((b & 0x00FF) << 8);
    
    if (*ptr != exp) {
        free((void *)ptr);
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%u *ptr=%u \n", exp, *ptr);
        return (1);
    }
  
    free((void *)ptr);
    DEBUG(1, passed);
    return (0);
}

int test_stwbrx()
{
    unsigned int *ptr, b, exp;
    DEBUG(1, "__stwbrx...........");

    ptr = (unsigned int *)malloc(sizeof(unsigned int));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    b = rand_local();
    __stwbrx(ptr, b);
    /* Reverse the endian order */
    exp = ((b & 0xFF000000) >> 24) |
          ((b & 0x00FF0000) >> 8) |
          ((b & 0x0000FF00) << 8) |
          ((b & 0x000000FF) << 24);
    
    if (*ptr != exp) {
        free((void *)ptr);
        DEBUG(1, failed);
        DEBUG(2, "\t\texp=%u *ptr=%u \n", exp, *ptr);
        return (1);
    }
  
    free((void *)ptr);
    DEBUG(1, passed);
    return (0);
}

int test_stwcx()
{
    unsigned int *ptr, b, c;
    unsigned int d;
    DEBUG(1, "__stwcx............");
    
    ptr = (unsigned int *)malloc(sizeof(unsigned int));
    if (ptr == NULL) {
      DEBUG(1, failed);
      DEBUG(2, "\t\tmemory allocation failed\n");
      return (1);      
    }
    *ptr = rand_local();
    c = __lwarx((void *)ptr);
    b = rand_local();
    d = __stwcx((void *)ptr, b);
    
    if (d == 0 || *ptr != b) {
        free((void *)ptr);
        DEBUG(1, failed);
        DEBUG(2, "\t\td=%d *ptr=%u b=%u c=%u\n", d, *ptr, b, c);
        return (1);
    }
  
    free((void *)ptr);
    DEBUG(1, passed);
    return (0);
}

int test_sync()
{
    DEBUG(1, "__sync.............");

    __sync();
  
    DEBUG(1, passed);
    return (0);
}



DECLARE_MAIN()
{
  DECLARE_TEST_VARS();

  INVOKE_TEST(test_cctph);
  INVOKE_TEST(test_cctpl);
  INVOKE_TEST(test_cctpm);
  INVOKE_TEST(test_cntlzd);
  INVOKE_TEST(test_cntlzw);
  INVOKE_TEST(test_db10cyc);
  INVOKE_TEST(test_db12cyc);
  INVOKE_TEST(test_db16cyc);
  INVOKE_TEST(test_db8cyc);
  INVOKE_TEST(test_dcbf);
  INVOKE_TEST(test_dcbst);
  INVOKE_TEST(test_dcbt);
  INVOKE_TEST(test_dcbt_TH1000);
  INVOKE_TEST(test_dcbt_TH1010);
  INVOKE_TEST(test_dcbtst);
  INVOKE_TEST(test_dcbz);
  INVOKE_TEST(test_eieio);
  INVOKE_TEST(test_fabs);
  INVOKE_TEST(test_fabsf);
  INVOKE_TEST(test_fcfid);
  INVOKE_TEST(test_fctid);
  INVOKE_TEST(test_fctidz);
  INVOKE_TEST(test_fctiw);
  INVOKE_TEST(test_fctiwz);
  INVOKE_TEST(test_fmadd);
  INVOKE_TEST(test_fmadds);
  INVOKE_TEST(test_fmsub);
  INVOKE_TEST(test_fmsubs);
  INVOKE_TEST(test_fmul);
  INVOKE_TEST(test_fmuls);
  INVOKE_TEST(test_fnabs);
  INVOKE_TEST(test_fnabsf);
  INVOKE_TEST(test_fnmadd);
  INVOKE_TEST(test_fnmadds);
  INVOKE_TEST(test_fnmsub);
  INVOKE_TEST(test_fnmsubs);
  INVOKE_TEST(test_fres);
  INVOKE_TEST(test_frsp);
  INVOKE_TEST(test_frsqrte);
  INVOKE_TEST(test_fsel);
  INVOKE_TEST(test_fsels);
  INVOKE_TEST(test_fsqrt);
  INVOKE_TEST(test_fsqrts);
  INVOKE_TEST(test_icbi);
  INVOKE_TEST(test_isync);
#if defined (__powerpc64__)
  INVOKE_TEST(test_ldarx); /* ldarx is only supported in 64-bit */
#endif
  INVOKE_TEST(test_ldbrx);
  INVOKE_TEST(test_lhbrx);
  INVOKE_TEST(test_lwarx);
  INVOKE_TEST(test_lwbrx);
  INVOKE_TEST(test_lwsync);
  INVOKE_TEST(test_mffs_setflm);
#if defined (__powerpc64__)
  INVOKE_TEST(test_mfspr_mtspr); /* mfspr, mtspr are only supported in 64-bit */
#endif
  INVOKE_TEST(test_mftb);
  INVOKE_TEST(test_mtfsb0);
  INVOKE_TEST(test_mtfsb1);
  INVOKE_TEST(test_mtfsf);
  INVOKE_TEST(test_mtfsfi);
#if defined (__powerpc64__)
  INVOKE_TEST(test_mulhdu); /* mulhdu and mulhd is only supported in 64-bit */
  INVOKE_TEST(test_mulhd);
#endif
  INVOKE_TEST(test_mulhwu);
  INVOKE_TEST(test_mulhw);
  INVOKE_TEST(test_nop);
#if defined (__powerpc64__)
  INVOKE_TEST(test_rldcl); /* rldcl, rldcr, rldic, rldicl, rldicr, rldimi */
  INVOKE_TEST(test_rldcr); /* intrinsics are only supported in 64-bit */
  INVOKE_TEST(test_rldic);
  INVOKE_TEST(test_rldicl);
  INVOKE_TEST(test_rldicr);
  INVOKE_TEST(test_rldimi);
#endif
  INVOKE_TEST(test_rlwinm);
  INVOKE_TEST(test_rlwnm);
  INVOKE_TEST(test_stdbrx);
#if defined (__powerpc64__)
  INVOKE_TEST(test_stdcx); /* stdcx is only supported in 64-bit */
#endif
  INVOKE_TEST(test_sthbrx);
  INVOKE_TEST(test_stwbrx);
  INVOKE_TEST(test_stwcx);
  INVOKE_TEST(test_sync);
  
  DISPLAY_RESULTS();
}

