/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2007,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#include <stdlib.h>
#include <stdio.h>
#include "spu_ext.h"
#include "libtest.h"
#include "libmisc.h"

static int samples = 100;

int test_ab()
{
  int i, j;
  union {
    unsigned char c[16];
    vector unsigned char v;
  } in_a, in_b, out;
  unsigned char exp;

  DEBUG(1, "AB..............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned char)rand_v();
    in_b.v = (vector unsigned char)rand_v();
    
    AB(out.v, in_a.v, in_b.v)
    
    for (j=0; j<16; j++) {
      exp = in_a.c[j] + in_b.c[j];
      if (exp != out.c[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x%x\n", exp, out.c[j], in_a.c[j], in_b.c[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_abi()
{
  int i, j;
  union {
    unsigned char c[16];
    vector unsigned char v;
  } in_a, out;
  unsigned char exp;

  DEBUG(1, "ABI.............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned char)rand_v();
    
    ABI(out.v, in_a.v, 23)
    
    for (j=0; j<16; j++) {
      exp = in_a.c[j] + 23;
      if (exp != out.c[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=23\n", exp, out.c[j], in_a.c[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}

int test_mpyah()
{
  int i, j;
  union {
    signed char c[16];
    vector signed char v;
  } in_a, in_b;
  union {
    signed short s[8];
    vector signed short v;
  } in_c, out;
  signed short exp;

  DEBUG(1, "MPYAH...........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector signed char)rand_v();
    in_b.v = (vector signed char)rand_v();
    in_c.v = (vector signed short)rand_v();
    
    MPYAH(out.v, in_a.v, in_b.v, in_c.v)
    
    for (j=0; j<8; j++) {
      exp = (signed short)(in_a.c[2*j+1] * in_b.c[2*j+1] + in_c.s[j]);;
      if (exp != out.s[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=%d got=%d ", exp, out.s[j]);
	DEBUG(2, "a=%d b=%d c=%d\n", in_a.c[2*j+1], in_b.c[2*j+1], in_c.s[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}



int test_rotb()
{
  int i, j;
  union {
    unsigned char c[16];
    vector unsigned char v;
  } in_a, in_b, out;
  unsigned char exp;
  unsigned char cnt;

  DEBUG(1, "ROTB............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned char)rand_v();
    in_b.v = (vector unsigned char)rand_v();

    ROTB(out.v, in_a.v, in_b.v)
    
    for (j=0; j<16; j++) {
      cnt = in_b.c[j] & 7;
      exp = in_a.c[j];
      if (cnt) {
	exp = (exp << cnt) | (exp >> (8-cnt));
      }

      if (exp != out.c[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x%x\n", exp, out.c[j], in_a.c[j], in_b.c[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_rotbi()
{
  int i, j;
  union {
    unsigned char c[16];
    vector unsigned char v;
  } in_a, out;
  unsigned char exp;

  DEBUG(1, "ROTBI...........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned char)rand_v();
    
    ROTBI(out.v, in_a.v, 11)
    
    for (j=0; j<16; j++) {
      exp = in_a.c[j];
      exp = (exp << 3) | (exp >> (8-3));

      if (exp != out.c[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=11\n", exp, out.c[j], in_a.c[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_rotbm()
{
  int i, j;
  union {
    unsigned char c[16];
    vector unsigned char v;
  } in_a, in_b, out;
  unsigned char exp;
  unsigned char cnt;

  DEBUG(1, "ROTBM...........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned char)rand_v();
    in_b.v = (vector unsigned char)rand_v();
    
    ROTBM(out.v, in_a.v, in_b.v)
    
    for (j=0; j<16; j++) {
      cnt = in_b.c[j] & 15;
      exp = in_a.c[j];
      if (cnt) {
	if (cnt < 9) {
	  exp = 0;
	} else {
	  exp = exp >> (16-cnt);
	}
      }

      if (exp != out.c[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x%x\n", exp, out.c[j], in_a.c[j], in_b.c[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_rotbmi()
{
  int i, j;
  union {
    unsigned char c[16];
    vector unsigned char v;
  } in_a, out;
  unsigned char exp;

  DEBUG(1, "ROTBMI..........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned char)rand_v();
    
    ROTBMI(out.v, in_a.v, -3)
    
    for (j=0; j<16; j++) {
      exp = in_a.c[j] >> 3;

      if (exp != out.c[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=-3\n", exp, out.c[j], in_a.c[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_rotmab()
{
  int i, j;
  union {
    signed char c[16];
    vector signed char v;
  } in_a, in_b, out;
  signed char exp;
  unsigned char cnt;

  DEBUG(1, "ROTMAB..........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector signed char)rand_v();
    in_b.v = (vector signed char)rand_v();
    
    ROTMAB(out.v, in_a.v, in_b.v)
    
    for (j=0; j<16; j++) {
      cnt = in_b.c[j] & 15;
      exp = in_a.c[j];
      if (cnt) {
	if (cnt < 9) {
	  exp = (exp < 0) ? 0xFF : 0;
	} else {
	  exp = exp >> (16-cnt);
	}
      }

      if (exp != out.c[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x%x\n", exp, out.c[j], in_a.c[j], in_b.c[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}

int test_rotmabi()
{
  int i, j;
  union {
    signed char c[16];
    vector signed char v;
  } in_a, out;
  signed char exp;

  DEBUG(1, "ROTMABI.........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector signed char)rand_v();
    
    ROTMABI(out.v, in_a.v, -5)
    
    for (j=0; j<16; j++) {
      exp = in_a.c[j] >> 5;

      if (exp != out.c[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=-5\n", exp, out.c[j], in_a.c[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_sfb()
{
  int i, j;
  union {
    signed char c[16];
    vector signed char v;
  } in_a, in_b, out;
  signed char exp;

  DEBUG(1, "SFB.............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector signed char)rand_v();
    in_b.v = (vector signed char)rand_v();
    
    SFB(out.v, in_a.v, in_b.v)
    
    for (j=0; j<16; j++) {
      exp = in_b.c[j] - in_a.c[j];
      if (exp != out.c[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x%x\n", exp, out.c[j], in_a.c[j], in_b.c[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_sfbi()
{
  int i, j;
  union {
    signed char c[16];
    vector signed char v;
  } in_a, out;
  signed char exp;

  DEBUG(1, "SFBI............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector signed char)rand_v();
     
    SFBI(out.v, in_a.v, 32)
    
    for (j=0; j<16; j++) {
      exp = 32 - in_a.c[j];
      if (exp != out.c[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=32\n", exp, out.c[j], in_a.c[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_shlb()
{
  int i, j;
  union {
    unsigned char c[16];
    vector unsigned char v;
  } in_a, in_b, out;
  unsigned char exp;
  unsigned char cnt;

  DEBUG(1, "SHLB............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned char)rand_v();
    in_b.v = (vector unsigned char)rand_v();
    
    SHLB(out.v, in_a.v, in_b.v)
    
    for (j=0; j<16; j++) {
      cnt = in_b.c[j] & 15;
      exp = in_a.c[j];
      if (cnt) {
	if (cnt > 7) {
	  exp = 0;
	} else {
	  exp = exp << cnt;
	}
      }

      if (exp != out.c[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x%x\n", exp, out.c[j], in_a.c[j], in_b.c[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}

int test_shlbi()
{
  int i, j;
  union {
    unsigned char c[16];
    vector unsigned char v;
  } in_a, out;
  unsigned char exp;

  DEBUG(1, "SHLBI...........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned char)rand_v();
    
    SHLBI(out.v, in_a.v, 0x31)
    
    for (j=0; j<16; j++) {
      exp = in_a.c[j] << 1;

      if (exp != out.c[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x31\n", exp, out.c[j], in_a.c[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_ad()
{
  int i, j;
  union {
    unsigned int i[4];
    vector unsigned long long v;
  } in_a, in_b, out;
  unsigned int exp, low, max;

  DEBUG(1, "AD..............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned long long)rand_v();
    in_b.v = (vector unsigned long long)rand_v();
    
    AD(out.v, in_a.v, in_b.v)
    
    for (j=0; j<4; j++) {
      exp = in_a.i[j] + in_b.i[j];
      if ((j & 1) == 0) {
	low = in_a.i[j|1] + in_b.i[j|1];
	max = (in_a.i[j|1] > in_b.i[j|1]) ? in_a.i[j|1] : in_b.i[j|1];
	if (low < max) exp += 1;
      }

      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x%x\n", exp, out.i[j], in_a.i[j], in_b.i[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_adi()
{
#ifdef __xlc__
  int i, j;
  union {
    unsigned int i[4];
    vector unsigned long long v;
  } in_a, out;
  unsigned int exp;

  DEBUG(1, "ADI.............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned long long)rand_v();
    
    ADI(out.v, in_a.v, -1750)
    
    for (j=0; j<4; j++) {
      if (j & 1) {
	exp = in_a.i[j] - 1750;
      } else {
	exp = in_a.i[j];
	if (in_a.i[j|1] < 1750) exp -= 1;
      }

      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=-1750\n", exp, out.i[j], in_a.i[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
#endif /* __xlc__ */
  return (0);
}

#define VERIFY_DBL(_out, _exp1, _exp2, _exp3, _exp4)						\
  if ((_out.i[0] != _exp1) ||									\
      (_out.i[1] != _exp2) ||									\
      (_out.i[2] != _exp3) ||									\
      (_out.i[3] != _exp4)) {									\
    DEBUG(1, "FAILED\n");							       		\
    DEBUG(2, "\t\tline %d", __LINE__);								\
    DEBUG(1, " out = 0x%x 0x%x 0x%x 0x%x\n", _out.i[0],  _out.i[1],  _out.i[2],  _out.i[3]);	\
    return (1);											\
  }

int test_cdflts()
{
  union {
    unsigned int i[4];
    vector signed long long v;
  } out;
  union {
    vector unsigned int iv;
    vector double v;
  } in;

  DEBUG(1, "CDFLTS..........");

  in.v = ((vector double) { 1.5, -2147483648.0});
  CDFLTS(out.v, in.v, 4)
  VERIFY_DBL(out, 0, 24, 0xFFFFFFF8, 0x0);

  in.v = ((vector double) { 1.0, 13.0});
  CDFLTS(out.v, in.v, 0)
  VERIFY_DBL(out, 0, 1, 0, 13);

  in.v = ((vector double) { -6.0, 13.99});
  CDFLTS(out.v, in.v, 0)
  VERIFY_DBL(out, 0xFFFFFFFF, 0xFFFFFFFA, 0, 13);

  in.v = ((vector double) { 0.0, -0.0});
  CDFLTS(out.v, in.v, 0)
  VERIFY_DBL(out, 0, 0, 0, 0);

  in.v = ((vector double) { 9.0, -9.0});
  CDFLTS(out.v, in.v, 60)
  VERIFY_DBL(out, 0x7FFFFFFF, 0xFFFFFFFF, 0x80000000, 0x0);


  in.v = ((vector double) { 0.034765625, -0.01953125});
  CDFLTS(out.v, in.v, 8)
  VERIFY_DBL(out, 0, 8, 0xFFFFFFFF, 0xFFFFFFFB);

  DEBUG(1, "passed\n");
  return (0);
}


int test_cdfltu()
{
  union {
    unsigned int i[4];
    vector unsigned long long v;
  } out;
  union {
    vector unsigned int iv;
    vector double v;
  } in;

  DEBUG(1, "CDFLTU..........");

  in.v = ((vector double) { 1.9, 1017.5});
  CDFLTU(out.v, in.v, 0)
  VERIFY_DBL(out, 0, 1, 0, 1017);

  in.v = ((vector double) { 1.75, 0.8});
  CDFLTU(out.v, in.v, 1)
  VERIFY_DBL(out, 0, 3, 0, 1);

  in.v = ((vector double) { 0.5, -0.0});
  CDFLTU(out.v, in.v, 0)
  VERIFY_DBL(out, 0, 0, 0, 0);

  in.v = ((vector double) { 1.0, 2.0});
  CDFLTU(out.v, in.v, 63)
  VERIFY_DBL(out, 0x80000000, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF);

  in.v = ((vector double) { -1.0, 1023.5});
  CDFLTU(out.v, in.v, 2)
  VERIFY_DBL(out, 0, 0, 0, 4094);

  DEBUG(1, "passed\n");
  return (0);
}


int test_ceqd()
{
  int i, j;
  union {
    unsigned int i[4];
    vector unsigned long long v;
  } in_a, in_b, out;
  unsigned int val;
  unsigned int exp = 0;

  DEBUG(1, "CEQD............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned long long)rand_v();
    in_b.v = (vector unsigned long long)rand_v();

    for (j=0, val = (unsigned int)(rand()); j<4; j++, val >>= 1) {
      if (val & 1) in_b.i[j] = in_a.i[j];
    }

    CEQD(out.v, in_a.v, in_b.v)

    for (j=0; j<4; j++) {
      if ((j & 1) == 0) {
	if ((in_a.i[j]   == in_b.i[j]) &&
	    (in_a.i[j+1] == in_b.i[j+1])) {
	  exp = 0xFFFFFFFF;
	} else {
	  exp = 0;
	}
      }

      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x 0x%x", exp, out.i[j], in_a.i[j&2], in_a.i[j|1]);
	DEBUG(2, " b=0x%x 0x%x\n", in_b.i[j&2], in_b.i[j|1]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


#define VERIFY_CMP_D(_out, _exp1, _exp2)							\
  if ((_out.i[0] != _exp1) ||									\
      (_out.i[1] != _exp1) ||									\
      (_out.i[2] != _exp2) ||									\
      (_out.i[3] != _exp2)) {									\
    DEBUG(1, "FAILED\n");										\
    DEBUG(2, "\t\tline %d", __LINE__);								\
    DEBUG(1, " out = 0x%x 0x%x 0x%x 0x%x\n", _out.i[0],  _out.i[1],  _out.i[2],  _out.i[3]);	\
    return (1);											\
  }
    
      
int test_ceqdi()
{
  union {
    int i[4];
    vector unsigned long long v;
  } in_a, out;

  DEBUG(1, "CEQDI...........");

  in_a.v = ((vector unsigned long long) { 0x0000000000000123ULL, 0x0000000000000123ULL});
  CEQDI(out.v, in_a.v, 291)
  VERIFY_CMP_D(out, -1, -1);
  in_a.v = ((vector unsigned long long) { 0x0010000000000123ULL, 0x0000000000000123ULL});
  CEQDI(out.v, in_a.v, 291)
  VERIFY_CMP_D(out, 0, -1);
  in_a.v = ((vector unsigned long long) { 0x0000000000000123ULL, 0x0000000000000124ULL});
  CEQDI(out.v, in_a.v, 291)
  VERIFY_CMP_D(out, -1, 0);
  in_a.v = ((vector unsigned long long) { 0x0000000001000123ULL, 0x0000100000000123ULL});
  CEQDI(out.v, in_a.v, 291)
  VERIFY_CMP_D(out, 0, 0);

  in_a.v = ((vector unsigned long long) { 0xFFFFFFFFFFFFFF23ULL, 0xFFFFFFFFFFFFFF23ULL});
  CEQDI(out.v, in_a.v, -221)
  VERIFY_CMP_D(out, -1, -1);
  in_a.v = ((vector unsigned long long) { 0xFFEFFFFFFFFFFF23ULL, 0xFFFFFFFFFFFFFF23ULL});
  CEQDI(out.v, in_a.v, -221)
  VERIFY_CMP_D(out, 0, -1);
  in_a.v = ((vector unsigned long long) { 0xFFFFFFFFFFFFFF23ULL, 0xFFFFFFFFFFFFFF24ULL});
  CEQDI(out.v, in_a.v, -221)
  VERIFY_CMP_D(out, -1, 0);
  in_a.v = ((vector unsigned long long) { 0xFFFFFFFFFEFFFF23ULL, 0xFFFFEFFFFFFFFF23ULL});
  CEQDI(out.v, in_a.v, -221)
  VERIFY_CMP_D(out, 0, 0);

  DEBUG(1, "passed\n");
  return (0);
}


int test_cgtd()
{
  int i, j;
  union {
    signed int i[4];
    unsigned int ui[4];
    vector signed long long v;
  } in_a, in_b;
  union {
    signed int i[4];
    unsigned int ui[4];
    vector unsigned long long v;
  } out;
  signed int exp = 0;

  DEBUG(1, "CGTD............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector signed long long)rand_v();
    in_b.v = (vector signed long long)rand_v();

    CGTD(out.v, in_a.v, in_b.v)

    for (j=0; j<4; j++) {
      if ((j & 1) == 0) {
	if ((in_a.i[j] > in_b.i[j]) ||
	    ((in_a.i[j] == in_b.i[j]) && (in_a.ui[j+1] > in_b.ui[j+1]))) {
	  exp = 0xFFFFFFFF;
	} else {
	  exp = 0;
	}
      }

      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x 0x%x", exp, out.i[j], in_a.i[j&2], in_a.i[j|1]);
	DEBUG(1, " b=0x%x 0x%x\n", in_b.i[j&2], in_b.i[j|1]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}



int test_cgtdi()
{
  union {
    int i[4];
    vector unsigned long long v;
  } out;
  union {
    int i[4];
    vector signed long long v;
  } in_a;

  DEBUG(1, "CGTDI...........");

  in_a.v = ((vector signed long long) { 0x0000000000000123LL, 0x0000000000000124LL});
  CGTDI(out.v, in_a.v, 291)
  VERIFY_CMP_D(out, 0, -1);
  in_a.v = ((vector signed long long) { 0x0000000000000122LL, 0x0010000000000120LL});
  CGTDI(out.v, in_a.v, 291)
  VERIFY_CMP_D(out, 0, -1);
  in_a.v = ((vector signed long long) { 0x00000000FFFFFFFFLL, 0x8000000000000124LL});
  CGTDI(out.v, in_a.v, 291)
  VERIFY_CMP_D(out, -1, 0);

  in_a.v = ((vector signed long long) { 0xFFFFFFFFFFFFFF23LL, 0xFFFFFFFFFFFFFF24LL});
  CGTDI(out.v, in_a.v, -221)
  VERIFY_CMP_D(out, 0, -1);
  in_a.v = ((vector signed long long) { 0x7FFFFFFFFFFFFF22LL, 0xFFFFFFFFFFFFFF22LL});
  CGTDI(out.v, in_a.v, -221)
  VERIFY_CMP_D(out, -1, 0);
  in_a.v = ((vector signed long long) { 0xFFFFFFFF7FFFFF23LL, 0xFFFFFFFFFFFFFFFFLL});
  CGTDI(out.v, in_a.v, -221)
  VERIFY_CMP_D(out, 0, -1);

  DEBUG(1, "passed\n");
  return (0);
}

int test_clgtd()
{
  int i, j;
  union {
    signed int i[4];
    unsigned int ui[4];
    vector unsigned long long v;
  } in_a, in_b, out;
  signed int exp = 0;

  DEBUG(1, "CLGTD...........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned long long)rand_v();
    in_b.v = (vector unsigned long long)rand_v();

    CLGTD(out.v, in_a.v, in_b.v)

    for (j=0; j<4; j++) {
      if ((j & 1) == 0) {
	if ((in_a.ui[j] > in_b.ui[j]) ||
	    ((in_a.ui[j] == in_b.ui[j]) && (in_a.ui[j+1] > in_b.ui[j+1]))) {
	  exp = 0xFFFFFFFF;
	} else {
	  exp = 0;
	}
      }

      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x 0x%x", exp, out.i[j], in_a.i[j&2], in_a.i[j|1]);
	DEBUG(1, " b=0x%x 0x%x\n", in_b.i[j&2], in_b.i[j|1]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_clgtdi()
{
  union {
    int i[4];
    vector unsigned long long v;
  } in_a, out;

  DEBUG(1, "CLGTDI..........");

  in_a.v = ((vector unsigned long long) { 0x0000000000000123LL, 0x0000000000000124LL});
  CLGTDI(out.v, in_a.v, 291)
  VERIFY_CMP_D(out, 0, -1);
  in_a.v = ((vector unsigned long long) { 0x0000000000000122LL, 0x0010000000000120LL});
  CLGTDI(out.v, in_a.v, 291)
  VERIFY_CMP_D(out, 0, -1);
  in_a.v = ((vector unsigned long long) { 0x00000000FFFFFFFFLL, 0x8000000000000124LL});
  CLGTDI(out.v, in_a.v, 291)
  VERIFY_CMP_D(out, -1, -1);

  in_a.v = ((vector unsigned long long) { 0xFFFFFFFFFFFFFF23LL, 0xFFFFFFFFFFFFFF24LL});
  CLGTDI(out.v, in_a.v, -221)
  VERIFY_CMP_D(out, 0, -1);
  in_a.v = ((vector unsigned long long) { 0x7FFFFFFFFFFFFF24LL, 0xFFFFFFFFFFFFFF22LL});
  CLGTDI(out.v, in_a.v, -221)
  VERIFY_CMP_D(out, 0, 0);
  in_a.v = ((vector unsigned long long) { 0xFFFFFFFF7FFFFF24LL, 0xFFFFFFFFFFFFFFFFLL});
  CLGTDI(out.v, in_a.v, -221)
  VERIFY_CMP_D(out, 0, -1);

  DEBUG(1, "passed\n");
  return (0);
}





int test_csdflt()
{
  union {
    unsigned int i[4];
    vector double v;
  } out;
  vector signed long long in;

  DEBUG(1, "CSDFLT..........");

  in = ((vector signed long long) { 0x0000000000000001LL, 0x8000000000000000LL});
  CSDFLT(out.v, in, 0)
  VERIFY_DBL(out, 0x3ff00000, 0x0, 0xc3e00000, 0x0);

  in = ((vector signed long long) { 0xFFFFFFFFFFFFFFFFLL, 0x0000000000000105LL});
  CSDFLT(out.v, in, 0)
  VERIFY_DBL(out, 0xbff00000, 0x0, 0x40705000, 0x0);

  in = ((vector signed long long) { 0x0000010000000005LL, 0x8000000000000105LL});
  CSDFLT(out.v, in, 0)
  VERIFY_DBL(out, 0x42700000, 0x00005000, 0xC3DFFFFF, 0xFFFFFFFF);

  in = ((vector signed long long) { 0x1000000000000000LL, 0xFFFFFFFFFFFFFFFFLL});
  CSDFLT(out.v, in, 60)
  VERIFY_DBL(out, 0x3ff00000, 0x0, 0xBC300000, 0x0);

  in = ((vector signed long long) { 0x0000000000000002LL, 0x0000000000000000LL});
  CSDFLT(out.v, in, 0)
  VERIFY_DBL(out, 0x40000000, 0x0, 0x0, 0x0);

  DEBUG(1, "passed\n");
  return (0);
}


int test_cudflt()
{
  union {
    unsigned int i[4];
    vector double v;
  } out;
  vector signed long long in;

  DEBUG(1, "CUDFLT..........");

  in = ((vector signed long long) { 0x0000000000000001LL, 0x8000000000000000LL});
  CUDFLT(out.v, in, 0)
  VERIFY_DBL(out, 0x3ff00000, 0x0, 0x43e00000, 0x0);

  in = ((vector signed long long) { 0xFFFFFFFFFFFFFFFFLL, 0x0000000000000105LL});
  CUDFLT(out.v, in, 0)
  VERIFY_DBL(out, 0x43efffff, 0xffffffff, 0x40705000, 0x0);

  in = ((vector signed long long) { 0x0000010000000005LL, 0x8000000000000805LL});
  CUDFLT(out.v, in, 0)
  VERIFY_DBL(out, 0x42700000, 0x00005000, 0x43e00000, 0x1);

  in = ((vector signed long long) { 0x1000000000000000LL, 0xFFFFFFFFFFFFFFFFLL});
  CUDFLT(out.v, in, 60)
  VERIFY_DBL(out, 0x3ff00000, 0x0, 0x402fffff, 0xffffffff);

  in = ((vector signed long long) { 0x0000000000000002LL, 0x0000000000000000LL});
  CUDFLT(out.v, in, 0)
  VERIFY_DBL(out, 0x40000000, 0x0, 0x0, 0x0);

  DEBUG(1, "passed\n");
  return (0);
}


int test_fsmd()
{
  int i, j;
  union {
    unsigned int ui[4];
    vector unsigned long long v;
  } out;
  unsigned int a;
  unsigned int exp;

  DEBUG(1, "FSMD............");

  for (i=0; i<samples; i++) {
    a = rand() ^ (rand() << 16);

    FSMD(out.v, a)

    for (j=0; j<4; j++) {
      exp = (a & (2-(j>>1))) ? 0xFFFFFFFF : 0x0;
      if (exp != out.ui[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x\n", exp, out.ui[j], a);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_gbd()
{
  int i, j;
  union {
    unsigned int ui[4];
    vector unsigned long long v;
  } in_a;
  union {
    unsigned int ui[4];
    vector unsigned int v;
  } out;
  unsigned int exp;

  DEBUG(1, "GBD.............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned long long)rand_v();
    
    GBD(out.v, in_a.v)
    
    for (j=0; j<4; j++) {
      exp = 0;
      if (j == 0) {
	if (in_a.ui[1] & 1) exp |= 2;
	if (in_a.ui[3] & 1) exp |= 1;
      }
      if (exp != out.ui[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x\n", exp, out.ui[j]);
	DEBUG(2, "\t\ta=0x%x 0x%x 0x%x 0x%x\n", in_a.ui[0], in_a.ui[1], in_a.ui[2], in_a.ui[3]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}

int test_mpyd()
{
  union {
    unsigned int i[4];
    vector signed long long v;
  } out;
  vector signed int in_a, in_b;

  DEBUG(1, "MPYD............");

  in_a = ((vector signed int) { 0, 0, 1, 1});
  in_b = ((vector signed int) { 1, 0x12345678, 1, 0xFEDCBA98});
  MPYD(out.v, in_a, in_b)
  VERIFY_DBL(out, 0, 0, 0xFFFFFFFF, 0xFEDCBA98);

  in_a = ((vector signed int) { 1, -1, 0xAAAAAAAA, 0x12345678});
  in_b = ((vector signed int) { 1, 0x12345678, 1, 0xFEDCBA98});
  MPYD(out.v, in_a, in_b)
  VERIFY_DBL(out, 0xFFFFFFFF, 0xEDCBA988, 0xFFEB4992, 0x35068740);

  in_a = ((vector signed int) { 2, 0x80000000, 32, 0x80000000});
  in_b = ((vector signed int) { 1, 0x7FFFFFFF, 0, 0x80000000});
  MPYD(out.v, in_a, in_b)
  VERIFY_DBL(out, 0xC0000000, 0x80000000, 0x40000000, 0x0);

  in_a = ((vector signed int) { 2, 0x7FFFFFFF, 32, 0x80000000});
  in_b = ((vector signed int) { 1, 0x7FFFFFFF, 0, 0x7FFFFFFF});
  MPYD(out.v, in_a, in_b)
  VERIFY_DBL(out, 0x3FFFFFFF, 0x1, 0xC0000000, 0x80000000);

  DEBUG(1, "passed\n");
  return (0);
}


int test_rotd()
{
  int i, j;
  union {
    unsigned int i[4];
    vector unsigned long long v;
  } in_a, in_b, out;
  unsigned int exp;
  unsigned int cnt, x, y;

  DEBUG(1, "ROTD............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned long long)rand_v();
    in_b.v = (vector unsigned long long)rand_v();

    ROTD(out.v, in_a.v, in_b.v)
    
    for (j=0; j<4; j++) {
      cnt = in_b.i[j|1] & 0x3F;
      x = in_a.i[j];
      y = in_a.i[j^1];
      if (cnt == 0) {
	exp = x;
      } else if (cnt == 32) {
	exp = y;
      } else if (cnt < 32) {
	exp = (x << cnt) | (y >> (32-cnt));
      } else {
	/* cnt > 32 */
	cnt -= 32;
	exp = (y << cnt) | (x >> (32-cnt));
      }
      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x%x\n", exp, out.i[j], in_a.i[j], in_b.i[j|1]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_rotdi()
{
  int i, j;
  union {
    unsigned int i[4];
    vector unsigned long long v;
  } in_a, out;
  unsigned int exp;
  unsigned int cnt, x, y;

  DEBUG(1, "ROTDI...........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned long long)rand_v();
    
    ROTDI(out.v, in_a.v, 113)
    
    for (j=0; j<4; j++) {
      cnt = 113 & 0x3F;
      x = in_a.i[j];
      y = in_a.i[j^1];
      exp = (y << 17) | (x >> 15);
      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=113\n", exp, out.i[j], in_a.i[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_rotdm()
{
  int i, j;
  union {
    unsigned int i[4];
    vector unsigned long long v;
  } in_a, out;
  union {
    unsigned int i[4];
    vector signed long long v;
  } in_b;
  unsigned int exp;
  unsigned int cnt;

  DEBUG(1, "ROTDM...........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned long long)rand_v();
    in_b.v = (vector signed   long long)rand_v();


    ROTDM(out.v, in_a.v, in_b.v)

    for (j=0; j<4; j++) {
      cnt = in_b.i[j|1] & 0x7F;
      if (cnt & 0x40) cnt |= 0xFFFFFF80;
      cnt = -cnt;
      if (j & 1) {
	if (cnt > 63) {
	  exp = 0;
	} else if (cnt > 31) {
	  exp = in_a.i[j-1] >> (cnt - 32);
	} else if (cnt == 32) {
	  exp = in_a.i[j-1];
	} else if (cnt == 0) {
	  exp = in_a.i[j];
	} else {
	  exp = (in_a.i[j] >> cnt) | (in_a.i[j-1] << (32 - cnt));
	}
      } else {
	if (cnt > 31) {
	  exp = 0;
	} else {
	  exp = in_a.i[j] >> cnt;
	}
      }
      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x%x\n", exp, out.i[j], in_a.i[j], in_b.i[j|1]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_rotdmi()
{
  int i, j;
  union {
    unsigned int i[4];
    vector unsigned long long v;
  } in_a, out;
  unsigned int exp;
  unsigned int cnt;

  DEBUG(1, "ROTDMI..........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned long long)rand_v();

    ROTDMI(out.v, in_a.v, -23)

    for (j=0; j<4; j++) {
      cnt = 23;
      if (j & 1) {
	exp = (in_a.i[j] >> cnt) | (in_a.i[j-1] << (32 - cnt));
      } else {
	exp = in_a.i[j] >> cnt;
      }
      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=23\n", exp, out.i[j], in_a.i[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_rotmad()
{
  int i, j;
  union {
    signed int i[4];
    vector signed long long v;
  } in_a, in_b, out;
  signed int exp;
  unsigned int cnt;

  DEBUG(1, "ROTMAD..........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector signed long long)rand_v();
    in_b.v = (vector signed long long)rand_v();

    ROTMAD(out.v, in_a.v, in_b.v)

    for (j=0; j<4; j++) {
      cnt = in_b.i[j|1] & 0x7F;
      if (cnt & 0x40) cnt |= 0xFFFFFF80;
      cnt = -cnt;
      if (j & 1) {
	if (cnt > 63) {
	  exp = in_a.i[j-1] >> 31;
	} else if (cnt > 31) {
	  exp = in_a.i[j-1] >> (cnt - 32);
	} else if (cnt == 32) {
	  exp = in_a.i[j-1];
	} else if (cnt == 0) {
	  exp = in_a.i[j];
	} else {
	  exp = ((unsigned)in_a.i[j] >> cnt) | (in_a.i[j-1] << (32 - cnt));
	}
      } else {
	if (cnt > 31) cnt = 31;
	exp = in_a.i[j] >> cnt;
      }
      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x%x\n", exp, out.i[j], in_a.i[j], in_b.i[j|1]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}

int test_rotmadi()
{
  int i, j;
  union {
    signed int i[4];
    vector signed long long v;
  } in_a, out;
  signed int exp;
  unsigned int cnt;

  DEBUG(1, "ROTMADI.........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector signed long long)rand_v();

    ROTMADI(out.v, in_a.v, -41)

    for (j=0; j<4; j++) {
      cnt = 41;
      if (j & 1) {
	if (cnt > 63) {
	  exp = in_a.i[j-1] >> 31;
	} else if (cnt > 31) {
	  exp = in_a.i[j-1] >> (cnt - 32);
	} else if (cnt == 32) {
	  exp = in_a.i[j-1];
	} else if (cnt == 0) {
	  exp = in_a.i[j];
	} else {
	  exp = ((unsigned)in_a.i[j] >> cnt) | (in_a.i[j-1] << (32 - cnt));
	}
      } else {
	if (cnt > 31) cnt = 31;
	exp = in_a.i[j] >> cnt;
      }
      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=-41\n", exp, out.i[j], in_a.i[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_sfd()
{
  int i, j;
  union {
    signed int i[4];
    unsigned int ui[4];
    vector signed long long v;
  } in_a, in_b, out;
  int exp;

  DEBUG(1, "SFD.............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector signed long long)rand_v();
    in_b.v = (vector signed long long)rand_v();
    
    SFD(out.v, in_a.v, in_b.v)
    
    for (j=0; j<4; j++) {
      exp = in_b.i[j] - in_a.i[j];
      if ((j & 1) == 0) {
	if (in_a.ui[j|1] > in_b.ui[j|1]) exp -= 1;
      }

      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x%x\n", exp, out.i[j], in_a.i[j], in_b.i[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


int test_sfdi()
{
#ifdef __xlc__
  int i, j;
  union {
    signed int i[4];
    unsigned int ui[4];
    vector signed long long v;
  } in_a, out;
  unsigned int exp;

  DEBUG(1, "SFDI............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector signed long long)rand_v();
    
    SFDI(out.v, in_a.v, -1750)
    
    for (j=0; j<4; j++) {
      if (j & 1) {
	exp = -1750 - in_a.i[j];
      } else {
	exp = ~in_a.i[j];
	if (in_a.ui[j|1] > (unsigned int)(-1750)) exp -= 1;
      }

      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=-1750\n", exp, out.i[j], in_a.i[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
#endif /* __xlc__ */
  return (0);
}


int test_shld()
{
  int i, j;
  union {
    unsigned int i[4];
    vector unsigned long long v;
  } in_a, in_b, out;
  unsigned int exp;
  unsigned int cnt;

  DEBUG(1, "SHLD............");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned long long)rand_v();
    in_b.v = (vector unsigned long long)rand_v();
    
    SHLD(out.v, in_a.v, in_b.v)
    
    for (j=0; j<4; j++) {
      cnt = in_b.i[j|1] & 0x7F;
      if (j & 1) {
	if (cnt > 31) {
	  exp = 0;
	} else {
	  exp = in_a.i[j] << cnt;
	}
      } else {
	if (cnt > 63) {
	  exp = 0;
	} else if (cnt > 31) {
	  exp = in_a.i[j|1] << (cnt - 32);
	} else if (cnt == 32) {
	  exp = in_a.i[j|1];
	} else if (cnt == 0) {
	  exp = in_a.i[j];
	} else {
	  exp = (in_a.i[j] << cnt) | (in_a.i[j|1] >> (32 - cnt));
	}
      }
      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x b=0x%x\n", exp, out.i[j], in_a.i[j], in_b.i[j|1]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}

int test_shldi()
{
  int i, j;
  union {
    unsigned int i[4];
    vector unsigned long long v;
  } in_a, out;
  unsigned int exp;

  DEBUG(1, "SHLDI...........");

  for (i=0; i<samples; i++) {
    in_a.v = (vector unsigned long long)rand_v();
    
    SHLDI(out.v, in_a.v, 17)
    
    for (j=0; j<4; j++) {
      if (j & 1) {
	  exp = in_a.i[j] << 17;
      } else {
	exp = (in_a.i[j] << 17) | (in_a.i[j|1] >> (32 - 17));
      }
      if (exp != out.i[j]) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\ti=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x a=0x%x i7=17\n", exp, out.i[j], in_a.i[j]);
	return (1);
	break;
      }
    }
  }
  DEBUG(1, "passed\n");
  return (0);
}


unsigned int samples_double[] = 
{
  0x00000000, 0x00000000,	/* 0  positive Zero */
  0x80000000, 0x00000000,	/* 1  negative Zero */
  0x7FF00000, 0x00000000,	/* 2  positive Infinity */
  0xFFF00000, 0x00000000,	/* 3  negative Infinity */
  0x7FF00000, 0x00000001,	/* 4  +NaN 01 */
  0xFFF00000, 0x00000001,	/* 5  -NaN 01 */
  0x7FF00300, 0x00000000,	/* 6  +NaN 30 */
  0xFFF00300, 0x00000000,	/* 7  -NaN 30 */
  0x7FF00300, 0x00000001,	/* 8  +NaN 31 */
  0xFFF00300, 0x00000001,	/* 9  -NaN 31 */
  0x3FF00000, 0x00000000,	/* 10 +1.0 */
  0xBFF00000, 0x00000000,	/* 11 -1.0 */
  0x40000000, 0x00000000,	/* 12 +2.0 */
  0xC0000000, 0x00000000,	/* 13 -2.0 */
  0x3FF00000, 0x00000001,	/* 14 +1.0 + e */
  0xBFF00000, 0x00000001,	/* 15 -1.0 + e */
  0x40000000, 0x00000001,	/* 16 +2.0 + e */
  0xC0000000, 0x00000001,	/* 17 -2.0 + e */
};

#define DOUBLE_SAMPLES	(int)(sizeof(samples_double)/8)



int test_dfceq()
{
  int i, j;
  int index;
  union {
    unsigned int i[4];
    vector double d;
  } in_a, in_b;
  unsigned int ah, al, bh, bl;
  union {
    unsigned int i[4];
    vector unsigned long long ll;
  } out;
  unsigned int exp;

  DEBUG(1, "DFCEQ...........");

  /* Test DOUBLE_IEEE */
#undef _SPU_EXT_H_
#include "spu_ext.h"
  for (i=0; i<DOUBLE_SAMPLES; i++) {
    index = (unsigned int)(rand())&1;
    in_a.i[2*index+0] = ah = samples_double[2*i+0];
    in_a.i[2*index+1] = al = samples_double[2*i+1];
    for (j=0; j<DOUBLE_SAMPLES; j++) {
      in_b.i[2*index+0] = bh = samples_double[2*j+0];
      in_b.i[2*index+1] = bl = samples_double[2*j+1];
      
      DFCEQ(out.ll, in_a.d, in_b.d)
      /* Compute expected */
      exp = 0;
      if (((ah & 0x7FFFFFFF) == 0) && (al == 0) &&
	  ((bh & 0x7FFFFFFF) == 0) && (bl == 0)) {
	/* zero compare equal */
	exp = 0xFFFFFFFF;
      }
      if ((ah == bh) && (al == bl)) {
	/* bits are equal */
	exp = 0xFFFFFFFF;
      }
      if (((ah & 0x7FF00000) == 0x7FF00000) &&
	  (((ah & 0x000FFFFF) != 0) || (al != 0))) {
	/* a is a nan */
	exp = 0;
      }
      if (((bh & 0x7FF00000) == 0x7FF00000) &&
	  (((bh & 0x000FFFFF) != 0) || (bl != 0))) {
	/* b is a nan */
	exp = 0;
      }
      if ((out.i[2*index+0] != exp) || (out.i[2*index+1] != exp)) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\tIEEE i=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x 0x%x\n", exp, out.i[2*index+0], out.i[2*index+1]);
	return (1);
	break;
      }
    }
  }

#undef _SPU_EXT_H_
#define DOUBLE_X86
#include "spu_ext.h"
#undef  DOUBLE_X86
  for (i=0; i<DOUBLE_SAMPLES; i++) {
    index = (unsigned int)(rand())&1;
    in_a.i[2*index+0] = ah = samples_double[2*i+0];
    in_a.i[2*index+1] = al = samples_double[2*i+1];
    for (j=0; j<DOUBLE_SAMPLES; j++) {
      in_b.i[2*index+0] = bh = samples_double[2*j+0];
      in_b.i[2*index+1] = bl = samples_double[2*j+1];
      
      DFCEQ(out.ll, in_a.d, in_b.d)
      /* Compute expected */
      exp = 0;
      if ((ah == bh) && (al == bl)) {
	/* bits are equal */
	exp = 0xFFFFFFFF;
      }
      if ((out.i[2*index+0] != exp) || (out.i[2*index+1] != exp)) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\tX86 i=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x 0x%x\n", exp, out.i[2*index+0], out.i[2*index+1]);
	return (1);
	break;
      }
    }
  }
#undef _SPU_EXT_H_
#define DOUBLE_ZERO
#include "spu_ext.h"
#undef  DOUBLE_ZERO
  for (i=0; i<DOUBLE_SAMPLES; i++) {
    index = (unsigned int)(rand())&1;
    in_a.i[2*index+0] = ah = samples_double[2*i+0];
    in_a.i[2*index+1] = al = samples_double[2*i+1];
    for (j=0; j<DOUBLE_SAMPLES; j++) {
      in_b.i[2*index+0] = bh = samples_double[2*j+0];
      in_b.i[2*index+1] = bl = samples_double[2*j+1];
      
      DFCEQ(out.ll, in_a.d, in_b.d)
      /* Compute expected */
      exp = 0;
      if (((ah & 0x7FFFFFFF) == 0) && (al == 0) &&
	  ((bh & 0x7FFFFFFF) == 0) && (bl == 0)) {
	/* zero compare equal */
	exp = 0xFFFFFFFF;
      }
      if ((ah == bh) && (al == bl)) {
	/* bits compare equal */
	exp = 0xFFFFFFFF;
      }
      if ((out.i[2*index+0] != exp) || (out.i[2*index+1] != exp)) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\tZERO i=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x 0x%x\n", exp, out.i[2*index+0], out.i[2*index+1]);
	return (1);
	break;
      }
    }
  }

  DEBUG(1, "passed\n");
  return (0);
}

int test_dfcgt()
{
  int i, j;
  int index;
  union {
    unsigned int i[4];
    vector double d;
  } in_a, in_b;
  unsigned int ah, al, bh, bl, abs_ah, abs_bh;
  union {
    unsigned int i[4];
    vector unsigned long long ll;
  } out;
  unsigned int exp;

  DEBUG(1, "DFCGT...........");

  /* Test DOUBLE_IEEE */
#undef _SPU_EXT_H_
#include "spu_ext.h"
  for (i=0; i<DOUBLE_SAMPLES; i++) {
    index = (unsigned int)(rand())&1;
    in_a.i[2*index+0] = ah = samples_double[2*i+0];
    in_a.i[2*index+1] = al = samples_double[2*i+1];
    for (j=0; j<DOUBLE_SAMPLES; j++) {
      in_b.i[2*index+0] = bh = samples_double[2*j+0];
      in_b.i[2*index+1] = bl = samples_double[2*j+1];
      
      DFCGT(out.ll, in_a.d, in_b.d)
      /* Compute expected */
      exp = 0;
      abs_ah = ah & 0x7FFFFFFF;
      abs_bh = bh & 0x7FFFFFFF;

      if ((abs_ah > abs_bh) || ((abs_ah == abs_bh) && (al > bl))) {
	/* abs(a) > abs(b) */
	exp = 0xFFFFFFFF;
      }
      if (abs_ah != ah) {
	if (abs_bh != bh) {
	  /* both negative, toggle expected */
	  exp ^= 0xFFFFFFFF;
	} else {
	  /* negative a, positive b */
	  exp = 0x0;
	}
      } else { 
	if (abs_bh != bh) {
	  /* positive a, negative b */
	  exp = 0xFFFFFFFF;
	}
      }
      if (((ah & 0x7FFFFFFF) == 0) && (al == 0) &&
	  ((bh & 0x7FFFFFFF) == 0) && (bl == 0)) {
	/* zero compare equal */
	exp = 0;
      }
      if ((ah == bh) && (al == bl)) {
	/* compare equal */
	exp = 0;
      }
      if (((ah & 0x7FF00000) == 0x7FF00000) &&
	  (((ah & 0x000FFFFF) != 0) || (al != 0))) {
	/* a is a NaN */
	exp = 0;
      }
      if (((bh & 0x7FF00000) == 0x7FF00000) &&
	  (((bh & 0x000FFFFF) != 0) || (bl != 0))) {
	/* b is a NaN */
	exp = 0;
      }
      if ((out.i[2*index+0] != exp) || (out.i[2*index+1] != exp)) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\tIEEE i=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x 0x%x\n", exp, out.i[2*index+0], out.i[2*index+1]);
	return (1);
	break;
      }
    }
  }

#undef _SPU_EXT_H_
#define DOUBLE_X86
#include "spu_ext.h"
#undef  DOUBLE_X86
  for (i=0; i<DOUBLE_SAMPLES; i++) {
    index = (unsigned int)(rand())&1;
    in_a.i[2*index+0] = ah = samples_double[2*i+0];
    in_a.i[2*index+1] = al = samples_double[2*i+1];
    for (j=0; j<DOUBLE_SAMPLES; j++) {
      in_b.i[2*index+0] = bh = samples_double[2*j+0];
      in_b.i[2*index+1] = bl = samples_double[2*j+1];
      
      DFCGT(out.ll, in_a.d, in_b.d)
      /* Compute expected */
      exp = 0;
      abs_ah = ah & 0x7FFFFFFF;
      abs_bh = bh & 0x7FFFFFFF;

      if ((abs_ah > abs_bh) || ((abs_ah == abs_bh) && (al > bl))) {
	/* abs(a) > abs(b) */
	exp = 0xFFFFFFFF;
      }
      if (abs_ah != ah) {
	if (abs_bh != bh) {
	  /* both negative, toggle expected */
	  exp ^= 0xFFFFFFFF;
	} else {
	  /* negative a, positive b */
	  exp = 0x0;
	}
      } else { 
	if (abs_bh != bh) {
	  /* positive a, negative b */
	  exp = 0xFFFFFFFF;
	}
      }
      if (((ah & 0x7FFFFFFF) == 0) && (al == 0) &&
	  ((bh & 0x7FFFFFFF) == 0) && (bl == 0)) {
	/* zero compare equal */
	exp = 0;
      }
      if ((ah == bh) && (al == bl)) {
	/* compare equal */
	exp = 0;
      }
      if ((out.i[2*index+0] != exp) || (out.i[2*index+1] != exp)) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\tX86 i=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x 0x%x\n", exp, out.i[2*index+0], out.i[2*index+1]);
	return (1);
	break;
      }
    }
  }

  DEBUG(1, "passed\n");
  return (0);
}

int test_dfcmeq()
{
  int i, j;
  int index;
  union {
    unsigned int i[4];
    vector double d;
  } in_a, in_b;
  unsigned int ah, al, bh, bl;
  union {
    unsigned int i[4];
    vector unsigned long long ll;
  } out;
  unsigned int exp;

  DEBUG(1, "DFCMEQ..........");

  /* Test DOUBLE_IEEE */
#undef _SPU_EXT_H_
#include "spu_ext.h"
  for (i=0; i<DOUBLE_SAMPLES; i++) {
    index = (unsigned int)(rand())&1;
    in_a.i[2*index+0] = ah = samples_double[2*i+0];
    in_a.i[2*index+1] = al = samples_double[2*i+1];
    for (j=0; j<DOUBLE_SAMPLES; j++) {
      in_b.i[2*index+0] = bh = samples_double[2*j+0];
      in_b.i[2*index+1] = bl = samples_double[2*j+1];
      
      DFCMEQ(out.ll, in_a.d, in_b.d)
      /* Compute expected */
      exp = 0;
      if (((ah & 0x7FFFFFFF) == (bh & 0x7FFFFFFF)) && (al == bl)) {
	/* abs(a) == abs(b) */
	exp = 0xFFFFFFFF;
      }
      if (((ah & 0x7FF00000) == 0x7FF00000) &&
	  (((ah & 0x000FFFFF) != 0) || (al != 0))) {
	/* a is a NaN */
	exp = 0;
      }
      if (((bh & 0x7FF00000) == 0x7FF00000) &&
	  (((bh & 0x000FFFFF) != 0) || (bl != 0))) {
	/* b is a NaN */
	exp = 0;
      }
      if ((out.i[2*index+0] != exp) || (out.i[2*index+1] != exp)) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\tIEEE i=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x 0x%x\n", exp, out.i[2*index+0], out.i[2*index+1]);
	return (1);
	break;
      }
    }
  }

#undef _SPU_EXT_H_
#define DOUBLE_X86
#include "spu_ext.h"
#undef  DOUBLE_X86
  for (i=0; i<DOUBLE_SAMPLES; i++) {
    index = (unsigned int)(rand())&1;
    in_a.i[2*index+0] = ah = samples_double[2*i+0];
    in_a.i[2*index+1] = al = samples_double[2*i+1];
    for (j=0; j<DOUBLE_SAMPLES; j++) {
      in_b.i[2*index+0] = bh = samples_double[2*j+0];
      in_b.i[2*index+1] = bl = samples_double[2*j+1];
      
      DFCMEQ(out.ll, in_a.d, in_b.d)
      /* Compute expected */
      exp = 0;
      if (((ah & 0x7FFFFFFF) == (bh & 0x7FFFFFFF)) && (al == bl)) {
	/* abs(a) == abs(b) */
	exp = 0xFFFFFFFF;
      }
      if ((out.i[2*index+0] != exp) || (out.i[2*index+1] != exp)) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\tX86 i=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x 0x%x\n", exp, out.i[2*index+0], out.i[2*index+1]);
	return (1);
	break;
      }
    }
  }

  DEBUG(1, "passed\n");
  return (0);
}

int test_dfcmgt()
{
  int i, j;
  int index;
  union {
    unsigned int i[4];
    vector double d;
  } in_a, in_b;
  unsigned int ah, al, bh, bl, abs_ah, abs_bh;
  union {
    unsigned int i[4];
    vector unsigned long long ll;
  } out;
  unsigned int exp;

  DEBUG(1, "DFCMGT..........");

  /* Test DOUBLE_IEEE */
#undef _SPU_EXT_H_
#include "spu_ext.h"
  for (i=0; i<DOUBLE_SAMPLES; i++) {
    index = (unsigned int)(rand())&1;
    in_a.i[2*index+0] = ah = samples_double[2*i+0];
    in_a.i[2*index+1] = al = samples_double[2*i+1];
    for (j=0; j<DOUBLE_SAMPLES; j++) {
      in_b.i[2*index+0] = bh = samples_double[2*j+0];
      in_b.i[2*index+1] = bl = samples_double[2*j+1];
      
      DFCMGT(out.ll, in_a.d, in_b.d)
      /* Compute expected */
      exp = 0;
      abs_ah = ah & 0x7FFFFFFF;
      abs_bh = bh & 0x7FFFFFFF;

      if ((abs_ah > abs_bh) || ((abs_ah == abs_bh) && (al > bl))) {
	/* abs(a) > abs(b) */
	exp = 0xFFFFFFFF;
      }
      if (((ah & 0x7FF00000) == 0x7FF00000) &&
	  (((ah & 0x000FFFFF) != 0) || (al != 0))) {
	/* a is a NaN */
	exp = 0;
      }
      if (((bh & 0x7FF00000) == 0x7FF00000) &&
	  (((bh & 0x000FFFFF) != 0) || (bl != 0))) {
	/* b is a NaN */
	exp = 0;
      }
      if ((out.i[2*index+0] != exp) || (out.i[2*index+1] != exp)) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\tIEEE i=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x 0x%x\n", exp, out.i[2*index+0], out.i[2*index+1]);
	return (1);
	break;
      }
    }
  }

#undef _SPU_EXT_H_
#define DOUBLE_X86
#include "spu_ext.h"
#undef  DOUBLE_X86
  for (i=0; i<DOUBLE_SAMPLES; i++) {
    index = (unsigned int)(rand())&1;
    for (j=0; j<DOUBLE_SAMPLES; j++) {
      in_a.i[2*index+0] = ah = samples_double[2*i+0];
      in_a.i[2*index+1] = al = samples_double[2*i+1];
      in_b.i[2*index+0] = bh = samples_double[2*j+0];
      in_b.i[2*index+1] = bl = samples_double[2*j+1];
      
      DFCMGT(out.ll, in_a.d, in_b.d)
      /* Compute expected */
      exp = 0;
      abs_ah = ah & 0x7FFFFFFF;
      abs_bh = bh & 0x7FFFFFFF;

      if ((abs_ah > abs_bh) || ((abs_ah == abs_bh) && (al > bl))) {
	/* abs(a) > abs(b) */
	exp = 0xFFFFFFFF;
      }
      if ((out.i[2*index+0] != exp) || (out.i[2*index+1] != exp)) {
	DEBUG(1, "FAILED\n");
	DEBUG(2, "\t\tX86 i=%d j=%d\n", i, j);
	DEBUG(2, "\t\texp=0x%x got=0x%x 0x%x\n", exp, out.i[2*index+0], out.i[2*index+1]);
	return (1);
	break;
      }
    }
  }

  DEBUG(1, "passed\n");
  return (0);
}


int test_byte()
{
  int failures = 0;

  failures += test_ab();
  failures += test_abi();
  failures += test_mpyah();
  failures += test_rotb();
  failures += test_rotbi();
  failures += test_rotbm();
  failures += test_rotbmi();
  failures += test_rotmab();
  failures += test_rotmabi();
  failures += test_sfb();
  failures += test_sfbi();
  failures += test_shlb();
  failures += test_shlbi();

  return (failures);
}

int test_doubleword()
{
  int failures = 0;

  failures += test_ad();
  failures += test_adi();
  failures += test_cdflts();
  failures += test_cdfltu();
  failures += test_ceqd();
  failures += test_ceqdi();
  failures += test_cgtd();
  failures += test_cgtdi();
  failures += test_clgtd();
  failures += test_clgtdi();
  failures += test_csdflt();
  failures += test_cudflt();
  failures += test_fsmd();
  failures += test_mpyd();
  failures += test_gbd();
  failures += test_rotd();
  failures += test_rotdi();
  failures += test_rotdm();
  failures += test_rotdmi();
  failures += test_rotmad();
  failures += test_rotmadi();
  failures += test_sfd();
  failures += test_sfdi();
  failures += test_shld();
  failures += test_shldi();

  return (failures);
}

int test_double()
{
  int failures = 0;

  failures += test_dfceq();
  failures += test_dfcgt();
  failures += test_dfcmeq();
  failures += test_dfcmgt();

  return (failures);
}



DECLARE_MAIN()
{
  DECLARE_TEST_VARS();

  SCALE_SAMPLES(samples);

  INVOKE_TEST(test_byte);
  INVOKE_TEST(test_doubleword);
  INVOKE_TEST(test_double);

  DISPLAY_RESULTS();
}

