/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2007,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#include <stdio.h>
#include <vec_types.h>
#ifdef __SPU__
#include <vmx2spu.h>
#else /* not __SPU__ */
#include <altivec.h>
#endif
#include "libtest.h"

typedef union {
  vec_uchar16	vuc;
  vec_char16	vc;
  vec_bchar16	vbc;
  vec_ushort8	vus;
  vec_bshort8	vbs;
  vec_short8	vs;
  vec_pixel8	vp;
  vec_uint4	vui;
  vec_int4	vi;
  vec_bint4	vbi;
  vec_float4	vf;
  unsigned char	uc[16];
  unsigned char	bc[16];
  signed char	c[16];
  unsigned short us[8];
  signed short	s[8];
  unsigned short bs[8];
  signed short	p[8];
  unsigned int	ui[4];
  unsigned int	bi[4];
  signed int	i[4];
  float		f[4];
} v128;
  
static char *passed = "passed\n";
static char *failed = "FAILED\n";
static v128 a, b, c, d, e;
static unsigned char data[256];


static int samples = 200;

unsigned int seed = 0;

unsigned int rand_local()
{
  seed = seed * 69607 + 54329;
  return (seed);
}

float rand_float(float min, float max)
{
  union {
    unsigned int ui;
    float f;
  } val;

  val.ui = (rand_local() & 0x7FFFFF) | 0x3F800000;
  val.f -= 1.0f;
  val.f *= (max - min);
  val.f += min;

  return (val.f);
}

vec_uint4 rand_vui()
{
  static v128 val;

  val.ui[0] = rand_local();
  val.ui[1] = rand_local();
  val.ui[2] = rand_local();
  val.ui[3] = rand_local();

  return (val.vui);
}

vec_float4 rand_vf(float min, float max)
{
  int i;
  static v128 val;

  for (i=0; i<4; i++) {
    val.ui[i] = (rand_local() & 0x7FFFFF) | 0x3F800000;
    val.f[i] -= 1.0f;
    val.f[i] *= (max - min);
    val.f[i] += min;
  }
  return (val.vf);
}


signed int sum3s(signed int a, signed int b, signed int c)
{
  signed int a1, a2, b1, b2, c1, c2, s1, s2, s;

  a1 = a >> 16;
  a2 = a & 0xFFFF;
  b1 = b >> 16;
  b2 = b & 0xFFFF;
  c1 = c >> 16;
  c2 = c & 0xFFFF;

  s2 = a2 + b2 + c2;
  s1 = a1 + b1 + c1 + (s2 >> 16);
  
  if ((s1 < 0) && (s1 < -32768)) {
    /* underflow */
    s = 0x80000000;
  } else if ((s1 > 0) && (s1 > 32767)) {
    /* overflow */
    s = 0x7FFFFFFF;
  } else {
    s = (s1 << 16) | (s2 & 0xFFFF);
  }
  return (s);
}


signed int sum5s(signed int a, signed int b, signed int c, signed int d, signed int e)
{
  signed int a1, a2, b1, b2, c1, c2, d1, d2, e1, e2, s1, s2, s;

  a1 = a >> 16;
  a2 = a & 0xFFFF;
  b1 = b >> 16;
  b2 = b & 0xFFFF;
  c1 = c >> 16;
  c2 = c & 0xFFFF;
  d1 = d >> 16;
  d2 = d & 0xFFFF;
  e1 = e >> 16;
  e2 = e & 0xFFFF;

  s2 = a2 + b2 + c2 + d2 + e2;
  s1 = a1 + b1 + c1 + d1 + e1 + (s2 >> 16);
  
  if ((s1 < 0) && (s1 < -32768)) {
    /* underflow */
    s = 0x80000000;
  } else if ((s1 > 0) && (s1 > 32767)) {
    /* overflow */
    s = 0x7FFFFFFF;
  } else {
    s = (s1 << 16) | (s2 & 0xFFFF);
  }
  return (s);
}


#define VERIFY_EQUAL(_a, _b)	if (verify_equal(__LINE__, _a, _b)) return (1);
#define VERIFY_DATA(_a, _index)	if (verify_data(__LINE__, _a, _index)) return (1);
#define VERIFY_DATA_LEFT(_a, _index)	if (verify_data_left(__LINE__, _a, _index)) return (1);
#define VERIFY_DATA_RIGHT(_a, _index)	if (verify_data_right(__LINE__, _a, _index)) return (1);
#define VERIFY_DATA_ELEMENT(_a, _index, _ele, _size)	\
    if (verify_data_element(__LINE__, _a, _index, _ele, _size)) return (1);

int verify_equal(int line, v128 d, v128 e)
{
  if ((d.ui[0] != e.ui[0]) ||
      (d.ui[1] != e.ui[1]) ||
      (d.ui[2] != e.ui[2]) ||
      (d.ui[3] != e.ui[3])) {
    DEBUG(1, failed);
    DEBUG(2, "\tequivalence not verified at line %d\n", line);
    DEBUG(2, "\td=0x%x 0x%x 0x%x 0x%x\n", d.ui[0], d.ui[1], d.ui[2], d.ui[3]);
    DEBUG(2, "\te=0x%x 0x%x 0x%x 0x%x\n", e.ui[0], e.ui[1], e.ui[2], e.ui[3]);
    return (1);
  }
  return (0);
}      

int verify_data(int line, v128 a, int index)
{
  int i;
  for (i=0; i<16; i++) {
    if (data[index + i] != a.uc[i]) {
      DEBUG(1, failed);
      DEBUG(2, "\tdata buffer not verified at line %d\n", line);
      DEBUG(2, "\ti=%d a=0x%x data=0x%x\n", i, a.uc[i], data[index+i]);
      return (1);
    }
  }
  return (0);
}      

int verify_data_left(int line, v128 a, int index)
{
  int i;
  unsigned char exp;
  unsigned int eb;

  eb = (unsigned int)index & 0xF;

  for (i=0; i<(int)(sizeof(data)); i++) {
    if ((i >= (int)index) && (i < (int)(index + 16-eb))) {
      exp = a.uc[i-index];
    } else {
      exp = i;
    }
    if (data[i] != exp) {
      DEBUG(1, failed);
      DEBUG(2, "\tdata buffer not verified at line %d\n", line);
      DEBUG(2, "\ti=%d exp=0x%x got=0x%x\n", i, exp, data[i]);
      return (1);
    }
    data[i] = i;
  }
  return (0);
}      


int verify_data_right(int line, v128 a, int index)
{
  int i;
  unsigned char exp;
  unsigned int eb;

  eb = (unsigned int)index & 0xF;

  for (i=0; i<(int)(sizeof(data)); i++) {
    if ((i >= (int)index-(int)eb) && (i < (int)(index))) {
      exp = a.uc[16+i-index];
    } else {
      exp = i;
    }
    if (data[i] != exp) {
      DEBUG(1, failed);
      DEBUG(2, "\tdata buffer not verified at line %d\n", line);
      DEBUG(2, "\ti=%d exp=0x%x got=0x%x\n", i, exp, data[i]);
      return (1);
    }
    data[i] = i;
  }
  return (0);
}      


int verify_data_element(int line, v128 a, int index, int element, int size)
{
  int i;

  index += element * size;
  index &= ~(size-1);

  for (i=0; i<256; i++) {
    if ((i >= index) && (i < index+size)) {
      if (data[i] != a.uc[i & 0xF]) {
	DEBUG(1, failed);
	DEBUG(2, "\tdata buffer not verified at line %d\n", line);
	DEBUG(2, "\ti=%d a=0x%x data=0x%x\n", i, a.uc[i], data[index+i]);
	return (1);
      }
      data[i] = 0;
    } else {
      if (data[i] != 0x0) {
	DEBUG(1, failed);
	DEBUG(2, "\tdata buffer at %d contains non-zero data\n", line);
	return (1);
      }
    }
  }
  return (0);
}      



int test_vec_abs()
{
  int i, j;

  DEBUG(1, "abs......");

  /* signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    d.vc = vec_abs(a.vc);
    for (j=0; j<16; j++) {
      signed char exp;
      exp = (a.c[j] < 0) ? -a.c[j] : a.c[j];
      if (exp != d.c[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar: element=%d in=%d out=%d exp=%d\n", 
	       j, a.c[j], d.c[j], exp);
	return (1);
      }
    }
  }

  /* signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    d.vs = vec_abs(a.vs);
    for (j=0; j<8; j++) {
      signed short exp;
      exp = (a.s[j] < 0) ? -a.s[j] : a.s[j];
      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d in=%d out=%d exp=%d\n", 
	       j, a.s[j], d.s[j], exp);
	return (1);
      }
    }
  }

  /* signed int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    d.vi = vec_abs(a.vi);
    for (j=0; j<4; j++) {
      signed int exp;
      exp = (a.i[j] < 0) ? -a.i[j] : a.i[j];
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint: element=%d in=%d out=%d exp=%d\n", 
	       j, a.i[j], d.i[j], exp);
	return (1);
      }
    }
  }

  /* float */
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-1.0f, 1.0f);
    d.vf = vec_abs(a.vf);
    for (j=0; j<4; j++) {
      float exp;
      exp = (a.f[j] < 0.0f) ? -a.f[j] : a.f[j];
      if (exp != d.f[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tfloat: element=%d in=%f out=%f exp=%f\n", 
	       j, a.f[j], d.f[j], exp);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_abss()
{
  int i, j;

  DEBUG(1, "abss.....");

  /* signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    if (a.c[i & 15] < 0) a.c[i & 15] = 0x80;
    d.vc = vec_abss(a.vc);
    for (j=0; j<16; j++) {
      signed char exp;
      exp = (a.c[j] < 0) ? -a.c[j] : a.c[j];
      if (exp < 0) exp = 0x7F;
      if (exp != d.c[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar: element=%d in=%d out=%d exp=%d\n", 
	       j, a.c[j], d.c[j], exp);
	return (1);
      }
    }
  }

  /* signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    if (a.s[i & 7] < 0) a.s[i & 7] = 0x8000;
    d.vs = vec_abss(a.vs);
    for (j=0; j<8; j++) {
      signed short exp;
      exp = (a.s[j] < 0) ? -a.s[j] : a.s[j];
      if (exp < 0) exp = 0x7FFF;
      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d in=%d out=%d exp=%d\n", 
	       j, a.s[j], d.s[j], exp);
	return (1);
      }
    }
  }

  /* signed int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    if (a.i[i & 3] < 0) a.i[i & 3] = 0x80000000;
    d.vi = vec_abss(a.vi);
    for (j=0; j<4; j++) {
      signed int exp;
      if (a.i[j] == -0x7FFFFFFF - 1) 
	exp = 0x7FFFFFFF;
      else
	exp = (a.i[j] < 0) ? -a.i[j] : a.i[j];
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint: element=%d in=%d out=%d exp=%d\n", 
	       j, a.i[j], d.i[j], exp);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_add()
{
  int i, j;

  DEBUG(1, "add......");

  /* unsigned char  unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_add(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = a.uc[j] + b.uc[j];
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar + uchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.uc[j], b.uc[j], d.uc[j], exp);
	return (1);
      }
    }
    
    /* verify equivalent variations */
    e.vuc = vec_add(a.vuc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_add(a.vbc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_add(a.vc, b.vc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_add(a.vc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_add(a.vbc, b.vc);
    VERIFY_EQUAL(d, e);
  }

  /* unsigned short  unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_add(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = a.us[j] + b.us[j];
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort + ushort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.us[j], b.us[j], d.us[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vus = vec_add(a.vus, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vus = vec_add(a.vbs, b.vus);
    VERIFY_EQUAL(d, e);
    e.vs = vec_add(a.vs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_add(a.vs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_add(a.vbs, b.vs);
    VERIFY_EQUAL(d, e);
  }


  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_add(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = a.ui[j] + b.ui[j];
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint + uint: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vui = vec_add(a.vui, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vui = vec_add(a.vbi, b.vui);
    VERIFY_EQUAL(d, e);
    e.vi = vec_add(a.vi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_add(a.vi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_add(a.vbi, b.vi);
    VERIFY_EQUAL(d, e);
  }

  /* float  float */
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-1.0f, 1.0f);
    b.vf = rand_vf(-1.0f, 1.0f);
    d.vf = vec_add(a.vf, b.vf);
    for (j=0; j<4; j++) {
      float exp;
      exp = a.f[j] + b.f[j];
      if (exp != d.f[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tfloat + float: element=%d ", j);
	DEBUG(2, "a=%f b=%f d=%f exp=%f\n", a.f[j], b.f[j], d.f[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_addc()
{
  int i, j;

  DEBUG(1, "addc.....");

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_addc(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp, sum;
      sum = a.ui[j] + b.ui[j];
      exp =  ((sum < a.ui[j]) || (sum < b.ui[j])) ? 1 : 0;
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint + uint: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_adds()
{
  int i, j;

  DEBUG(1, "adds.....");

  /* unsigned char  unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_adds(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = a.uc[j] + b.uc[j];
      if (((a.uc[j] & b.uc[j]) | ((a.uc[j] ^ b.uc[j]) & ~exp)) & 0x80) {
	exp = 0xFF;
      }
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar + uchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.uc[j], b.uc[j], d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vuc = vec_adds(a.vuc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_adds(a.vbc, b.vuc);
    VERIFY_EQUAL(d, e);
  }


  /* signed char  signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vc = vec_adds(a.vc, b.vc);
    for (j=0; j<16; j++) {
      signed char exp;
      exp = a.c[j] + b.c[j];
      if ((a.c[j] <  0) && (b.c[j] <  0) && (exp >= 0)) exp = 0x80;
      if ((a.c[j] >= 0) && (b.c[j] >= 0) && (exp <  0)) exp = 0x7F;
      if (exp != d.c[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar + char: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.c[j], b.c[j], d.c[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vc = vec_adds(a.vc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_adds(a.vbc, b.vc);
    VERIFY_EQUAL(d, e);
  }

  /* unsigned short  unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_adds(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = a.us[j] + b.us[j];
      if (((a.us[j] & b.us[j]) | ((a.us[j] ^ b.us[j]) & ~exp)) & 0x8000) {
	exp = 0xFFFF;
      }
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort + ushort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.us[j], b.us[j], d.us[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vus = vec_adds(a.vus, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vus = vec_adds(a.vbs, b.vus);
    VERIFY_EQUAL(d, e);
  }

  /* signed short  signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vs = vec_adds(a.vs, b.vs);
    for (j=0; j<8; j++) {
      signed short exp;
      exp = a.s[j] + b.s[j];
      if ((a.s[j] <  0) && (b.s[j] <  0) && (exp >= 0)) exp = 0x8000;
      if ((a.s[j] >= 0) && (b.s[j] >= 0) && (exp <  0)) exp = 0x7FFF;
      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort + short: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.s[j], b.s[j], d.s[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vs = vec_adds(a.vs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_adds(a.vbs, b.vs);
    VERIFY_EQUAL(d, e);
  }

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_adds(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = a.ui[j] + b.ui[j];
      if (((a.ui[j] & b.ui[j]) | ((a.ui[j] ^ b.ui[j]) & ~exp)) & 0x80000000) {
	exp = 0xFFFFFFFF;
      }
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint + uint: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vui = vec_adds(a.vui, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vui = vec_adds(a.vbi, b.vui);
    VERIFY_EQUAL(d, e);
  }

  /* signed int  signed int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vi = vec_adds(a.vi, b.vi);
    for (j=0; j<4; j++) {
      signed int exp;
      exp = a.i[j] + b.i[j];
      if ((a.i[j] <  0) && (b.i[j] <  0) && (exp >= 0)) exp = 0x80000000;
      if ((a.i[j] >= 0) && (b.i[j] >= 0) && (exp <  0)) exp = 0x7FFFFFFF;
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint + int: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.i[j], b.i[j], d.i[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vi = vec_adds(a.vi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_adds(a.vbi, b.vi);
    VERIFY_EQUAL(d, e);
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_and()
{
  int i, j;

  DEBUG(1, "and......");

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_and(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = a.ui[j] & b.ui[j];
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint & uint : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vuc = vec_and(a.vuc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_and(a.vuc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_and(a.vbc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_and(a.vbc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_and(a.vc, b.vc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_and(a.vc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_and(a.vbc, b.vc);
    VERIFY_EQUAL(d, e);

    e.vus = vec_and(a.vus, b.vus);
    VERIFY_EQUAL(d, e);
    e.vus = vec_and(a.vbs, b.vus);
    VERIFY_EQUAL(d, e);
    e.vus = vec_and(a.vus, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_and(a.vbs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_and(a.vs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_and(a.vbs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_and(a.vs, b.vbs);
    VERIFY_EQUAL(d, e);

    e.vui = vec_and(a.vbi, b.vui);
    VERIFY_EQUAL(d, e);
    e.vui = vec_and(a.vui, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_and(a.vbi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_and(a.vi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_and(a.vbi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_and(a.vi, b.vbi);
    VERIFY_EQUAL(d, e);

    e.vf = vec_and(a.vf, b.vf);
    VERIFY_EQUAL(d, e);
    e.vf = vec_and(a.vbi, b.vf);
    VERIFY_EQUAL(d, e);
    e.vf = vec_and(a.vf, b.vbi);
    VERIFY_EQUAL(d, e);
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_andc()
{
  int i, j;

  DEBUG(1, "andc.....");

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_andc(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = a.ui[j] & ~b.ui[j];
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint & ~uint : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vuc = vec_andc(a.vuc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_andc(a.vuc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_andc(a.vbc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_andc(a.vbc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_andc(a.vc, b.vc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_andc(a.vc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_andc(a.vbc, b.vc);
    VERIFY_EQUAL(d, e);

    e.vus = vec_andc(a.vus, b.vus);
    VERIFY_EQUAL(d, e);
    e.vus = vec_andc(a.vbs, b.vus);
    VERIFY_EQUAL(d, e);
    e.vus = vec_andc(a.vus, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_andc(a.vbs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_andc(a.vs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_andc(a.vbs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_andc(a.vs, b.vbs);
    VERIFY_EQUAL(d, e);

    e.vui = vec_andc(a.vbi, b.vui);
    VERIFY_EQUAL(d, e);
    e.vui = vec_andc(a.vui, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_andc(a.vbi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_andc(a.vi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_andc(a.vbi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_andc(a.vi, b.vbi);
    VERIFY_EQUAL(d, e);

    e.vf = vec_andc(a.vf, b.vf);
    VERIFY_EQUAL(d, e);
    e.vf = vec_andc(a.vbi, b.vf);
    VERIFY_EQUAL(d, e);
    e.vf = vec_andc(a.vf, b.vbi);
    VERIFY_EQUAL(d, e);
  }

  DEBUG(1, passed);
  return (0);
}



int test_vec_avg()
{
  int i, j;

  DEBUG(1, "avg......");

  /* unsigned char  unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_avg(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = (a.uc[j] >> 1) + (b.uc[j] >> 1) + ((a.uc[j] | b.uc[j]) & 1);
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.uc[j], b.uc[j], d.uc[j], exp);
	return (1);
      }
    }
  }

  /* signed char  signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vc = vec_avg(a.vc, b.vc);
    for (j=0; j<16; j++) {
      signed char exp;
      exp = (a.c[j] >> 1) + (b.c[j] >> 1) + ((a.c[j] | b.c[j]) & 1);
      if (exp != d.c[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.c[j], b.c[j], d.c[j], exp);
	return (1);
      }
    }
  }

  /* unsigned short  unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_avg(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = (a.us[j] >> 1) + (b.us[j] >> 1) + ((a.us[j] | b.us[j]) & 1);
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.us[j], b.us[j], d.us[j], exp);
	return (1);
      }
    }
  }

  /* signed short  signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vs = vec_avg(a.vs, b.vs);
    for (j=0; j<8; j++) {
      signed short exp;
      exp = (a.s[j] >> 1) + (b.s[j] >> 1) + ((a.s[j] | b.s[j]) & 1);
      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.s[j], b.s[j], d.s[j], exp);
	return (1);
      }
    }
  }

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_avg(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.ui[j] >> 1) + (b.ui[j] >> 1) + ((a.ui[j] | b.ui[j]) & 1);
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
  }

  /* signed int  signed int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vi = vec_avg(a.vi, b.vi);
    for (j=0; j<4; j++) {
      signed int exp;
      exp = (a.i[j] >> 1) + (b.i[j] >> 1) + ((a.i[j] | b.i[j]) & 1);
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.i[j], b.i[j], d.i[j], exp);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}

static struct _ceil_floor {
  float in;
  float ceil_out;
  float floor_out;
} ceil_floor_samples[] = {
  { 0.0f, 0.0f, 0.0f },
  { 0.0001f, 1.0f, 0.0f },
  { -0.0f, 0.0f, 0.0f },
  { -0.0001f, 0.0f, -1.0f },

  { -0.9999f, 0.0f, -1.0f },
  { 0.9999f, 1.0f, 0.0f },
  { -0.5f, 0.0f, -1.0f },
  { 0.5f, 1.0f, 0.0f },

  { 250.2f, 251.0f, 250.0f },
  { -250.3f, -250.0f, -251.0f },
  { 129.01f, 130.0f, 129.0f },
  { -129.01f, -129.0f, -130.0f },

  { 18.99f, 19.0f, 18.0f },
  { 123456792.0f, 123456792.0f, 123456792.0f },
  { -123456792.0f, -123456792.0f, -123456792.0f },
};


int test_vec_ceil()
{
  int i, j;

  DEBUG(1, "ceil.....");
  for (i=0; i<(int)(sizeof(ceil_floor_samples)/(4*sizeof(struct _ceil_floor))); i++) {
    a.f[0] = ceil_floor_samples[4*i+0].in;
    a.f[1] = ceil_floor_samples[4*i+1].in;
    a.f[2] = ceil_floor_samples[4*i+2].in;
    a.f[3] = ceil_floor_samples[4*i+3].in;
    d.vf = vec_ceil(a.vf);

    for (j=0; j<4; j++) {
      float exp;

      exp = ceil_floor_samples[4*i+j].ceil_out;
      if (d.f[j] != exp) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%f d=%f exp=%f\n", a.f[j], d.f[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}



int test_vec_cmpb()
{
  int i, j;

  DEBUG(1, "cmpb.....");

  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-3.0, 3.0);
    b.vf = rand_vf(-3.0, 3.0);
    d.vi = vec_cmpb(a.vf, b.vf);
    for (j=0; j<4; j++) {
      signed int exp;
      exp = 0;
      if (a.f[j] >  b.f[j]) exp |= 0x80000000;
      if (a.f[j] < -b.f[j]) exp |= 0x40000000;

      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%f b=%f d=0x%x exp=0x%x\n", a.f[j], b.f[j], d.i[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_cmpeq()
{
  int i, j;

  DEBUG(1, "cmpeq....");

  /* unsigned char  unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.uc[i&15] = a.uc[i&15];
    d.vbc = vec_cmpeq(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = (a.uc[j] == b.uc[j]) ? 0xFF : 0;
      if (exp != d.bc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.uc[j], b.uc[j], d.bc[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vbc = vec_cmpeq(a.vc, b.vc);
    VERIFY_EQUAL(d, e);
  }

  /* unsigned short  unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.us[i&7] = a.us[i&7];
    d.vbs = vec_cmpeq(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = (a.us[j] == b.us[j]) ? 0xFFFF : 0;
      if (exp != d.bs[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.us[j], b.us[j], d.bs[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vbs = vec_cmpeq(a.vs, b.vs);
    VERIFY_EQUAL(d, e);
  }

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.ui[i&3] = a.ui[i&3];
    d.vbi = vec_cmpeq(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.ui[j] == b.ui[j]) ? 0xFFFFFFFF : 0;
      if (exp != d.bi[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.ui[j], b.ui[j], d.bi[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vbi = vec_cmpeq(a.vi, b.vi);
    VERIFY_EQUAL(d, e);
  }

  /* float  float */
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-2.0f, 2.0f);
    b.vf = rand_vf(-2.0f, 2.0f);
    if (i % 3) b.f[i&3] = a.f[i&3];
    d.vbi = vec_cmpeq(a.vf, b.vf);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.f[j] == b.f[j]) ? 0xFFFFFFFF : 0;
      if (exp != d.bi[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tfloat : element=%d ", j);
	DEBUG(2, "a=%f b=%f d=0x%x exp=0x%x\n", a.f[j], b.f[j], d.bi[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_cmpge()
{
  int i, j;

  DEBUG(1, "cmpge....");

  /* float  float */
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-2.0f, 2.0f);
    b.vf = rand_vf(-2.0f, 2.0f);
    if (i % 3) b.f[i&3] = a.f[i&3];
    d.vbi = vec_cmpge(a.vf, b.vf);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.f[j] >= b.f[j]) ? 0xFFFFFFFF : 0;
      if (exp != d.bi[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tfloat : element=%d ", j);
	DEBUG(2, "a=%f b=%f d=0x%x exp=0x%x\n", a.f[j], b.f[j], d.bi[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_cmpgt()
{
  int i, j;

  DEBUG(1, "cmpgt....");

  /* unsigned char  unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.uc[i&15] = a.uc[i&15];
    d.vbc = vec_cmpgt(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = (a.uc[j] > b.uc[j]) ? 0xFF : 0;
      if (exp != d.bc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.uc[j], b.uc[j], d.bc[j], exp);
	return (1);
      }
    }
  }

  /* signed char  signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.c[i&15] = a.c[i&15];
    d.vbc = vec_cmpgt(a.vc, b.vc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = (a.c[j] > b.c[j]) ? 0xFF : 0;
      if (exp != d.bc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar : element=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", a.c[j], b.c[j], d.bc[j], exp);
	return (1);
      }
    }
  }

  /* unsigned short  unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.us[i&7] = a.us[i&7];
    d.vbs = vec_cmpgt(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = (a.us[j] > b.us[j]) ? 0xFFFF : 0;
      if (exp != d.bs[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.us[j], b.us[j], d.bs[j], exp);
	return (1);
      }
    }
  }

  /* signed short  signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.s[i&7] = a.s[i&7];
    d.vbs = vec_cmpgt(a.vs, b.vs);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = (a.s[j] > b.s[j]) ? 0xFFFF : 0;
      if (exp != d.bs[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort : element=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", a.s[j], b.s[j], d.bs[j], exp);
	return (1);
      }
    }
  }

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.ui[i&3] = a.ui[i&3];
    d.vbi = vec_cmpgt(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.ui[j] > b.ui[j]) ? 0xFFFFFFFF : 0;
      if (exp != d.bi[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.ui[j], b.ui[j], d.bi[j], exp);
	return (1);
      }
    }
  }

  /* signed int  signed int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.i[i&3] = a.i[i&3];
    d.vbi = vec_cmpgt(a.vi, b.vi);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.i[j] > b.i[j]) ? 0xFFFFFFFF : 0;
      if (exp != d.bi[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint : element=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", a.i[j], b.i[j], d.bi[j], exp);
	return (1);
      }
    }
  }

  /* float  float */
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-2.0f, 2.0f);
    b.vf = rand_vf(-2.0f, 2.0f);
    if (i % 3) b.f[i&3] = a.f[i&3];
    d.vbi = vec_cmpgt(a.vf, b.vf);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.f[j] > b.f[j]) ? 0xFFFFFFFF : 0;
      if (exp != d.bi[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tfloat : element=%d ", j);
	DEBUG(2, "a=%f b=%f d=0x%x exp=0x%x\n", a.f[j], b.f[j], d.bi[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_cmple()
{
  int i, j;

  DEBUG(1, "cmple....");


  /* float  float */
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-2.0f, 2.0f);
    b.vf = rand_vf(-2.0f, 2.0f);
    if (i % 3) b.f[i&3] = a.f[i&3];
    d.vbi = vec_cmple(a.vf, b.vf);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.f[j] <= b.f[j]) ? 0xFFFFFFFF : 0;
      if (exp != d.bi[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tfloat : element=%d ", j);
	DEBUG(2, "a=%f b=%f d=0x%x exp=0x%x\n", a.f[j], b.f[j], d.bi[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_cmplt()
{
  int i, j;

  DEBUG(1, "cmplt....");

  /* unsigned char  unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.uc[i&15] = a.uc[i&15];
    d.vbc = vec_cmplt(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = (a.uc[j] < b.uc[j]) ? 0xFF : 0;
      if (exp != d.bc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.uc[j], b.uc[j], d.bc[j], exp);
	return (1);
      }
    }
  }

  /* signed char  signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.c[i&15] = a.c[i&15];
    d.vbc = vec_cmplt(a.vc, b.vc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = (a.c[j] < b.c[j]) ? 0xFF : 0;
      if (exp != d.bc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar : element=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", a.c[j], b.c[j], d.bc[j], exp);
	return (1);
      }
    }
  }

  /* unsigned short  unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.us[i&7] = a.us[i&7];
    d.vbs = vec_cmplt(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = (a.us[j] < b.us[j]) ? 0xFFFF : 0;
      if (exp != d.bs[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.us[j], b.us[j], d.bs[j], exp);
	return (1);
      }
    }
  }

  /* signed short  signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.s[i&7] = a.s[i&7];
    d.vbs = vec_cmplt(a.vs, b.vs);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = (a.s[j] < b.s[j]) ? 0xFFFF : 0;
      if (exp != d.bs[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort : element=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", a.s[j], b.s[j], d.bs[j], exp);
	return (1);
      }
    }
  }

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.ui[i&3] = a.ui[i&3];
    d.vbi = vec_cmplt(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.ui[j] < b.ui[j]) ? 0xFFFFFFFF : 0;
      if (exp != d.bi[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.ui[j], b.ui[j], d.bi[j], exp);
	return (1);
      }
    }
  }

  /* signed int  signed int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    if (i % 3) b.i[i&3] = a.i[i&3];
    d.vbi = vec_cmplt(a.vi, b.vi);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.i[j] < b.i[j]) ? 0xFFFFFFFF : 0;
      if (exp != d.bi[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint : element=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", a.i[j], b.i[j], d.bi[j], exp);
	return (1);
      }
    }
  }

  /* float  float */
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-2.0f, 2.0f);
    b.vf = rand_vf(-2.0f, 2.0f);
    if (i % 3) b.f[i&3] = a.f[i&3];
    d.vbi = vec_cmplt(a.vf, b.vf);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.f[j] < b.f[j]) ? 0xFFFFFFFF : 0;
      if (exp != d.bi[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tfloat : element=%d ", j);
	DEBUG(2, "a=%f b=%f d=0x%x exp=0x%x\n", a.f[j], b.f[j], d.bi[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_ctf()
{
  int i, j;
  int scale;
  unsigned int exponent;
  union {
    float f;
    unsigned int ui;
  } u_exp;
  

  DEBUG(1, "ctf......");

  /* unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    if (i & 1) {
      d.vf = vec_ctf(a.vui, 31);
      scale = 31;
    } else {
      d.vf = vec_ctf(a.vui, 0);
      scale = 0;
    }
    for (j=0; j<4; j++) {
      float exp;
      u_exp.f = (float)(a.ui[j]);
      
      exponent = (u_exp.ui >> 23) & 0xFF;
      exponent -= scale;
      u_exp.ui = (u_exp.ui & 0x807FFFFF) | (exponent << 23);
      exp = u_exp.f;
      if (exp != d.f[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=0x%x scale=%d d=%f exp=%f\n", a.ui[j], scale, d.f[j], exp);
	return (1);
      }
    }
  }

  /* signed int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    if (i & 1) {
      d.vf = vec_ctf(a.vi, 31);
      scale = 31;
    } else {
      d.vf = vec_ctf(a.vi, 0);
      scale = 0;
    }
    for (j=0; j<4; j++) {
      float exp;
      u_exp.f = (float)(a.i[j]);
      
      exponent = (u_exp.ui >> 23) & 0xFF;
      exponent -= scale;
      u_exp.ui = (u_exp.ui & 0x807FFFFF) | (exponent << 23);
      exp = u_exp.f;
      if (exp != d.f[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%d scale=%d d=%f exp=%f\n", a.i[j], scale, d.f[j], exp);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}

int test_vec_cts()
{
  int i, j;
  int scale;
  unsigned int exponent;
  union {
    float f;
    unsigned int ui;
  } u_exp;

  DEBUG(1, "cts......");

  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-8.0f, 8.0f);
    if (i & 1) {
      d.vi = vec_cts(a.vf, 31);
      scale = 31;
    } else {
      d.vi = vec_cts(a.vf, 0);
      scale = 0;
    }
    for (j=0; j<4; j++) {
      signed int exp;

      u_exp.f = a.f[j];
      exponent = (u_exp.ui >> 23) & 0xFF;
      exponent += scale;
      if (exponent > 157) {
	exp = (u_exp.ui & 0x80000000) ? 0x80000000 : 0x7FFFFFFF;
      } else {
	u_exp.ui = (u_exp.ui & 0x807FFFFF) | (exponent << 23);
	exp = (signed int)(u_exp.f);
      }

      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%f scale=%d d=%d exp=%d\n", a.f[j], scale, d.i[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_ctu()
{
  int i, j;
  int scale;
  unsigned int exponent;
  union {
    float f;
    unsigned int ui;
  } u_exp;


  DEBUG(1, "ctu......");
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-1.0f, 8.0f);
    if (i & 1) {
      d.vui = vec_ctu(a.vf, 31);
      scale = 31;
    } else {
      d.vui = vec_ctu(a.vf, 0);
      scale = 0;
    }
    for (j=0; j<4; j++) {
      unsigned int exp;

      u_exp.f = a.f[j];
      exponent = (u_exp.ui >> 23) & 0xFF;
      exponent += scale;
      if (exponent > 158) {
	exp = 0xFFFFFFFF;
      } else if (u_exp.f < 0.0f) {
	exp = 0x0;
      } else {
	u_exp.ui = (u_exp.ui & 0x807FFFFF) | (exponent << 23);
	exp = (unsigned int)(u_exp.f);
      }

      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%f scale=%d d=0x%x exp=0x%x\n", a.f[j], scale, d.ui[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_dss()
{
  DEBUG(1, "dss......");
  vec_dss(0);
  vec_dss(1);
  vec_dss(2);
  vec_dss(3);
  DEBUG(1, passed);
  return (0);
}


int test_vec_dssall()
{
  DEBUG(1, "dssall...");
  vec_dssall();
  DEBUG(1, passed);
  return (0);
}

int test_vec_dst()
{
  DEBUG(1, "dst......");
  vec_dst(data, 40, 3);
  DEBUG(1, passed);
  return (0);
}


int test_vec_dstst()
{
  DEBUG(1, "dstst....");
  vec_dstst(data, 13, 1);
  DEBUG(1, passed);
  return (0);
}

int test_vec_dststt()
{
  DEBUG(1, "dststt...");
  vec_dststt(data, 0, 2);
  DEBUG(1, passed);
  return (0);
}

int test_vec_dstt()
{
  DEBUG(1, "dstt.....");
  vec_dstst(data, 0, 0);
  DEBUG(1, passed);
  return (0);
}


static struct _expte {
  float in;
  float out;
} expte_samples[] = {
  { -3.0f, 0.125f },
  { -2.0f, 0.25f },
  { 2.0f, 4.0f },
  { 5.0f, 32.0f },

  { 0.0f, 1.0f },
  { 0.5f, 1.4142135f },
  { 4.1f, 17.148375f },
  { -4.8f, 0.03589682f },

  { -1.4f, 0.378929f },
  { 0.7f, 1.6245f },
  { 0.1f, 1.07177f },
  { -0.8f, 0.574349f },
};


int test_vec_expte()
{
  int i, j;

  DEBUG(1, "expte....");

  for (i=0; i<(int)(sizeof(expte_samples)/(4*sizeof(struct _expte))); i++) {
    a.f[0] = expte_samples[4*i+0].in;
    a.f[1] = expte_samples[4*i+1].in;
    a.f[2] = expte_samples[4*i+2].in;
    a.f[3] = expte_samples[4*i+3].in;
    d.vf = vec_expte(a.vf);

    for (j=0; j<4; j++) {
      float exp;
      float delta;

      exp = expte_samples[4*i+j].out;
      delta = (d.f[j] - exp) / exp;
      if (delta < 0.0f) delta = -delta;

      if (delta > 0.021f) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%f d=%f exp=%f delta=%f\n", a.f[j], d.f[j], exp, delta);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_floor()
{
  int i, j;

  DEBUG(1, "floor....");
  for (i=0; i<(int)(sizeof(ceil_floor_samples)/(4*sizeof(struct _ceil_floor))); i++) {
    a.f[0] = ceil_floor_samples[4*i+0].in;
    a.f[1] = ceil_floor_samples[4*i+1].in;
    a.f[2] = ceil_floor_samples[4*i+2].in;
    a.f[3] = ceil_floor_samples[4*i+3].in;
    d.vf = vec_floor(a.vf);

    for (j=0; j<4; j++) {
      float exp;

      exp = ceil_floor_samples[4*i+j].floor_out;
      if (d.f[j] != exp) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%f d=%f exp=%f\n", a.f[j], d.f[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_ld()
{
  int i, j;
  int base, offset;

  DEBUG(1, "ld.......");

  for (i=0; i<(int)(sizeof(data)); i++) data[i] = i;

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    d.vuc = vec_ld(offset, (vec_uchar16 *)(&data[base])); 

    for (j=0; j<16; j++) {
      unsigned char exp;

      exp = (((unsigned)base + (unsigned)offset) & ~0xF) + j;

      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", offset, base, d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vuc = vec_ld(offset, (unsigned char *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vc = vec_ld(offset, (vec_char16 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vc = vec_ld(offset, (signed char *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbc = vec_ld(offset, (vec_bchar16 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vus = vec_ld(offset, (vec_ushort8 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vus = vec_ld(offset, (unsigned short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vs = vec_ld(offset, (vec_short8 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vs = vec_ld(offset, (signed short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbs = vec_ld(offset, (vec_bshort8 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vp = vec_ld(offset, (vec_pixel8 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vui = vec_ld(offset, (vec_uint4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vui = vec_ld(offset, (unsigned int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vi = vec_ld(offset, (vec_int4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vi = vec_ld(offset, (signed int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbi = vec_ld(offset, (vec_bint4 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vf = vec_ld(offset, (vec_float4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vf = vec_ld(offset, (float *)(&data[base]));
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_lde()
{
  int i, j;
  int base, offset;
  int ea;

  DEBUG(1, "lde......");

  for (i=0; i<(int)(sizeof(data)); i++) data[i] = i;

  /* char */
  for (i=0; i<samples; i++) {
    unsigned char exp;

    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    d.vuc = vec_lde(offset, (unsigned char *)(&data[base])); 

    ea = (unsigned)base + (unsigned)offset;
    j = ea & 0xF;
    exp = (((unsigned)base + (unsigned)offset) & ~0xF) + j;

    if (exp != d.uc[j]) {
      DEBUG(1, failed);
      DEBUG(2, "\telement=%d ", j);
      DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", offset, base, d.uc[j], exp);
      return (1);
    }
    /* verify equivalent variations */
    e.vc = vec_lde(offset, (signed char *)(&data[base]));
    VERIFY_EQUAL(d, e);
  }


  /* short */
  for (i=0; i<samples; i++) {
    unsigned short exp;

    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    d.vus = vec_lde(offset, (unsigned short *)(&data[base])); 

    ea = (unsigned)base + (unsigned)offset;
    j = (ea >> 1) & 0x7;
    exp = (((unsigned)base + (unsigned)offset)  & ~0xF)  + 2*j;
    exp = (exp << 8) | (exp + 1);

    if (exp != d.us[j]) {
      DEBUG(1, failed);
      DEBUG(2, "\telement=%d ", j);
      DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", offset, base, d.us[j], exp);
      return (1);
    }
    /* verify equivalent variations */
    e.vs = vec_lde(offset, (signed short *)(&data[base]));
    VERIFY_EQUAL(d, e);
  }

  /* int and float */
  for (i=0; i<samples; i++) {
    unsigned int exp;

    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    d.vui = vec_lde(offset, (unsigned int *)(&data[base])); 

    ea = (unsigned)base + (unsigned)offset;
    j = (ea >> 2) & 0x3;
    exp = (((unsigned)base + (unsigned)offset)  & ~0xF)  + 4*j;
    exp = ((exp) << 24) | ((exp+1) << 16) | ((exp+2) << 8) | (exp+3);

    if (exp != d.ui[j]) {
      DEBUG(1, failed);
      DEBUG(2, "\telement=%d ", j);
      DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", offset, base, d.ui[j], exp);
      return (1);
    }
    /* verify equivalent variations */
    e.vi = vec_lde(offset, (signed int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vf = vec_lde(offset, (float *)(&data[base]));
    VERIFY_EQUAL(d, e);
  }

  DEBUG(1, passed);
  return (0);
}

int test_vec_ldl()
{
  int i, j;
  int base, offset;

  DEBUG(1, "ldl......");

  for (i=0; i<(int)(sizeof(data)); i++) data[i] = i;

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    d.vuc = vec_ldl(offset, (vec_uchar16 *)(&data[base])); 

    for (j=0; j<16; j++) {
      unsigned char exp;

      exp = (((unsigned)base + (unsigned)offset) & ~0xF) + j;

      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", offset, base, d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vuc = vec_ldl(offset, (unsigned char *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vc = vec_ldl(offset, (vec_char16 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vc = vec_ldl(offset, (signed char *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbc = vec_ldl(offset, (vec_bchar16 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vus = vec_ldl(offset, (vec_ushort8 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vus = vec_ldl(offset, (unsigned short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vs = vec_ldl(offset, (vec_short8 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vs = vec_ldl(offset, (signed short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbs = vec_ldl(offset, (vec_bshort8 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vp = vec_ldl(offset, (vec_pixel8 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vui = vec_ldl(offset, (vec_uint4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vui = vec_ldl(offset, (unsigned int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vi = vec_ldl(offset, (vec_int4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vi = vec_ldl(offset, (signed int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbi = vec_ldl(offset, (vec_bint4 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vf = vec_ldl(offset, (vec_float4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vf = vec_ldl(offset, (float *)(&data[base]));
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_loge()
{
  int i, j;

  DEBUG(1, "loge.....");

  for (i=0; i<(int)(sizeof(expte_samples)/(4*sizeof(struct _expte))); i++) {
    a.f[0] = expte_samples[4*i+0].out;
    a.f[1] = expte_samples[4*i+1].out;
    a.f[2] = expte_samples[4*i+2].out;
    a.f[3] = expte_samples[4*i+3].out;
    d.vf = vec_loge(a.vf);

    for (j=0; j<4; j++) {
      float exp;
      float delta;

      exp = expte_samples[4*i+j].in;
      if (exp > 0.00001) {
	delta = (d.f[j] - exp) / exp;
      } else {
	delta = d.f[j] - exp;
      }
      if (delta < 0.0f) delta = -delta;

      if (delta > 0.06f) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%f d=%f exp=%f delta=%f\n", a.f[j], d.f[j], exp, delta);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_lvsl()
{
  int i, j;
  int base, offset;

  DEBUG(1, "lvsl.....");

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    d.vuc = vec_lvsl(offset, (unsigned char *)(&data[base])); 

    for (j=0; j<16; j++) {
      unsigned char exp;

      exp = (((unsigned)base + (unsigned)offset) & 0xF) + j;

      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", offset, base, d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vuc = vec_lvsl(offset, (signed char *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vuc = vec_lvsl(offset, (unsigned short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vuc = vec_lvsl(offset, (signed short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vuc = vec_lvsl(offset, (unsigned int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vuc = vec_lvsl(offset, (signed int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vuc = vec_lvsl(offset, (float *)(&data[base]));
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_lvsr()
{
  int i, j;
  int base, offset;

  DEBUG(1, "lvsr.....");

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    d.vuc = vec_lvsr(offset, (unsigned char *)(&data[base])); 

    for (j=0; j<16; j++) {
      unsigned char exp;

      exp = 16 - (((unsigned)base + (unsigned)offset) & 0xF) + j;

      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", offset, base, d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vuc = vec_lvsr(offset, (signed char *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vuc = vec_lvsr(offset, (unsigned short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vuc = vec_lvsr(offset, (signed short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vuc = vec_lvsr(offset, (unsigned int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vuc = vec_lvsr(offset, (signed int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vuc = vec_lvsr(offset, (float *)(&data[base]));
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_madd()
{
  int i, j;

  DEBUG(1, "madd.....");

  /* float */
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-1.0f, 1.0f);
    b.vf = rand_vf(-1.0f, 1.0f);
    c.vf = rand_vf(-1.0f, 1.0f);
    d.vf = vec_madd(a.vf, b.vf, c.vf);
    for (j=0; j<4; j++) {
      float exp;
      float delta;
      exp = a.f[j] * b.f[j] + c.f[j];
      delta = exp - d.f[j];
      if (delta < 0.0f) delta = -delta;
      if (delta > 0.00001f) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d a=%f ", j, a.f[j]);
	DEBUG(2, "b=%f c=%f d=%f exp=%f\n", b.f[j], c.f[j], d.f[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_madds()
{
  int i, j;

  DEBUG(1, "madds....");

  /* short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    c.vui = rand_vui();
    d.vs = vec_madds(a.vs, b.vs, c.vs);
    for (j=0; j<8; j++) {
      signed int exp;
      exp = (((signed int)a.s[j] * (signed int)b.s[j]) >> 15) + (signed int)c.s[j];
      if (exp >  32767) exp =  32767;
      if (exp < -32768) exp = -32768;
      if ((signed short)(exp) != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d a=%d ", j, a.s[j]);
	DEBUG(2, "b=%d c=%d d=%d exp=%d\n", b.s[j], c.s[j], d.s[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_max()
{
  int i, j;

  DEBUG(1, "max......");

  /* unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_max(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = (a.uc[j] > b.uc[j]) ? a.uc[j] : b.uc[j];
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.uc[j], b.uc[j], d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vuc = vec_max(a.vuc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_max(a.vbc, b.vuc);
    VERIFY_EQUAL(d, e);
  }

  /* signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vc = vec_max(a.vc, b.vc);
    for (j=0; j<16; j++) {
      signed char exp;
      exp = (a.c[j] > b.c[j]) ? a.c[j] : b.c[j];
      if (exp != d.c[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.c[j], b.c[j], d.c[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vc = vec_max(a.vc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_max(a.vbc, b.vc);
    VERIFY_EQUAL(d, e);
  }

  /* unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_max(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = (a.us[j] > b.us[j]) ? a.us[j] : b.us[j];
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.us[j], b.us[j], d.us[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vus = vec_max(a.vus, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vus = vec_max(a.vbs, b.vus);
    VERIFY_EQUAL(d, e);
  }

  /* signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vs = vec_max(a.vs, b.vs);
    for (j=0; j<8; j++) {
      signed short exp;
      exp = (a.s[j] > b.s[j]) ? a.s[j] : b.s[j];
      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.s[j], b.s[j], d.s[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vs = vec_max(a.vs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_max(a.vbs, b.vs);
    VERIFY_EQUAL(d, e);
  }

  /* unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_max(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.ui[j] > b.ui[j]) ? a.ui[j] : b.ui[j];
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vui = vec_max(a.vui, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vui = vec_max(a.vbi, b.vui);
    VERIFY_EQUAL(d, e);
  }

  /* signed int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vi = vec_max(a.vi, b.vi);
    for (j=0; j<4; j++) {
      signed int exp;
      exp = (a.i[j] > b.i[j]) ? a.i[j] : b.i[j];
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.i[j], b.i[j], d.i[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vi = vec_max(a.vi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_max(a.vbi, b.vi);
    VERIFY_EQUAL(d, e);
  }

  /* float */
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-1.0f, 1.0f);
    b.vf = rand_vf(-1.0f, 1.0f);
    d.vf = vec_max(a.vf, b.vf);
    for (j=0; j<4; j++) {
      float exp;
      exp = (a.f[j] > b.f[j]) ? a.f[j] : b.f[j];
      if (exp != d.f[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tfloat: element=%d ", j);
	DEBUG(2, "a=%f b=%f d=%f exp=%f\n", a.f[j], b.f[j], d.f[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_mergeh()
{
  int i, j;

  DEBUG(1, "mergeh...");

  /* byte */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_mergeh(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = (j & 1) ? b.uc[j>>1] : a.uc[j>>1];
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tbyte: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vc = vec_mergeh(a.vc, b.vc);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_mergeh(a.vbc, b.vbc);
    VERIFY_EQUAL(d, e);
  }

  /* halfword */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_mergeh(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = (j & 1) ? b.us[j>>1] : a.us[j>>1];
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\thalfword: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.us[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vs = vec_mergeh(a.vs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_mergeh(a.vbs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vp = vec_mergeh(a.vp, b.vp);
    VERIFY_EQUAL(d, e);
  }

  /* word */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_mergeh(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (j & 1) ? b.ui[j>>1] : a.ui[j>>1];
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tword: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.ui[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vi = vec_mergeh(a.vi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_mergeh(a.vbi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vf = vec_mergeh(a.vf, b.vf);
    VERIFY_EQUAL(d, e);
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_mergel()
{
  int i, j;

  DEBUG(1, "mergel...");

  /* byte */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_mergel(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = (j & 1) ? b.uc[8+(j>>1)] : a.uc[8+(j>>1)];
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tbyte: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vc = vec_mergel(a.vc, b.vc);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_mergel(a.vbc, b.vbc);
    VERIFY_EQUAL(d, e);
  }

  /* halfword */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_mergel(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = (j & 1) ? b.us[4+(j>>1)] : a.us[4+(j>>1)];
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\thalfword: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.us[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vs = vec_mergel(a.vs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_mergel(a.vbs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vp = vec_mergel(a.vp, b.vp);
    VERIFY_EQUAL(d, e);
  }

  /* word */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_mergel(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (j & 1) ? b.ui[2+(j>>1)] : a.ui[2+(j>>1)];
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tword: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.ui[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vi = vec_mergel(a.vi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_mergel(a.vbi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vf = vec_mergel(a.vf, b.vf);
    VERIFY_EQUAL(d, e);
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_mfvscr()
{
  vec_ushort8 exp;

  DEBUG(1, "mfvscr...");
  exp = vec_mfvscr();
  DEBUG(1, passed);
  return (0);
}


int test_vec_min()
{
  int i, j;

  DEBUG(1, "min......");

  /* unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_min(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = (a.uc[j] < b.uc[j]) ? a.uc[j] : b.uc[j];
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.uc[j], b.uc[j], d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vuc = vec_min(a.vuc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_min(a.vbc, b.vuc);
    VERIFY_EQUAL(d, e);
  }

  /* signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vc = vec_min(a.vc, b.vc);
    for (j=0; j<16; j++) {
      signed char exp;
      exp = (a.c[j] < b.c[j]) ? a.c[j] : b.c[j];
      if (exp != d.c[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.c[j], b.c[j], d.c[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vc = vec_min(a.vc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_min(a.vbc, b.vc);
    VERIFY_EQUAL(d, e);
  }

  /* unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_min(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = (a.us[j] < b.us[j]) ? a.us[j] : b.us[j];
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.us[j], b.us[j], d.us[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vus = vec_min(a.vus, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vus = vec_min(a.vbs, b.vus);
    VERIFY_EQUAL(d, e);
  }

  /* signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vs = vec_min(a.vs, b.vs);
    for (j=0; j<8; j++) {
      signed short exp;
      exp = (a.s[j] < b.s[j]) ? a.s[j] : b.s[j];
      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.s[j], b.s[j], d.s[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vs = vec_min(a.vs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_min(a.vbs, b.vs);
    VERIFY_EQUAL(d, e);
  }

  /* unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_min(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.ui[j] < b.ui[j]) ? a.ui[j] : b.ui[j];
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vui = vec_min(a.vui, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vui = vec_min(a.vbi, b.vui);
    VERIFY_EQUAL(d, e);
  }

  /* signed int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vi = vec_min(a.vi, b.vi);
    for (j=0; j<4; j++) {
      signed int exp;
      exp = (a.i[j] < b.i[j]) ? a.i[j] : b.i[j];
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.i[j], b.i[j], d.i[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vi = vec_min(a.vi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_min(a.vbi, b.vi);
    VERIFY_EQUAL(d, e);
  }

  /* float */
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-1.0f, 1.0f);
    b.vf = rand_vf(-1.0f, 1.0f);
    d.vf = vec_min(a.vf, b.vf);
    for (j=0; j<4; j++) {
      float exp;
      exp = (a.f[j] < b.f[j]) ? a.f[j] : b.f[j];
      if (exp != d.f[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tfloat: element=%d ", j);
	DEBUG(2, "a=%f b=%f d=%f exp=%f\n", a.f[j], b.f[j], d.f[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_mladd()
{
  int i, j;

  DEBUG(1, "mladd....");

  /* float */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    c.vui = rand_vui();
    d.vus = vec_mladd(a.vus, b.vus, c.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = a.us[j] * b.us[j] + c.us[j];
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d a=%d ", j, a.us[j]);
	DEBUG(2, "b=%d c=%d d=%d exp=%d\n", b.us[j], c.us[j], d.us[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vs = vec_mladd(a.vus, b.vs, c.vs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_mladd(a.vs, b.vus, c.vus);
    VERIFY_EQUAL(d, e);
    e.vs = vec_mladd(a.vs, b.vs, c.vs);
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}



int test_vec_mradds()
{
  int i, j;

  DEBUG(1, "mradds...");

  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    c.vui = rand_vui();
    d.vs = vec_mradds(a.vs, b.vs, c.vs);
    for (j=0; j<8; j++) {
      signed int exp;
      exp = (((signed int)a.s[j] * (signed int)b.s[j] + 16384) >> 15) + (signed int)c.s[j];
      if (exp >  32767) exp = 32767;
      if (exp < -32768) exp = -32768;
      if ((signed short)(exp) != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\t: element=%d a=%d ", j, a.s[j]);
	DEBUG(2, "b=%d c=%d d=%d exp=%d\n", b.s[j], c.s[j], d.s[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_msum()
{
  int i, j;

  DEBUG(1, "msum.....");

  /* unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    c.vui = rand_vui();
    d.vui = vec_msum(a.vuc, b.vuc, c.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (((unsigned int)a.uc[4*j+0] * (unsigned int)b.uc[4*j+0]) +
	     ((unsigned int)a.uc[4*j+1] * (unsigned int)b.uc[4*j+1]) +
	     ((unsigned int)a.uc[4*j+2] * (unsigned int)b.uc[4*j+2]) +
	     ((unsigned int)a.uc[4*j+3] * (unsigned int)b.uc[4*j+3]) +
	     (unsigned int)c.ui[j]);
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar: element=%d ", j);
	DEBUG(2, "a=%d %d %d %d ", a.uc[4*j+0], a.uc[4*j+1], a.uc[4*j+2], a.uc[4*j+3]);
	DEBUG(2, "b=%d %d %d %d ", b.uc[4*j+0], b.uc[4*j+1], b.uc[4*j+2], b.uc[4*j+3]);
	DEBUG(2, "c=%d d=%d exp=%d\n",  c.ui[j], d.ui[j], exp);
	return (1);
      }
    }
  }

  /* signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    c.vui = rand_vui();
    d.vi = vec_msum(a.vc, b.vuc, c.vi);
    for (j=0; j<4; j++) {
      signed int exp;
      exp = (((signed int)a.c[4*j+0] * (signed int)b.uc[4*j+0]) +
	     ((signed int)a.c[4*j+1] * (signed int)b.uc[4*j+1]) +
	     ((signed int)a.c[4*j+2] * (signed int)b.uc[4*j+2]) +
	     ((signed int)a.c[4*j+3] * (signed int)b.uc[4*j+3]) +
	     (signed int)c.i[j]);
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar: element=%d ", j);
	DEBUG(2, "a=%d %d %d %d ", a.c[4*j+0], a.c[4*j+1], a.c[4*j+2], a.c[4*j+3]);
	DEBUG(2, "b=%d %d %d %d ", b.uc[4*j+0], b.uc[4*j+1], b.uc[4*j+2], b.uc[4*j+3]);
	DEBUG(2, "c=%d d=%d exp=%d\n",  c.i[j], d.i[j], exp);
	return (1);
      }
    }
  }

  /* unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    c.vui = rand_vui();
    d.vui = vec_msum(a.vus, b.vus, c.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (((unsigned int)a.us[2*j+0] * (unsigned int)b.us[2*j+0]) +
	     ((unsigned int)a.us[2*j+1] * (unsigned int)b.us[2*j+1]) +
	     (unsigned int)c.ui[j]);
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort: element=%d ", j);
	DEBUG(2, "a=%d %d b=%d %d ", a.us[2*j+0], a.us[2*j+1], b.us[2*j+0], a.us[2*j+1]);
	DEBUG(2, "c=%d d=%d exp=%d\n",  c.ui[j], d.ui[j], exp);
	return (1);
      }
    }
  }

  /* signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    c.vui = rand_vui();
    d.vi = vec_msum(a.vs, b.vs, c.vi);
    for (j=0; j<4; j++) {
      signed int exp;
      exp = (((signed int)a.s[2*j+0] * (signed int)b.s[2*j+0]) +
	     ((signed int)a.s[2*j+1] * (signed int)b.s[2*j+1]) +
	     (signed int)c.i[j]);
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d ", j);
	DEBUG(2, "a=%d %d b=%d %d ", a.s[2*j+0], a.s[2*j+1], b.s[2*j+0], a.s[2*j+1]);
	DEBUG(2, "c=%d d=%d exp=%d\n",  c.i[j], d.i[j], exp);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_msums()
{
  int i, j;

  DEBUG(1, "msums....");

  /* unsigned short */
  for (i=0; i<samples; i++) {
    if (i==0) {
      /* test extent */
#ifdef __SPU__
      a.vui = b.vui = c.vui = spu_splats((unsigned int)0xFFFFFFFF);
#else /* !__SPU__ */
      a.vui = b.vui = c.vui = ((vector unsigned int) {0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF});
#endif /* __SPU__ */
    } else {
      a.vui = rand_vui();
      b.vui = rand_vui();
      c.vui = rand_vui();
    }
    d.vui = vec_msums(a.vus, b.vus, c.vui);

    for (j=0; j<4; j++) {
      unsigned int exp, prod0, prod1, sum;
      prod0 = (unsigned int)a.us[2*j+0] * (unsigned int)b.us[2*j+0];
      prod1 = (unsigned int)a.us[2*j+1] * (unsigned int)b.us[2*j+1];

      sum = prod0 + prod1;
      if (((prod0 & prod1) | ((prod0 ^ prod1) & ~sum)) & 0x80000000) {
	sum = 0xFFFFFFFF;
      }
      exp = sum + c.ui[j];
      if (((sum & c.ui[j]) | ((sum ^ c.ui[j]) & ~exp)) & 0x80000000) {
	exp = 0xFFFFFFFF;
      }

      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort: element=%d ", j);
	DEBUG(2, "a=%d %d b=%d %d ", a.us[2*j+0], a.us[2*j+1], b.us[2*j+0], a.us[2*j+1]);
	DEBUG(2, "c=%d d=%d exp=%d\n",  c.ui[j], d.ui[j], exp);
	return (1);
      }
    }
  }

  /* signed short */
  for (i=0; i<samples; i++) {
    if (i==0) {
      /* test extent */
#ifdef __SPU__
      a.vui = b.vui = c.vui = spu_splats((unsigned int)0x80000000);
#else /* !__SPU__ */
      a.vui = b.vui = c.vui = ((vector unsigned int) {0x80000000,0x80000000,0x80000000,0x80000000});
#endif /* __SPU__ */
    } else if (i==1) {
      /* test extent */
#ifdef __SPU__
      a.vui = b.vui = c.vui = spu_splats((unsigned int)0x7FFFFFFF);
#else /* !__SPU__ */
      a.vui = b.vui = c.vui = ((vector unsigned int) {0x7FFFFFFF,0x7FFFFFFF,0x7FFFFFFF,0x7FFFFFFF});
#endif /* __SPU__ */
    } else {
      a.vui = rand_vui();
      b.vui = rand_vui();
      c.vui = rand_vui();
    }
    d.vi = vec_msums(a.vs, b.vs, c.vi);
    for (j=0; j<4; j++) {
      signed int exp, prod0, prod1, sum;
      prod0 = (signed int)a.s[2*j+0] * (signed int)b.s[2*j+0];
      prod1 = (signed int)a.s[2*j+1] * (signed int)b.s[2*j+1];

      sum = prod0 + prod1;
      exp = sum + c.i[j];
      if ((sum < 0) && (c.i[j] < 0) && (exp >= 0)) exp = 0x80000000;
      if ((sum > 0) && (c.i[j] > 0) && (exp <  0)) exp = 0x7FFFFFFF;

      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d ", j);
	DEBUG(2, "a=%d %d b=%d %d ", a.s[2*j+0], a.s[2*j+1], b.s[2*j+0], a.s[2*j+1]);
	DEBUG(2, "c=%d d=%d exp=%d\n",  c.i[j], d.i[j], exp);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}

int test_vec_mtvscr()
{
  DEBUG(1, "mfvscr...");
  a.vus = vec_mfvscr();
  vec_mtvscr(a.vuc);
  vec_mtvscr(a.vc);
  vec_mtvscr(a.vbc);
  vec_mtvscr(a.vus);
  vec_mtvscr(a.vs);
  vec_mtvscr(a.vbs);
  vec_mtvscr(a.vp);
  vec_mtvscr(a.vui);
  vec_mtvscr(a.vi);
  vec_mtvscr(a.vbi);
  DEBUG(1, passed);
  return (0);
}

int test_vec_mule()
{
  int i, j;

  DEBUG(1, "mule.....");

  /* unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_mule(a.vuc, b.vuc);
    for (j=0; j<8; j++) {
      unsigned short exp;

      exp = (unsigned short)(a.uc[2*j]) * (unsigned short)(b.uc[2*j]);
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n",  a.uc[2*j], b.uc[2*j], d.us[j], exp);
	return (1);
      }
    }
  }

  /* signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vs = vec_mule(a.vc, b.vc);
    for (j=0; j<8; j++) {
      signed short exp;

      exp = (signed short)(a.c[2*j]) * (signed short)(b.c[2*j]);
      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n",  a.c[2*j], b.c[2*j], d.s[j], exp);
	return (1);
      }
    }
  }

  /* unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_mule(a.vus, b.vus);
    for (j=0; j<4; j++) {
      unsigned int exp;

      exp = (unsigned int)(a.us[2*j]) * (unsigned int)(b.us[2*j]);
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n",  a.us[2*j], b.us[2*j], d.ui[j], exp);
	return (1);
      }
    }
  }

  /* signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vi = vec_mule(a.vs, b.vs);
    for (j=0; j<4; j++) {
      signed int exp;

      exp = (signed int)(a.s[2*j]) * (signed int)(b.s[2*j]);
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n",  a.s[2*j], b.s[2*j], d.i[j], exp);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_mulo()
{
  int i, j;

  DEBUG(1, "mulo.....");

  /* unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_mulo(a.vuc, b.vuc);
    for (j=0; j<8; j++) {
      unsigned short exp;

      exp = (unsigned short)(a.uc[2*j+1]) * (unsigned short)(b.uc[2*j+1]);
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n",  a.uc[2*j+1], b.uc[2*j+1], d.us[j], exp);
	return (1);
      }
    }
  }

  /* signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vs = vec_mulo(a.vc, b.vc);
    for (j=0; j<8; j++) {
      signed short exp;

      exp = (signed short)(a.c[2*j+1]) * (signed short)(b.c[2*j+1]);
      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n",  a.c[2*j+1], b.c[2*j+1], d.s[j], exp);
	return (1);
      }
    }
  }

  /* unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_mulo(a.vus, b.vus);
    for (j=0; j<4; j++) {
      unsigned int exp;

      exp = (unsigned int)(a.us[2*j+1]) * (unsigned int)(b.us[2*j+1]);
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n",  a.us[2*j+1], b.us[2*j+1], d.ui[j], exp);
	return (1);
      }
    }
  }

  /* signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vi = vec_mulo(a.vs, b.vs);
    for (j=0; j<4; j++) {
      signed int exp;

      exp = (signed int)(a.s[2*j+1]) * (signed int)(b.s[2*j+1]);
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n",  a.s[2*j+1], b.s[2*j+1], d.i[j], exp);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}

int test_vec_nmsub()
{
  int i, j;

  DEBUG(1, "nmsub....");

  /* float */
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-1.0f, 1.0f);
    b.vf = rand_vf(-1.0f, 1.0f);
    c.vf = rand_vf(-1.0f, 1.0f);
    d.vf = vec_nmsub(a.vf, b.vf, c.vf);
    for (j=0; j<4; j++) {
      float exp;
      float delta;
      exp = c.f[j] - (a.f[j] * b.f[j]);
      delta = exp - d.f[j];
      if (delta < 0.0f) delta = -delta;
      if (delta > 0.00001f) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d a=%f ", j, a.f[j]);
	DEBUG(2, "b=%f c=%f d=%f exp=%f\n", b.f[j], c.f[j], d.f[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_nor()
{
  int i, j;

  DEBUG(1, "nor......");

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_nor(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = ~(a.ui[j] | b.ui[j]);
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint: element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vuc = vec_nor(a.vuc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_nor(a.vbc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_nor(a.vc, b.vc);
    VERIFY_EQUAL(d, e);

    e.vus = vec_nor(a.vus, b.vus);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_nor(a.vbs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_nor(a.vs, b.vs);
    VERIFY_EQUAL(d, e);

    e.vbi = vec_nor(a.vbi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_nor(a.vi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vf = vec_nor(a.vf, b.vf);
    VERIFY_EQUAL(d, e);
  }

  DEBUG(1, passed);
  return (0);
}

int test_vec_or()
{
  int i, j;

  DEBUG(1, "or.......");

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_or(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = a.ui[j] | b.ui[j];
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint | uint : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vuc = vec_or(a.vuc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_or(a.vuc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_or(a.vbc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_or(a.vbc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_or(a.vc, b.vc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_or(a.vc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_or(a.vbc, b.vc);
    VERIFY_EQUAL(d, e);

    e.vus = vec_or(a.vus, b.vus);
    VERIFY_EQUAL(d, e);
    e.vus = vec_or(a.vbs, b.vus);
    VERIFY_EQUAL(d, e);
    e.vus = vec_or(a.vus, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_or(a.vbs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_or(a.vs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_or(a.vbs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_or(a.vs, b.vbs);
    VERIFY_EQUAL(d, e);

    e.vui = vec_or(a.vbi, b.vui);
    VERIFY_EQUAL(d, e);
    e.vui = vec_or(a.vui, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_or(a.vbi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_or(a.vi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_or(a.vbi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_or(a.vi, b.vbi);
    VERIFY_EQUAL(d, e);

    e.vf = vec_or(a.vf, b.vf);
    VERIFY_EQUAL(d, e);
    e.vf = vec_or(a.vbi, b.vf);
    VERIFY_EQUAL(d, e);
    e.vf = vec_or(a.vf, b.vbi);
    VERIFY_EQUAL(d, e);
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_pack()
{
  int i, j;

  DEBUG(1, "pack.....");

  /* ushort */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_pack(a.vus, b.vus);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = (unsigned char)((j<8) ? a.us[j] : b.us[j-8]);
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vc = vec_pack(a.vs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_pack(a.vbs, b.vbs);
    VERIFY_EQUAL(d, e);
  }

  /* uint */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_pack(a.vui, b.vui);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = (unsigned short)((j<4) ? a.ui[j] : b.ui[j-4]);
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.us[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vs = vec_pack(a.vi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_pack(a.vbi, b.vbi);
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_packpx()
{
  int i, j;

  DEBUG(1, "packpx...");

  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vp = vec_packpx(a.vui, b.vui);
    for (j=0; j<8; j++) {
      signed short exp;
      unsigned int word;
      word = (j<4) ? a.ui[j] : b.ui[j-4];
      exp = (signed short)(((word & 0x01000000) >> 9) | ((word & 0x00F80000) >> 9) |
			   ((word & 0x0000F800) >> 6) | ((word & 0x000000F8) >> 3));
      if (exp != d.p[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "pixel=0x%x d=0x%x exp=0x%x\n", word, d.p[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_packs()
{
  int i, j;

  DEBUG(1, "packs....");

  /* ushort */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_packs(a.vus, b.vus);
    for (j=0; j<16; j++) {
      unsigned char exp;
      unsigned short half;

      half = (j<8) ? a.us[j] : b.us[j-8];
      exp = (unsigned char)((half > 0xFF) ? 0xFF : half);
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[j], exp);
	return (1);
      }
    }
  }

  /* short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vc = vec_packs(a.vs, b.vs);
    for (j=0; j<16; j++) {
      signed char exp;
      signed short half;

      half = (j<8) ? a.s[j] : b.s[j-8];
      if (half > 127)       exp =  127;
      else if (half < -128) exp = -128;
      else                  exp = (signed char)(half);

      if (exp != d.c[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d ", j);
	DEBUG(2, "d=%d exp=%d\n", d.c[j], exp);
	return (1);
      }
    }
  }

  /* uint */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_packs(a.vui, b.vui);
    for (j=0; j<8; j++) {
      unsigned short exp;
      unsigned int word;

      word = (j<4) ? a.ui[j] : b.ui[j-4];
      exp = (unsigned short)((word > 0xFFFF) ? 0xFFFF : word);
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.us[j], exp);
	return (1);
      }
    }
  }

  /* int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vs = vec_packs(a.vi, b.vi);
    for (j=0; j<8; j++) {
      signed short exp;
      signed int word;

      word = (j<4) ? a.i[j] : b.i[j-4];
      if (word > 32767)       exp =  32767;
      else if (word < -32768) exp = -32768;
      else                    exp = (signed short)(word);

      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint: element=%d ", j);
	DEBUG(2, "d=%d exp=%d\n", d.s[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_packsu()
{
  int i, j;

  DEBUG(1, "packsu...");

  /* ushort */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_packsu(a.vus, b.vus);
    for (j=0; j<16; j++) {
      unsigned char exp;
      unsigned short half;

      half = (j<8) ? a.us[j] : b.us[j-8];
      exp = (unsigned char)((half > 0xFF) ? 0xFF : half);
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[j], exp);
	return (1);
      }
    }
  }

  /* short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_packsu(a.vs, b.vs);
    for (j=0; j<16; j++) {
      signed char exp;
      signed short half;

      half = (j<8) ? a.s[j] : b.s[j-8];
      if (half > 255) exp = 255;
      else if (half < 0) exp = 0;
      else exp = (unsigned char)(half);

      if (exp != d.c[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d ", j);
	DEBUG(2, "d=%d exp=%d\n", d.uc[j], exp);
	return (1);
      }
    }
  }

  /* uint */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_packsu(a.vui, b.vui);
    for (j=0; j<8; j++) {
      unsigned short exp;
      unsigned int word;

      word = (j<4) ? a.ui[j] : b.ui[j-4];
      exp = (unsigned short)((word > 0xFFFF) ? 0xFFFF : word);
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.us[j], exp);
	return (1);
      }
    }
  }

  /* int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_packsu(a.vi, b.vi);
    for (j=0; j<8; j++) {
      unsigned short exp;
      signed int word;

      word = (j<4) ? a.i[j] : b.i[j-4];
      if (word > 65535) exp = 65535;
      else if (word < 0) exp = 0;
      else exp = (unsigned short)(word);

      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint: element=%d ", j);
	DEBUG(2, "in=%d d=%d exp=%d\n", word, d.us[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_perm()
{
  int i, j;

  DEBUG(1, "perm.....");

  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    c.vui = rand_vui();
    d.vuc = vec_perm(a.vuc, b.vuc, c.vuc);
    for (j=0; j<16; j++) {
      unsigned int index;
      unsigned char exp;

      index = c.uc[j];
      if (index & 0x10) {
	exp = b.uc[index & 0xF];
      } else {
	exp = a.uc[index & 0xF];
      }
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vc = vec_perm(a.vc, b.vc, c.vuc);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_perm(a.vbc, b.vbc, c.vuc);
    VERIFY_EQUAL(d, e);
    e.vus = vec_perm(a.vus, b.vus, c.vuc);
    VERIFY_EQUAL(d, e);
    e.vs = vec_perm(a.vs, b.vs, c.vuc);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_perm(a.vbs, b.vbs, c.vuc);
    VERIFY_EQUAL(d, e);
    e.vp = vec_perm(a.vp, b.vp, c.vuc);
    VERIFY_EQUAL(d, e);
    e.vui = vec_perm(a.vui, b.vui, c.vuc);
    VERIFY_EQUAL(d, e);
    e.vi = vec_perm(a.vi, b.vi, c.vuc);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_perm(a.vbi, b.vbi, c.vuc);
    VERIFY_EQUAL(d, e);
    e.vf = vec_perm(a.vf, b.vf, c.vuc);
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_re()
{
  int i, j;

  DEBUG(1, "re.......");

  for (i=0; i<samples; i++) {
    a.vf = rand_vf(1.0f, 3.0f);
    d.vf = vec_re(a.vf);
    for (j=0; j<4; j++) {
      float err;
      err = a.f[j] * d.f[j];
      if ((err > 1.00025f) || (err < 0.99975)) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%f d=%f err=%f\n", a.f[j], d.f[j], err);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_rl()
{
  int i, j;

  DEBUG(1, "rl.......");

  /* byte */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_rl(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      int rotate;
      unsigned char exp;

      rotate = (int)(b.uc[j] & 0x7);
      exp = (a.uc[j] << rotate) | (a.uc[j] >> (8-rotate));
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tbyte element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.uc[j], b.uc[j], d.uc[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vc = vec_rl(a.vc, b.vuc);
    VERIFY_EQUAL(d, e);
  }

  /* halfword */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_rl(a.vus, b.vus);
    for (j=0; j<8; j++) {
      int rotate;
      unsigned short exp;

      rotate = (int)(b.us[j] & 0xF);
      exp = (a.us[j] << rotate) | (a.us[j] >> (16-rotate));
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\thalfword element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.us[j], b.us[j], d.us[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vs = vec_rl(a.vs, b.vus);
    VERIFY_EQUAL(d, e);
  }

  /* word */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_rl(a.vui, b.vui);
    for (j=0; j<4; j++) {
      int rotate;
      unsigned int exp;

      rotate = (int)(b.ui[j] & 0x1F);
      exp = (a.ui[j] << rotate) | (a.ui[j] >> (32-rotate));
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tword element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vi = vec_rl(a.vi, b.vui);
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}

static struct _round_trunc {
  float in;
  float out_round;
  float out_trunc;
} round_trunc_samples[] = {
  { -3.0f, -3.0f, -3.0f },
  { -1.999f, -2.0f, -1.0f },
  { -2.5f, -2.0f, -2.0f },
  { 2.5f, 2.0f, 2.0f }, 

  { -0.001f, 0.0f, 0.0f },
  { 73.49f, 73.0f, 73.0f },
  { 73.51f, 74.0f, 73.0f },
  { 73.5f,  74.0f, 73.0f },

  { -13.5f, -14.0f, -13.0f },
  { 0.999f, 1.0f, 0.0f },
  { 123456792.4f, 123456792.0f, 123456792.0f },
  { -23.51f, -24.0f, -23.0f },

  { -0.5f, 0.0f, 0.0f },
  { 0.5f, 0.0f, 0.0f },
  { 4194310.5f, 4194310.0f, 4194310.0f },
  { 8386607.5f, 8386608.0f, 8386607.0f },
};


int test_vec_round()
{
  int i, j;

  DEBUG(1, "round....");

  for (i=0; i<(int)(sizeof(round_trunc_samples)/(4*sizeof(struct _round_trunc))); i++) {
    a.f[0] = round_trunc_samples[4*i+0].in;
    a.f[1] = round_trunc_samples[4*i+1].in;
    a.f[2] = round_trunc_samples[4*i+2].in;
    a.f[3] = round_trunc_samples[4*i+3].in;
    d.vf = vec_round(a.vf);

    for (j=0; j<4; j++) {
      float exp;

      exp = round_trunc_samples[4*i+j].out_round;
      if (exp != d.f[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%f d=%f exp=%f\n", a.f[j], d.f[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_rsqrte()
{
  int i, j;

  DEBUG(1, "rsqrte...");

  for (i=0; i<samples; i++) {
    a.vf = rand_vf(1.0f, 3.0f);
    d.vf = vec_rsqrte(a.vf);
    for (j=0; j<4; j++) {
      float err;
      err = a.f[j] * d.f[j] * d.f[j];
      if ((err > 1.0004f) || (err < 0.9994)) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%f d=%f error=%f\n", a.f[j], d.f[j], err);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_sel()
{
  int i, j;

  DEBUG(1, "sel......");

  /* unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    c.vui = rand_vui();
    d.vui = vec_sel(a.vui, b.vui, c.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = (a.ui[j]& ~c.ui[j]) | (b.ui[j] & c.ui[j]);
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d a=0x%x ", j, a.ui[j]);
	DEBUG(2, "b=0x%x c=0x%x d=0x%x exp=0x%x\n", b.ui[j], c.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vuc = vec_sel(a.vuc, b.vuc, c.vuc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_sel(a.vuc, b.vuc, c.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_sel(a.vc, b.vc, c.vuc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_sel(a.vc, b.vc, c.vbc);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_sel(a.vbc, b.vbc, c.vuc);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_sel(a.vbc, b.vbc, c.vbc);
    VERIFY_EQUAL(d, e);

    e.vus = vec_sel(a.vus, b.vus, c.vus);
    VERIFY_EQUAL(d, e);
    e.vus = vec_sel(a.vus, b.vus, c.vbs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_sel(a.vs, b.vs, c.vus);
    VERIFY_EQUAL(d, e);
    e.vs = vec_sel(a.vs, b.vs, c.vbs);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_sel(a.vbs, b.vbs, c.vus);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_sel(a.vbs, b.vbs, c.vbs);
    VERIFY_EQUAL(d, e);

    e.vui = vec_sel(a.vui, b.vui, c.vbi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_sel(a.vi, b.vi, c.vui);
    VERIFY_EQUAL(d, e);
    e.vi = vec_sel(a.vi, b.vi, c.vbi);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_sel(a.vbi, b.vbi, c.vui);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_sel(a.vbi, b.vbi, c.vbi);
    VERIFY_EQUAL(d, e);
    e.vf = vec_sel(a.vf, b.vf, c.vui);
    VERIFY_EQUAL(d, e);
    e.vf = vec_sel(a.vf, b.vf, c.vbi);
    VERIFY_EQUAL(d, e);
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_sl()
{
  int i, j;

  DEBUG(1, "sl.......");

  /* byte */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_sl(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;

      exp = a.uc[j] << (b.uc[j] & 7);
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tbyte element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.uc[j], b.uc[j], d.uc[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vc = vec_sl(a.vc, b.vuc);
    VERIFY_EQUAL(d, e);
  }

  /* halfword */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_sl(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;

      exp = a.us[j] << (b.us[j] & 15);
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\thalfword element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.us[j], b.us[j], d.us[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vs = vec_sl(a.vs, b.vus);
    VERIFY_EQUAL(d, e);
  }

  /* word */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_sl(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;

      exp = a.ui[j] << (b.ui[j] & 31);
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tword element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vi = vec_sl(a.vi, b.vui);
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_sld()
{
  int i, j;

  DEBUG(1, "sld......");

  /* 0 */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_sld(a.vuc, b.vuc, 0);
    for (j=0; j<16; j++) {
      if (a.uc[j] != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshift 0 element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[j], a.uc[j]);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vc = vec_sld(a.vc, b.vc, 0);
    VERIFY_EQUAL(d, e);
    e.vus = vec_sld(a.vus, b.vus, 0);
    VERIFY_EQUAL(d, e);
    e.vs = vec_sld(a.vs, b.vs, 0);
    VERIFY_EQUAL(d, e);
    e.vp = vec_sld(a.vp, b.vp, 0);
    VERIFY_EQUAL(d, e);
    e.vui = vec_sld(a.vui, b.vui, 0);
    VERIFY_EQUAL(d, e);
    e.vi = vec_sld(a.vi, b.vi, 0);
    VERIFY_EQUAL(d, e);
    e.vf = vec_sld(a.vf, b.vf, 0);
    VERIFY_EQUAL(d, e);
  }

  /* 5 */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_sld(a.vuc, b.vuc, 5);
    for (j=0; j<16; j++) {
      unsigned char exp;

      exp = (j >= 11) ? b.uc[j-11] : a.uc[j+5];
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshift 5 element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vc = vec_sld(a.vc, b.vc, 5);
    VERIFY_EQUAL(d, e);
    e.vus = vec_sld(a.vus, b.vus, 5);
    VERIFY_EQUAL(d, e);
    e.vs = vec_sld(a.vs, b.vs, 5);
    VERIFY_EQUAL(d, e);
    e.vp = vec_sld(a.vp, b.vp, 5);
    VERIFY_EQUAL(d, e);
    e.vui = vec_sld(a.vui, b.vui, 5);
    VERIFY_EQUAL(d, e);
    e.vi = vec_sld(a.vi, b.vi, 5);
    VERIFY_EQUAL(d, e);
    e.vf = vec_sld(a.vf, b.vf, 5);
    VERIFY_EQUAL(d, e);
  }

  /* 15 */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_sld(a.vuc, b.vuc, 15);
    for (j=0; j<16; j++) {
      unsigned char exp;

      exp = (j >= 1) ? b.uc[j-1] : a.uc[15];
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshift 15 element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vc = vec_sld(a.vc, b.vc, 15);
    VERIFY_EQUAL(d, e);
    e.vus = vec_sld(a.vus, b.vus, 15);
    VERIFY_EQUAL(d, e);
    e.vs = vec_sld(a.vs, b.vs, 15);
    VERIFY_EQUAL(d, e);
    e.vp = vec_sld(a.vp, b.vp, 15);
    VERIFY_EQUAL(d, e);
    e.vui = vec_sld(a.vui, b.vui, 15);
    VERIFY_EQUAL(d, e);
    e.vi = vec_sld(a.vi, b.vi, 15);
    VERIFY_EQUAL(d, e);
    e.vf = vec_sld(a.vf, b.vf, 15);
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}



int test_vec_sll()
{
  int i, j;
  int shift;

  DEBUG(1, "sll......");

  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.uc[15] = (unsigned char)(rand_local());
    for (j=0; j<15; j++) b.uc[j] = b.uc[15];
    d.vuc = vec_sll(a.vuc, b.vuc);
    shift = b.uc[15] & 7;
    for (j=0; j<16; j++) {
      unsigned char exp;
      
      exp = a.uc[j] << shift;
      if (j<15) {
	exp |= a.uc[j+1] >> (8-shift);
      }
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "shift=%d d=0x%x exp=0x%x\n", shift, d.uc[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vuc = vec_sll(a.vuc, b.vus);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_sll(a.vuc, b.vui);
    VERIFY_EQUAL(d, e);

    e.vc = vec_sll(a.vc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_sll(a.vc, b.vus);
    VERIFY_EQUAL(d, e);
    e.vc = vec_sll(a.vc, b.vui);
    VERIFY_EQUAL(d, e);

    e.vbc = vec_sll(a.vbc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_sll(a.vbc, b.vus);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_sll(a.vbc, b.vui);
    VERIFY_EQUAL(d, e);

    e.vus = vec_sll(a.vus, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vus = vec_sll(a.vus, b.vus);
    VERIFY_EQUAL(d, e);
    e.vus = vec_sll(a.vus, b.vui);
    VERIFY_EQUAL(d, e);

    e.vs = vec_sll(a.vs, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vs = vec_sll(a.vs, b.vus);
    VERIFY_EQUAL(d, e);
    e.vs = vec_sll(a.vs, b.vui);
    VERIFY_EQUAL(d, e);

    e.vbs = vec_sll(a.vbs, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_sll(a.vbs, b.vus);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_sll(a.vbs, b.vui);
    VERIFY_EQUAL(d, e);

    e.vp = vec_sll(a.vp, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vp = vec_sll(a.vp, b.vus);
    VERIFY_EQUAL(d, e);
    e.vp = vec_sll(a.vp, b.vui);
    VERIFY_EQUAL(d, e);

    e.vui = vec_sll(a.vui, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vui = vec_sll(a.vui, b.vus);
    VERIFY_EQUAL(d, e);
    e.vui = vec_sll(a.vui, b.vui);
    VERIFY_EQUAL(d, e);

    e.vi = vec_sll(a.vi, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vi = vec_sll(a.vi, b.vus);
    VERIFY_EQUAL(d, e);
    e.vi = vec_sll(a.vi, b.vui);
    VERIFY_EQUAL(d, e);

    e.vbi = vec_sll(a.vbi, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_sll(a.vbi, b.vus);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_sll(a.vbi, b.vui);
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_slo()
{
  int i, j;
  int shift;

  DEBUG(1, "slo......");

  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_slo(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;

      shift = (b.uc[15] >> 3) & 0xF;
      exp = ((j+shift) > 15) ? 0 : a.uc[j+shift];
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "shift=%d d=0x%x exp=0x%x\n", shift, d.uc[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vuc = vec_slo(a.vuc, b.vc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_slo(a.vc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_slo(a.vc, b.vc);
    VERIFY_EQUAL(d, e);

    e.vus = vec_slo(a.vus, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vus = vec_slo(a.vus, b.vc);
    VERIFY_EQUAL(d, e);
    e.vs = vec_slo(a.vs, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vs = vec_slo(a.vs, b.vc);
    VERIFY_EQUAL(d, e);

    e.vp = vec_slo(a.vp, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vp = vec_slo(a.vp, b.vc);
    VERIFY_EQUAL(d, e);

    e.vui = vec_slo(a.vui, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vui = vec_slo(a.vui, b.vc);
    VERIFY_EQUAL(d, e);
    e.vi = vec_slo(a.vi, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vi = vec_slo(a.vi, b.vc);
    VERIFY_EQUAL(d, e);

    e.vf = vec_slo(a.vf, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vf = vec_slo(a.vf, b.vc);
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_splat()
{
  int i, j;

  DEBUG(1, "splat....");

  /* byte 13 */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    d.vuc = vec_splat(a.vuc, 13);
    for (j=0; j<16; j++) {
      if (a.uc[13] != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tsplat byte 13 element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[j], a.uc[13]);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vc = vec_splat(a.vc, 13);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_splat(a.vbc, 13);
    VERIFY_EQUAL(d, e);
  }

  /* halfword 5 */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    d.vus = vec_splat(a.vus, 5);
    for (j=0; j<8; j++) {
      if (a.us[5] != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tsplat halfword 5 element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.us[j], a.us[5]);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vs = vec_splat(a.vs, 5);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_splat(a.vbs, 5);
    VERIFY_EQUAL(d, e);
    e.vp = vec_splat(a.vp, 5);
    VERIFY_EQUAL(d, e);
  }

  /* word 1 */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    d.vui = vec_splat(a.vui, 1);
    for (j=0; j<4; j++) {
      if (a.ui[1] != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tsplat word 1 element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.ui[j], a.ui[j]);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vi = vec_splat(a.vi, 1);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_splat(a.vbi, 1);
    VERIFY_EQUAL(d, e);
    e.vf = vec_splat(a.vf, 1);
    VERIFY_EQUAL(d, e);
  }

  DEBUG(1, passed);
  return (0);
}

int test_vec_splat_s8()
{
  int j;

  DEBUG(1, "splats8..");

  /* 3 */
  d.vc = vec_splat_s8(3);
  for (j=0; j<16; j++) {
    if (d.uc[j] != 3) {
      DEBUG(1, failed);
      DEBUG(2, "\telement=%d ", j);
      DEBUG(2, "d=0x%x exp=0x3\n", d.uc[j]);
      return (1);
    }
  }

  /* 18 */
  d.vc = vec_splat_s8(-14);
  for (j=0; j<16; j++) {
    if (d.uc[j] != 0xF2) {
      DEBUG(1, failed);
      DEBUG(2, "\telement=%d ", j);
      DEBUG(2, "d=0x%x exp=0xF2\n", d.uc[j]);
      return (1);
    }
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_splat_s16()
{
  int j;

  DEBUG(1, "splats16.");

  /* 5 */
  d.vs = vec_splat_s16(5);
  for (j=0; j<8; j++) {
    if (d.us[j] != 5) {
      DEBUG(1, failed);
      DEBUG(2, "\telement=%d ", j);
      DEBUG(2, "d=0x%x exp=0x5\n", d.us[j]);
      return (1);
    }
  }

  /* 19 */
  d.vs = vec_splat_s16(-13);
  for (j=0; j<8; j++) {
    if (d.us[j] != 0xFFF3) {
      DEBUG(1, failed);
      DEBUG(2, "\telement=%d ", j);
      DEBUG(2, "d=0x%x exp=0xFFF3\n", d.us[j]);
      return (1);
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_splat_s32()
{
  int j;

  DEBUG(1, "splats32.");

  /* 1 */
  d.vi = vec_splat_s32(1);
  for (j=0; j<4; j++) {
    if (d.ui[j] != 1) {
      DEBUG(1, failed);
      DEBUG(2, "\telement=%d ", j);
      DEBUG(2, "d=0x%x exp=0x1\n", d.ui[j]);
      return (1);
    }
  }

  /* 30 */
  d.vi = vec_splat_s32(-2);
  for (j=0; j<4; j++) {
    if (d.ui[j] != 0xFFFFFFFE) {
      DEBUG(1, failed);
      DEBUG(2, "\telement=%d ", j);
      DEBUG(2, "d=0x%x exp=0xFFFFFFFE\n", d.ui[j]);
      return (1);
    }
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_splat_u8()
{
  int j;

  DEBUG(1, "splatu8..");

  /* -5 */
  d.vuc = vec_splat_u8(-5);
  for (j=0; j<16; j++) {
    if (d.uc[j] != 0xFB) {
      DEBUG(1, failed);
      DEBUG(2, "\telement=%d ", j);
      DEBUG(2, "d=0x%x exp=0xFB\n", d.uc[j]);
      return (1);
    }
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_splat_u16()
{
  int j;

  DEBUG(1, "splatu16.");

  /* -6 */
  d.vus = vec_splat_u16(-6);
  for (j=0; j<8; j++) {
    if (d.us[j] != 0xFFFA) {
      DEBUG(1, failed);
      DEBUG(2, "\telement=%d ", j);
      DEBUG(2, "d=0x%x exp=0xFFFA\n", d.us[j]);
      return (1);
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_splat_u32()
{
  int j;

  DEBUG(1, "splatu32.");

  /* -3 */
  d.vui = vec_splat_u32(-3);
  for (j=0; j<4; j++) {
    if (d.ui[j] != 0xFFFFFFFD) {
      DEBUG(1, failed);
      DEBUG(2, "\telement=%d ", j);
      DEBUG(2, "d=0x%x exp=0xFFFFFFFD\n", d.ui[j]);
      return (1);
    }
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_sr()
{
  int i, j;

  DEBUG(1, "sr.......");

  /* byte */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_sr(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;

      exp = a.uc[j] >> (b.uc[j] & 7);
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tbyte element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.uc[j], b.uc[j], d.uc[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vc = vec_sr(a.vc, b.vuc);
    VERIFY_EQUAL(d, e);
  }

  /* halfword */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_sr(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;

      exp = a.us[j] >> (b.us[j] & 15);
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\thalfword element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.us[j], b.us[j], d.us[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vs = vec_sr(a.vs, b.vus);
    VERIFY_EQUAL(d, e);
  }

  /* word */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_sr(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;

      exp = a.ui[j] >> (b.ui[j] & 31);
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tword element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vi = vec_sr(a.vi, b.vui);
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_sra()
{
  int i, j;

  DEBUG(1, "sra......");

  /* byte */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vc = vec_sra(a.vc, b.vuc);
    for (j=0; j<16; j++) {
      signed char exp;

      exp = a.c[j] >> (b.uc[j] & 7);
      if (exp != d.c[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tbyte element=%d ", j);
	DEBUG(2, "a=%d b=0x%x d=%d exp=%d\n", a.c[j], b.uc[j], d.c[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vuc = vec_sra(a.vuc, b.vuc);
    VERIFY_EQUAL(d, e);
  }

  /* halfword */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vs = vec_sra(a.vs, b.vus);
    for (j=0; j<8; j++) {
      signed short exp;

      exp = a.s[j] >> (b.us[j] & 15);
      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\thalfword element=%d ", j);
	DEBUG(2, "a=%d b=0x%x d=%d exp=%d\n", a.s[j], b.us[j], d.s[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vus = vec_sra(a.vus, b.vus);
    VERIFY_EQUAL(d, e);
  }

  /* word */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vi = vec_sra(a.vi, b.vui);
    for (j=0; j<4; j++) {
      signed int exp;

      exp = a.i[j] >> (b.ui[j] & 31);
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tword element=%d ", j);
	DEBUG(2, "a=%d b=0x%x d=%d exp=0%d\n", a.i[j], b.ui[j], d.i[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vui = vec_sra(a.vui, b.vui);
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_srl()
{
  int i, j;
  int shift;

  DEBUG(1, "srl......");

  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.uc[15] = (unsigned char)(rand_local());
    for (j=0; j<15; j++) b.uc[j] = b.uc[15];
    d.vuc = vec_sll(a.vuc, b.vuc);
    shift = b.uc[15] & 7;
    d.vuc = vec_srl(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      
      exp = a.uc[j] >> shift;
      if (j>0) {
	exp |= a.uc[j-1] << (8-shift);
      }

      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "shift=%d d=0x%x exp=0x%x\n", shift, d.uc[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vuc = vec_srl(a.vuc, b.vus);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_srl(a.vuc, b.vui);
    VERIFY_EQUAL(d, e);

    e.vc = vec_srl(a.vc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_srl(a.vc, b.vus);
    VERIFY_EQUAL(d, e);
    e.vc = vec_srl(a.vc, b.vui);
    VERIFY_EQUAL(d, e);

    e.vbc = vec_srl(a.vbc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_srl(a.vbc, b.vus);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_srl(a.vbc, b.vui);
    VERIFY_EQUAL(d, e);

    e.vus = vec_srl(a.vus, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vus = vec_srl(a.vus, b.vus);
    VERIFY_EQUAL(d, e);
    e.vus = vec_srl(a.vus, b.vui);
    VERIFY_EQUAL(d, e);

    e.vs = vec_srl(a.vs, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vs = vec_srl(a.vs, b.vus);
    VERIFY_EQUAL(d, e);
    e.vs = vec_srl(a.vs, b.vui);
    VERIFY_EQUAL(d, e);

    e.vbs = vec_srl(a.vbs, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_srl(a.vbs, b.vus);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_srl(a.vbs, b.vui);
    VERIFY_EQUAL(d, e);

    e.vp = vec_srl(a.vp, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vp = vec_srl(a.vp, b.vus);
    VERIFY_EQUAL(d, e);
    e.vp = vec_srl(a.vp, b.vui);
    VERIFY_EQUAL(d, e);

    e.vui = vec_srl(a.vui, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vui = vec_srl(a.vui, b.vus);
    VERIFY_EQUAL(d, e);
    e.vui = vec_srl(a.vui, b.vui);
    VERIFY_EQUAL(d, e);

    e.vi = vec_srl(a.vi, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vi = vec_srl(a.vi, b.vus);
    VERIFY_EQUAL(d, e);
    e.vi = vec_srl(a.vi, b.vui);
    VERIFY_EQUAL(d, e);

    e.vbi = vec_srl(a.vbi, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_srl(a.vbi, b.vus);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_srl(a.vbi, b.vui);
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_sro()
{
  int i, j;
  int shift;

  DEBUG(1, "sro......");

  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_sro(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;

      shift = (b.uc[15] >> 3) & 0xF;
      exp = (j < shift) ? 0 : a.uc[j-shift];
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "shift=%d d=0x%x exp=0x%x\n", shift, d.uc[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vuc = vec_sro(a.vuc, b.vc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_sro(a.vc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_sro(a.vc, b.vc);
    VERIFY_EQUAL(d, e);

    e.vus = vec_sro(a.vus, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vus = vec_sro(a.vus, b.vc);
    VERIFY_EQUAL(d, e);
    e.vs = vec_sro(a.vs, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vs = vec_sro(a.vs, b.vc);
    VERIFY_EQUAL(d, e);

    e.vp = vec_sro(a.vp, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vp = vec_sro(a.vp, b.vc);
    VERIFY_EQUAL(d, e);

    e.vui = vec_sro(a.vui, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vui = vec_sro(a.vui, b.vc);
    VERIFY_EQUAL(d, e);
    e.vi = vec_sro(a.vi, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vi = vec_sro(a.vi, b.vc);
    VERIFY_EQUAL(d, e);

    e.vf = vec_sro(a.vf, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vf = vec_sro(a.vf, b.vc);
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_st()
{
  int i;
  int base, offset;
  int index;

  DEBUG(1, "st.......");

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);
    index = ((unsigned)base + (unsigned)offset) & ~0xF;

    a.vui  = rand_vui();
    vec_st(a.vuc, offset, (vec_uchar16 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vuc, offset, (unsigned char *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vc, offset, (vec_char16 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vc, offset, (signed char *)(&data[base])); 
    VERIFY_DATA(a, index);

    a.vui  = rand_vui();
    vec_st(a.vbc, offset, (vec_bchar16 *)(&data[base])); 
    VERIFY_DATA(a, index);
#ifdef __SPU__
    a.vui  = rand_vui();
    vec_st(a.vbc, offset, (unsigned char *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vbc, offset, (signed char *)(&data[base])); 
    VERIFY_DATA(a, index);
#endif

    a.vui  = rand_vui();
    vec_st(a.vus, offset, (vec_ushort8 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vus, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vs, offset, (vec_short8 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vs, offset, (signed short *)(&data[base])); 
    VERIFY_DATA(a, index);

    a.vui  = rand_vui();
    vec_st(a.vbs, offset, (vec_bshort8 *)(&data[base])); 
    VERIFY_DATA(a, index);
#ifdef __SPU__
    a.vui  = rand_vui();
    vec_st(a.vbs, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vbs, offset, (short *)(&data[base])); 
    VERIFY_DATA(a, index);
#endif

    a.vui  = rand_vui();
    vec_st(a.vp, offset, (vec_pixel8 *)(&data[base])); 
    VERIFY_DATA(a, index);
#ifdef __SPU__
    a.vui  = rand_vui();
    vec_st(a.vp, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vp, offset, (short *)(&data[base])); 
    VERIFY_DATA(a, index);
#endif

    a.vui  = rand_vui();
    vec_st(a.vui, offset, (vec_uint4 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vui, offset, (unsigned int *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vi, offset, (vec_int4 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vi, offset, (signed int *)(&data[base])); 
    VERIFY_DATA(a, index);

    a.vui  = rand_vui();
    vec_st(a.vbi, offset, (vec_bint4 *)(&data[base])); 
    VERIFY_DATA(a, index);
#ifdef __SPU__
    a.vui  = rand_vui();
    vec_st(a.vbi, offset, (unsigned int *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vbi, offset, (int *)(&data[base])); 
    VERIFY_DATA(a, index);
#endif

    a.vui  = rand_vui();
    vec_st(a.vf, offset, (vec_float4 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_st(a.vf, offset, (float *)(&data[base])); 
    VERIFY_DATA(a, index);
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_ste()
{
  int i;
  int base, offset;
  int index, element;

  DEBUG(1, "ste......");

  for (i=0; i<256; i++) data[i] = 0;

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    index = ((unsigned)base + (unsigned)offset);
    element = index & 0xF;
    index &= ~0xF;

    a.vui  = rand_vui();
    vec_ste(a.vuc, offset, (unsigned char *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 1);
    a.vui  = rand_vui();
    vec_ste(a.vc, offset, (signed char *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 1);
#ifdef __SPU__
    a.vui  = rand_vui();
    vec_ste(a.vbc, offset, (unsigned char *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 1);
    a.vui  = rand_vui();
    vec_ste(a.vbc, offset, (signed char *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 1);
#endif
  }

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    index = ((unsigned)base + (unsigned)offset);
    element = (index & 0xF) >> 1;
    index &= ~0xF;

    a.vui  = rand_vui();
    vec_ste(a.vus, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 2);
    a.vui  = rand_vui();
    vec_ste(a.vs, offset, (short *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 2);
#ifdef __SPU__
    a.vui  = rand_vui();
    vec_ste(a.vbs, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 2);
    a.vui  = rand_vui();
    vec_ste(a.vbs, offset, (short *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 2);

    a.vui  = rand_vui();
    vec_ste(a.vp, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 2);
    a.vui  = rand_vui();
    vec_ste(a.vp, offset, (short *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 2);
#endif
  }

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    index = ((unsigned)base + (unsigned)offset);
    element = (index & 0xF) >> 2;
    index &= ~0xF;

    a.vui  = rand_vui();
    vec_ste(a.vui, offset, (unsigned int *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 4);
    a.vui  = rand_vui();
    vec_ste(a.vui, offset, (unsigned int *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 4);
    a.vui  = rand_vui();
    vec_ste(a.vi, offset, (int *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 4);
    a.vui  = rand_vui();
    vec_ste(a.vi, offset, (int *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 4);

#ifdef __SPU__
    a.vui  = rand_vui();
    vec_ste(a.vbi, offset, (unsigned int *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 4);
    a.vui  = rand_vui();
    vec_ste(a.vbi, offset, (unsigned int *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 4);
    a.vui  = rand_vui();
    vec_ste(a.vbi, offset, (int *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 4);
    a.vui  = rand_vui();
    vec_ste(a.vbi, offset, (int *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 4);
#endif

    a.vui  = rand_vui();
    vec_ste(a.vf, offset, (float *)(&data[base])); 
    VERIFY_DATA_ELEMENT(a, index, element, 4);
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_stl()
{
  int i;
  int base, offset;
  int index;

  DEBUG(1, "stl......");

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);
    index = ((unsigned)base + (unsigned)offset) & ~0xF;

    a.vui  = rand_vui();
    vec_stl(a.vuc, offset, (vec_uchar16 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vuc, offset, (unsigned char *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vc, offset, (vec_char16 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vc, offset, (signed char *)(&data[base])); 
    VERIFY_DATA(a, index);

    a.vui  = rand_vui();
    vec_stl(a.vbc, offset, (vec_bchar16 *)(&data[base])); 
    VERIFY_DATA(a, index);
#ifdef __SPU__
    a.vui  = rand_vui();
    vec_stl(a.vbc, offset, (unsigned char *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vbc, offset, (signed char *)(&data[base])); 
    VERIFY_DATA(a, index);
#endif

    a.vui  = rand_vui();
    vec_stl(a.vus, offset, (vec_ushort8 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vus, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vs, offset, (vec_short8 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vs, offset, (signed short *)(&data[base])); 
    VERIFY_DATA(a, index);

    a.vui  = rand_vui();
    vec_stl(a.vbs, offset, (vec_bshort8 *)(&data[base])); 
    VERIFY_DATA(a, index);
#ifdef __SPU__
    a.vui  = rand_vui();
    vec_stl(a.vbs, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vbs, offset, (signed short *)(&data[base])); 
    VERIFY_DATA(a, index);
#endif

    a.vui  = rand_vui();
    vec_stl(a.vp, offset, (vec_pixel8 *)(&data[base])); 
    VERIFY_DATA(a, index);
#ifdef __SPU__
    a.vui  = rand_vui();
    vec_stl(a.vp, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vp, offset, (signed short *)(&data[base])); 
    VERIFY_DATA(a, index);
#endif

    a.vui  = rand_vui();
    vec_stl(a.vui, offset, (vec_uint4 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vui, offset, (unsigned int *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vi, offset, (vec_int4 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vi, offset, (signed int *)(&data[base])); 
    VERIFY_DATA(a, index);

    a.vui  = rand_vui();
    vec_stl(a.vbi, offset, (vec_bint4 *)(&data[base])); 
    VERIFY_DATA(a, index);
#ifdef __SPU__
    a.vui  = rand_vui();
    vec_stl(a.vbi, offset, (unsigned int *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vbi, offset, (signed int *)(&data[base])); 
    VERIFY_DATA(a, index);
#endif

    a.vui  = rand_vui();
    vec_stl(a.vf, offset, (vec_float4 *)(&data[base])); 
    VERIFY_DATA(a, index);
    a.vui  = rand_vui();
    vec_stl(a.vf, offset, (float *)(&data[base])); 
    VERIFY_DATA(a, index);
  }
  DEBUG(1, passed);
  return (0);
}



int test_vec_sub()
{
  int i, j;

  DEBUG(1, "sub......");

  /* unsigned char  unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_sub(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = a.uc[j] - b.uc[j];
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar - uchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.uc[j], b.uc[j], d.uc[j], exp);
	return (1);
      }
    }
    
    /* verify equivalent variations */
    e.vuc = vec_sub(a.vuc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_sub(a.vbc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_sub(a.vc, b.vc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_sub(a.vc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_sub(a.vbc, b.vc);
    VERIFY_EQUAL(d, e);
  }

  /* unsigned short  unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_sub(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = a.us[j] - b.us[j];
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort - ushort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.us[j], b.us[j], d.us[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vus = vec_sub(a.vus, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vus = vec_sub(a.vbs, b.vus);
    VERIFY_EQUAL(d, e);
    e.vs = vec_sub(a.vs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_sub(a.vs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_sub(a.vbs, b.vs);
    VERIFY_EQUAL(d, e);
  }


  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_sub(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = a.ui[j] - b.ui[j];
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint - uint: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vui = vec_sub(a.vui, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vui = vec_sub(a.vbi, b.vui);
    VERIFY_EQUAL(d, e);
    e.vi = vec_sub(a.vi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_sub(a.vi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_sub(a.vbi, b.vi);
    VERIFY_EQUAL(d, e);
  }

  /* float  float */
  for (i=0; i<samples; i++) {
    a.vf = rand_vf(-1.0f, 1.0f);
    b.vf = rand_vf(-1.0f, 1.0f);
    d.vf = vec_sub(a.vf, b.vf);
    for (j=0; j<4; j++) {
      float exp;
      exp = a.f[j] - b.f[j];
      if (exp != d.f[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tfloat - float: element=%d ", j);
	DEBUG(2, "a=%f b=%f d=%f exp=%f\n", a.f[j], b.f[j], d.f[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_subc()
{
  int i, j;

  DEBUG(1, "subc.....");

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_subc(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp =  (a.ui[j] < b.ui[j]) ? 0 : 1;
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint - uint: element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_subs()
{
  int i, j;

  DEBUG(1, "subs.....");

  /* unsigned char  unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vuc = vec_subs(a.vuc, b.vuc);
    for (j=0; j<16; j++) {
      unsigned char exp;
      exp = a.uc[j] - b.uc[j];
      if (b.uc[j] > a.uc[j]) exp = 0x0;
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar - uchar: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.uc[j], b.uc[j], d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vuc = vec_subs(a.vuc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_subs(a.vbc, b.vuc);
    VERIFY_EQUAL(d, e);
  }


  /* signed char  signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vc = vec_subs(a.vc, b.vc);
    for (j=0; j<16; j++) {
      signed char exp;
      exp = a.c[j] - b.c[j];
      if ((a.c[j] <  0) && (b.c[j] >= 0) && (exp >= 0)) exp = 0x80;
      if ((a.c[j] >= 0) && (b.c[j] <  0) && (exp <  0)) exp = 0x7F;
      if (exp != d.c[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar - char: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.c[j], b.c[j], d.c[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vc = vec_subs(a.vc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_subs(a.vbc, b.vc);
    VERIFY_EQUAL(d, e);
  }

  /* unsigned short  unsigned short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vus = vec_subs(a.vus, b.vus);
    for (j=0; j<8; j++) {
      unsigned short exp;
      exp = a.us[j] - b.us[j];
      if (b.us[j] > a.us[j]) exp = 0x0;
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort - ushort: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.us[j], b.us[j], d.us[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vus = vec_subs(a.vus, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vus = vec_subs(a.vbs, b.vus);
    VERIFY_EQUAL(d, e);
  }

  /* signed short  signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vs = vec_subs(a.vs, b.vs);
    for (j=0; j<8; j++) {
      signed short exp;
      exp = a.s[j] - b.s[j];
      if ((a.s[j] <  0) && (b.s[j] >= 0) && (exp >= 0)) exp = 0x8000;
      if ((a.s[j] >= 0) && (b.s[j] <  0) && (exp <  0)) exp = 0x7FFF;
      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort - short: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.s[j], b.s[j], d.s[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vs = vec_subs(a.vs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_subs(a.vbs, b.vs);
    VERIFY_EQUAL(d, e);
  }

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_subs(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = a.ui[j] - b.ui[j];
      if (b.ui[j] > a.ui[j]) exp = 0x0;
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint - uint: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vui = vec_subs(a.vui, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vui = vec_subs(a.vbi, b.vui);
    VERIFY_EQUAL(d, e);
  }

  /* signed int  signed int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vi = vec_subs(a.vi, b.vi);
    for (j=0; j<4; j++) {
      signed int exp;
      exp = a.i[j] - b.i[j];
      if ((a.i[j] <  0) && (b.i[j] >= 0) && (exp >= 0)) exp = 0x80000000;
      if ((a.i[j] >= 0) && (b.i[j] <  0) && (exp <  0)) exp = 0x7FFFFFFF;
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint - int: element=%d ", j);
	DEBUG(2, "a=%d b=%d d=%d exp=%d\n", a.i[j], b.i[j], d.i[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vi = vec_subs(a.vi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_subs(a.vbi, b.vi);
    VERIFY_EQUAL(d, e);
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_sum4s()
{
  int i, j;

  DEBUG(1, "sum4s....");

  /* unsigned char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_sum4s(a.vuc, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp, sum;
      sum = ((unsigned int)a.uc[4*j+0] + (unsigned int)a.uc[4*j+1] +
	     (unsigned int)a.uc[4*j+2] + (unsigned int)a.uc[4*j+3]);
      exp = sum + b.ui[j];
      if (((sum & b.ui[j]) | ((sum ^ b.ui[j]) & ~exp)) & 0x80000000) {
	exp = 0xFFFFFFFF;
      }
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar: element=%d ", j);
	DEBUG(2, "a=%d %d %d %d ", a.uc[4*j], a.uc[4*j+1], a.uc[4*j+2], a.uc[4*j+3]);
	DEBUG(2, "b=%d d=%d exp=%d\n", b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
  }

  /* signed char */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vi = vec_sum4s(a.vc, b.vi);
    for (j=0; j<4; j++) {
      signed int exp, sum;
      sum = ((signed int)a.c[4*j+0] + (signed int)a.c[4*j+1] +
	     (signed int)a.c[4*j+2] + (signed int)a.c[4*j+3]);
      exp = sum + b.ui[j];
      if ((sum <  0) && (b.i[j] <  0) && (exp >= 0)) exp = 0x80000000;
      if ((sum >= 0) && (b.i[j] >= 0) && (exp <  0)) exp = 0x7FFFFFFF;
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tchar: element=%d ", j);
	DEBUG(2, "a=%d %d %d %d ", a.c[4*j], a.c[4*j+1], a.c[4*j+2], a.c[4*j+3]);
	DEBUG(2, "b=%d d=%d exp=%d\n", b.i[j], d.i[j], exp);
	return (1);
      }
    }
  }

  /* signed short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vi = vec_sum4s(a.vs, b.vi);
    for (j=0; j<4; j++) {
      signed int exp, sum;
      sum = (signed int)a.s[2*j+0] + (signed int)a.s[2*j+1];
      exp = sum + b.ui[j];
      if ((sum <  0) && (b.i[j] <  0) && (exp >= 0)) exp = 0x80000000;
      if ((sum >= 0) && (b.i[j] >= 0) && (exp <  0)) exp = 0x7FFFFFFF;
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d ", j);
	DEBUG(2, "a=%d %d ", a.c[2*j], a.c[2*j+1]);
	DEBUG(2, "b=%d d=%d exp=%d\n", b.i[j], d.i[j], exp);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_sum2s()
{
  int i, j;

  DEBUG(1, "sum2s....");

  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vi = vec_sum2s(a.vi, b.vi);
    for (j=0; j<4; j++) {
      signed int exp;
      if (j & 1) {
	exp = sum3s(a.i[j], a.i[j-1], b.i[j]);
      } else {
	exp = 0;
      }
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	if (j & 1) DEBUG(2, "a=%d %d b=%d ", a.i[j], a.i[j-1], b.i[j]);
	DEBUG(2, "d=%d exp=%d\n", d.i[j], exp);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}

int test_vec_sums()
{
  int i, j;

  DEBUG(1, "sums.....");

  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vi = vec_sums(a.vi, b.vi);
    for (j=0; j<4; j++) {
      signed int exp;
      if (j == 3) {
	exp = sum5s(a.i[0], a.i[1], a.i[2], a.i[3], b.i[3]);
      } else {
	exp = 0;
      }
      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	if (j == 3) {
	  DEBUG(2, "a=%d %d %d %d ", a.i[0], a.i[1], a.i[2], a.i[3]);
	  DEBUG(2, "b=%d ", b.i[j]);
	}
	DEBUG(2, "d=%d exp=%d\n", d.i[j], exp);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_trunc()
{
  int i, j;

  DEBUG(1, "trunc....");

  for (i=0; i<(int)(sizeof(round_trunc_samples)/(4*sizeof(struct _round_trunc))); i++) {
    a.f[0] = round_trunc_samples[4*i+0].in;
    a.f[1] = round_trunc_samples[4*i+1].in;
    a.f[2] = round_trunc_samples[4*i+2].in;
    a.f[3] = round_trunc_samples[4*i+3].in;
    d.vf = vec_trunc(a.vf);

    for (j=0; j<4; j++) {
      float exp;

      exp = round_trunc_samples[4*i+j].out_trunc;
      if (exp != d.f[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%f d=%f exp=%f\n", a.f[j], d.f[j], exp);
	return (1);
      }
    }
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_unpackh()
{
  int i, j;

  DEBUG(1, "unpackh..");

  /* short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    d.vs = vec_unpackh(a.vc);
    for (j=0; j<8; j++) {
      signed short exp;

      exp = (signed short)(a.c[j]);

      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d ", j);
	DEBUG(2, "d=%d exp=%d\n", d.s[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vbs = vec_unpackh(a.vbc);
    VERIFY_EQUAL(d, e);
  }

  /* pixel */
#ifdef SUPPORT_UNPACK_PIXEL
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    d.vui = vec_unpackh(a.vp);
    for (j=0; j<4; j++) {
      unsigned int exp;

      exp = (((unsigned int)((signed int)(a.s[j])) & 0xFF000000) |
	     ((unsigned int)(a.us[j] & 0x7C00) << 6) |
	     ((unsigned int)(a.us[j] & 0x03E0) << 3) |
	     ((unsigned int)(a.us[j] & 0x001F)));

      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tpixel: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.ui[j], exp);
	return (1);
      }
    }
  }
#endif

  /* int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    d.vi = vec_unpackh(a.vs);
    for (j=0; j<4; j++) {
      signed int exp;

      exp = (signed int)(a.s[j]);

      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint: element=%d ", j);
	DEBUG(2, "d=%d exp=%d\n", d.i[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
#ifndef SUPPORT_UNPACK_PIXEL
    e.vbi = vec_unpackh(a.vbs);
    VERIFY_EQUAL(d, e);
#endif
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_unpackl()
{
  int i, j;

  DEBUG(1, "unpackl..");

  /* short */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    d.vs = vec_unpackl(a.vc);
    for (j=0; j<8; j++) {
      signed short exp;

      exp = (signed int)a.c[8+j];

      if (exp != d.s[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tshort: element=%d ", j);
	DEBUG(2, "d=%d exp=%d\n", d.s[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vbs = vec_unpackl(a.vbc);
    VERIFY_EQUAL(d, e);
  }

#ifdef SUPPORT_UNPACK_PIXEL
  /* pixel */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    d.vui = vec_unpackl(a.vp);
    for (j=0; j<4; j++) {
      unsigned int exp;

      exp = (((unsigned int)((signed int)(a.s[4+j])) & 0xFF000000) |
	     ((unsigned int)(a.us[4+j] & 0x7C00) << 6) |
	     ((unsigned int)(a.us[4+j] & 0x03E0) << 3) |
	     ((unsigned int)(a.us[4+j] & 0x001F)));

      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tpixel: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.ui[j], exp);
	return (1);
      }
    }
  }
#endif

  /* int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    d.vi = vec_unpackl(a.vs);
    for (j=0; j<4; j++) {
      signed int exp;

      exp = (signed int)(a.s[4+j]);

      if (exp != d.i[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tint: element=%d ", j);
	DEBUG(2, "d=%d exp=%d\n", d.i[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
#ifndef SUPPORT_UNPACK_PIXEL
    e.vbi = vec_unpackl(a.vbs);
    VERIFY_EQUAL(d, e);
#endif
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_xor()
{
  int i, j;

  DEBUG(1, "xor......");

  /* unsigned int  unsigned int */
  for (i=0; i<samples; i++) {
    a.vui = rand_vui();
    b.vui = rand_vui();
    d.vui = vec_xor(a.vui, b.vui);
    for (j=0; j<4; j++) {
      unsigned int exp;
      exp = a.ui[j] ^ b.ui[j];
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint ^ uint : element=%d ", j);
	DEBUG(2, "a=0x%x b=0x%x d=0x%x exp=0x%x\n", a.ui[j], b.ui[j], d.ui[j], exp);
	return (1);
      }
    }
    /* verify all equivalent variations */
    e.vuc = vec_xor(a.vuc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_xor(a.vuc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vuc = vec_xor(a.vbc, b.vuc);
    VERIFY_EQUAL(d, e);
    e.vbc = vec_xor(a.vbc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_xor(a.vc, b.vc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_xor(a.vc, b.vbc);
    VERIFY_EQUAL(d, e);
    e.vc = vec_xor(a.vbc, b.vc);
    VERIFY_EQUAL(d, e);

    e.vus = vec_xor(a.vus, b.vus);
    VERIFY_EQUAL(d, e);
    e.vus = vec_xor(a.vbs, b.vus);
    VERIFY_EQUAL(d, e);
    e.vus = vec_xor(a.vus, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vbs = vec_xor(a.vbs, b.vbs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_xor(a.vs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_xor(a.vbs, b.vs);
    VERIFY_EQUAL(d, e);
    e.vs = vec_xor(a.vs, b.vbs);
    VERIFY_EQUAL(d, e);

    e.vui = vec_xor(a.vbi, b.vui);
    VERIFY_EQUAL(d, e);
    e.vui = vec_xor(a.vui, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vbi = vec_xor(a.vbi, b.vbi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_xor(a.vi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_xor(a.vbi, b.vi);
    VERIFY_EQUAL(d, e);
    e.vi = vec_xor(a.vi, b.vbi);
    VERIFY_EQUAL(d, e);

    e.vf = vec_xor(a.vf, b.vf);
    VERIFY_EQUAL(d, e);
    e.vf = vec_xor(a.vbi, b.vf);
    VERIFY_EQUAL(d, e);
    e.vf = vec_xor(a.vf, b.vbi);
    VERIFY_EQUAL(d, e);
  }

  DEBUG(1, passed);
  return (0);
}


/* CBEA extended VMX test.
 */
#if 0  /* Commented out until compilers support has been added */
int test_vec_extract()
{
  int i, element;

  DEBUG(1, "extract..");
  
  /* uchar */
  for (i=0; i<samples; i++) {
    unsigned char scalar, exp;
    a.vui = rand_vui();
    element = rand_local();
    scalar = vec_extract(a.vuc, element);
    exp = a.uc[element & 15];
    if (scalar != exp) {
      DEBUG(1, failed);
      DEBUG(2, "\tuchar: element=%d ", element);
      DEBUG(2, "d=0x%x exp=0x%x\n", scalar, exp);
      return (1);
    }
  }

  /* char */
  for (i=0; i<samples; i++) {
    signed char scalar, exp;
    a.vui = rand_vui();
    element = rand_local();
    scalar = vec_extract(a.vc, element);
    exp = a.c[element & 15];
    if (scalar != exp) {
      DEBUG(1, failed);
      DEBUG(2, "\tchar: element=%d ", element);
      DEBUG(2, "d=%d exp=%d\n", scalar, exp);
      return (1);
    }
  }

  /* ushort */
  for (i=0; i<samples; i++) {
    unsigned short scalar, exp;
    a.vui = rand_vui();
    element = rand_local();
    scalar = vec_extract(a.vus, element);
    exp = a.us[element & 7];
    if (scalar != exp) {
      DEBUG(1, failed);
      DEBUG(2, "\tushort: element=%d ", element);
      DEBUG(2, "d=0x%x exp=0x%x\n", scalar, exp);
      return (1);
    }
  }

  /* short */
  for (i=0; i<samples; i++) {
    signed short scalar, exp;
    a.vui = rand_vui();
    element = rand_local();
    scalar = vec_extract(a.vs, element);
    exp = a.s[element & 7];
    if (scalar != exp) {
      DEBUG(1, failed);
      DEBUG(2, "\tshort: element=%d ", element);
      DEBUG(2, "d=%d exp=%d\n", scalar, exp);
      return (1);
    }
  }

  /* uint */
  for (i=0; i<samples; i++) {
    unsigned int scalar, exp;
    a.vui = rand_vui();
    element = rand_local();
    scalar = vec_extract(a.vui, element);
    exp = a.ui[element & 3];
    if (scalar != exp) {
      DEBUG(1, failed);
      DEBUG(2, "\tuint: element=%d ", element);
      DEBUG(2, "d=0x%x exp=0x%x\n", scalar, exp);
      return (1);
    }
  }

  /* int */
  for (i=0; i<samples; i++) {
    signed int scalar, exp;
    a.vui = rand_vui();
    element = rand_local();
    scalar = vec_extract(a.vi, element);
    exp = a.i[element & 3];
    if (scalar != exp) {
      DEBUG(1, failed);
      DEBUG(2, "\tint: element=%d ", element);
      DEBUG(2, "d=%d exp=%d\n", scalar, exp);
      return (1);
    }
  }

  /* float */
  for (i=0; i<samples; i++) {
    float scalar, exp;
    a.vf = rand_vf(-1.0f, 1.0f);
    element = rand_local();
    scalar = vec_extract(a.vf, element);
    exp = a.f[element & 3];
    if (scalar != exp) {
      DEBUG(1, failed);
      DEBUG(2, "\tfloat: element=%d ", element);
      DEBUG(2, "d=%f exp=%f\n", scalar, exp);
      return (1);
    }
  }


  DEBUG(1, passed);
  return (0);
}


int test_vec_insert()
{
  int i, j, element;

  DEBUG(1, "insert...");
  

  /* byte */
  for (i=0; i<samples; i++) {
    unsigned char a, exp;
    b.vui = rand_vui();
    a = (unsigned char)rand_local();
    element = (int)rand_local();
    d.vuc = vec_insert(a, b.vuc, element);
    for (j=0; j<16; j++) {
      exp = (j == (element & 15)) ? a : b.uc[j];
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar: element=%d ", element);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vc = vec_insert((signed char)(a), b.vc, element);
    VERIFY_EQUAL(d, e);
  }

  /* halfword */
  for (i=0; i<samples; i++) {
    unsigned short a, exp;
    b.vui = rand_vui();
    a = (unsigned short)rand_local();
    element = (int)rand_local();
    d.vus = vec_insert(a, b.vus, element);
    for (j=0; j<8; j++) {
      exp = (j == (element & 7)) ? a : b.us[j];
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort: element=%d ", element);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.us[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vs = vec_insert((signed short)(a), b.vs, element);
    VERIFY_EQUAL(d, e);
  }

  /* word */
  for (i=0; i<samples; i++) {
    unsigned int a, exp;
    b.vui = rand_vui();
    a = (unsigned int)rand_local();
    element = (int)rand_local();
    d.vui = vec_insert(a, b.vui, element);
    for (j=0; j<4; j++) {
      exp = (j == (element & 3)) ? a : b.ui[j];
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint: element=%d ", element);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.ui[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vi = vec_insert((signed int)(a), b.vi, element);
    VERIFY_EQUAL(d, e);
  }


  /* float */
  for (i=0; i<samples; i++) {
    float a, exp;
    b.vf = rand_vf(-4.0f, 4.0f);
    a = rand_float(-4.0f, 4.0f);
    element = (int)rand_local();
    d.vf = vec_insert(a, b.vf, element);
    for (j=0; j<4; j++) {
      exp = (j == (element & 3)) ? a : b.f[j];
      if (exp != d.f[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tfloat: element=%d ", element);
	DEBUG(2, "d=%f exp=%f\n", d.f[j], exp);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_promote()
{
  int i, element;

  DEBUG(1, "promote..");
  
  /* uchar */
  for (i=0; i<samples; i++) {
    unsigned char a, exp;
    a = (unsigned char)rand_local();
    element = (int)rand_local();
    d.vuc = vec_promote(a, element);
    exp = a;
    if (exp != d.uc[element & 15]) {
      DEBUG(1, failed);
      DEBUG(2, "\tuchar: element=0x%x ", element);
      DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[element & 15], exp);
      return (1);
    }
  }

  /* char */
  for (i=0; i<samples; i++) {
    signed char a, exp;
    a = (signed char)rand_local();
    element = (int)rand_local();
    d.vc = vec_promote(a, element);
    exp = a;
    if (exp != d.c[element & 15]) {
      DEBUG(1, failed);
      DEBUG(2, "\tchar: element=0x%x ", element);
      DEBUG(2, "d=0x%x exp=0x%x\n", d.c[element & 15], exp);
      return (1);
    }
  }

  /* ushort */
  for (i=0; i<samples; i++) {
    unsigned short a, exp;
    a = (unsigned short)rand_local();
    element = (int)rand_local();
    d.vus = vec_promote(a, element);
    exp = a;
    if (exp != d.us[element & 7]) {
      DEBUG(1, failed);
      DEBUG(2, "\tushort: element=0x%x ", element);
      DEBUG(2, "d=0x%x exp=0x%x\n", d.us[element & 7], exp);
      return (1);
    }
  }

  /* short */
  for (i=0; i<samples; i++) {
    signed short a, exp;
    a = (signed short)rand_local();
    element = (int)rand_local();
    d.vs = vec_promote(a, element);
    exp = a;
    if (exp != d.s[element & 7]) {
      DEBUG(1, failed);
      DEBUG(2, "\tshort: element=0x%x ", element);
      DEBUG(2, "d=0x%x exp=0x%x\n", d.s[element & 7], exp);
      return (1);
    }
  }

  /* uint */
  for (i=0; i<samples; i++) {
    unsigned int a, exp;
    a = (unsigned int)rand_local();
    element = (int)rand_local();
    d.vui = vec_promote(a, element);
    exp = a;
    if (exp != d.ui[element & 3]) {
      DEBUG(1, failed);
      DEBUG(2, "\tuint: element=0x%x ", element);
      DEBUG(2, "d=0x%x exp=0x%x\n", d.ui[element & 3], exp);
      return (1);
    }
  }

  /* int */
  for (i=0; i<samples; i++) {
    signed int a, exp;
    a = (signed int)rand_local();
    element = (int)rand_local();
    d.vi = vec_promote(a, element);
    exp = a;
    if (exp != d.i[element & 3]) {
      DEBUG(1, failed);
      DEBUG(2, "\tint: element=0x%x ", element);
      DEBUG(2, "d=0x%x exp=0x%x\n", d.i[element & 3], exp);
      return (1);
    }
  }

  /* float */
  for (i=0; i<samples; i++) {
    float a, exp;
    a = rand_float(-4.0f, 4.0f);
    element = (int)rand_local();
    d.vf = vec_promote(a, element);
    exp = a;
    if (exp != d.f[element & 3]) {
      DEBUG(1, failed);
      DEBUG(2, "\tfloat: element=0x%x ", element);
      DEBUG(2, "d=%f exp=%f\n", d.f[element & 3], exp);
      return (1);
    }
  }

  DEBUG(1, passed);
  return (0);
}


int test_vec_lvlx()
{
  int i, j;
  int base, offset;

  DEBUG(1, "lvlx.....");

  for (i=0; i<(int)(sizeof(data)); i++) data[i] = i;

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    d.vuc = vec_lvlx(offset, (unsigned char *)(&data[base])); 

    for (j=0; j<16; j++) {
      unsigned char exp;
      int eb;

      eb = (base + offset) & 0xF;

      if (j < (16-eb)) {
	exp = (base + offset) + j;
      } else {
	exp = 0;
      }

      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", offset, base, d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vuc = vec_lvlx(offset, (vector unsigned char *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vc = vec_lvlx(offset, (signed char *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vc = vec_lvlx(offset, (vec_char16 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbc = vec_lvlx(offset, (vec_bchar16 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vus = vec_lvlx(offset, (unsigned short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vus = vec_lvlx(offset, (vec_ushort8 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vs = vec_lvlx(offset, (signed short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vs = vec_lvlx(offset, (vec_short8 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbs = vec_lvlx(offset, (vec_bshort8 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vp = vec_lvlx(offset, (vec_pixel8 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    
    e.vui = vec_lvlx(offset, (unsigned int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vui = vec_lvlx(offset, (vec_uint4 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vi = vec_lvlx(offset, (signed int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vi = vec_lvlx(offset, (vec_int4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbi = vec_lvlx(offset, (vec_bint4 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vf = vec_lvlx(offset, (float *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vf = vec_lvlx(offset, (vec_float4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}

int test_vec_lvlxl()
{
  int i, j;
  int base, offset;

  DEBUG(1, "lvlxl....");

  for (i=0; i<(int)(sizeof(data)); i++) data[i] = i;

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    d.vuc = vec_lvlxl(offset, (unsigned char *)(&data[base])); 

    for (j=0; j<16; j++) {
      unsigned char exp;
      int eb;

      eb = (base + offset) & 0xF;

      if (j < (16-eb)) {
	exp = (base + offset) + j;
      } else {
	exp = 0;
      }

      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", offset, base, d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vuc = vec_lvlxl(offset, (vec_uchar16 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vc = vec_lvlxl(offset, (signed char *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vc = vec_lvlxl(offset, (vec_char16 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbc = vec_lvlxl(offset, (vec_bchar16 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vus = vec_lvlxl(offset, (unsigned short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vus = vec_lvlxl(offset, (vec_ushort8 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vs = vec_lvlxl(offset, (signed short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vs = vec_lvlxl(offset, (vec_short8 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbs = vec_lvlxl(offset, (vec_bshort8 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vp = vec_lvlxl(offset, (vec_pixel8 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    
    e.vui = vec_lvlxl(offset, (unsigned int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vui = vec_lvlxl(offset, (vec_uint4 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vi = vec_lvlxl(offset, (signed int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vi = vec_lvlxl(offset, (vec_int4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbi = vec_lvlxl(offset, (vec_bint4 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vf = vec_lvlxl(offset, (float *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vf = vec_lvlxl(offset, (vec_float4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}


int test_vec_lvrx()
{
  int i, j;
  int base, offset;

  DEBUG(1, "lvrx.....");

  for (i=0; i<(int)(sizeof(data)); i++) data[i] = i;

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    d.vuc = vec_lvrx(offset, (unsigned char *)(&data[base])); 

    for (j=0; j<16; j++) {
      unsigned char exp;
      int eb;

      eb = (base + offset) & 0xF;

      if (j >= (16-eb)) exp = base + offset + j - 16;
      else exp = 0;

      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", offset, base, d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vuc = vec_lvrx(offset, (vec_uchar16 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vc = vec_lvrx(offset, (signed char *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vc = vec_lvrx(offset, (vec_char16 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbc = vec_lvrx(offset, (vec_bchar16 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vus = vec_lvrx(offset, (unsigned short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vus = vec_lvrx(offset, (vec_ushort8 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vs = vec_lvrx(offset, (signed short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vs = vec_lvrx(offset, (vec_short8 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbs = vec_lvrx(offset, (vec_bshort8 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vp = vec_lvrx(offset, (vec_pixel8 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    
    e.vui = vec_lvrx(offset, (unsigned int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vui = vec_lvrx(offset, (vec_uint4 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vi = vec_lvrx(offset, (signed int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vi = vec_lvrx(offset, (vec_int4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbi = vec_lvrx(offset, (vec_bint4 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vf = vec_lvrx(offset, (float *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vf = vec_lvrx(offset, (vec_float4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}



int test_vec_lvrxl()
{
  int i, j;
  int base, offset;

  DEBUG(1, "lvrxl....");

  for (i=0; i<(int)(sizeof(data)); i++) data[i] = i;

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);

    d.vuc = vec_lvrxl(offset, (unsigned char *)(&data[base])); 

    for (j=0; j<16; j++) {
      unsigned char exp;
      int eb;

      eb = (base + offset) & 0xF;

      if (j >= (16-eb)) exp = base + offset + j - 16;
      else exp = 0;

      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\telement=%d ", j);
	DEBUG(2, "a=%d b=%d d=0x%x exp=0x%x\n", offset, base, d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vuc = vec_lvrxl(offset, (vec_uchar16 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vc = vec_lvrxl(offset, (signed char *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vc = vec_lvrxl(offset, (vec_char16 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbc = vec_lvrxl(offset, (vec_bchar16 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vus = vec_lvrxl(offset, (unsigned short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vus = vec_lvrxl(offset, (vec_ushort8 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vs = vec_lvrxl(offset, (signed short *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vs = vec_lvrxl(offset, (vec_short8 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbs = vec_lvrxl(offset, (vec_bshort8 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vp = vec_lvrxl(offset, (vec_pixel8 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    
    e.vui = vec_lvrxl(offset, (unsigned int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vui = vec_lvrxl(offset, (vec_uint4 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vi = vec_lvrxl(offset, (signed int *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vi = vec_lvrxl(offset, (vec_int4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vbi = vec_lvrxl(offset, (vec_bint4 *)(&data[base]));
    VERIFY_EQUAL(d, e);

    e.vf = vec_lvrxl(offset, (float *)(&data[base]));
    VERIFY_EQUAL(d, e);
    e.vf = vec_lvrxl(offset, (vec_float4 *)(&data[base]));
    VERIFY_EQUAL(d, e);
  }
  DEBUG(1, passed);
  return (0);
}



int test_vec_splats()
{
  int i, j;

  DEBUG(1, "splats...");
  
  /* byte */
  for (i=0; i<samples; i++) {
    unsigned char a, exp;
    a = (unsigned char)rand_local();
    d.vuc = vec_splats(a);
    exp = a;
    for (j=0; j<16; j++) {
      if (exp != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.uc[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vc = vec_splats((signed char)(a));
    VERIFY_EQUAL(d, e);
  }

  /* halfword */
  for (i=0; i<samples; i++) {
    unsigned short a, exp;
    a = (unsigned short)rand_local();
    d.vus = vec_splats(a);
    exp = a;
    for (j=0; j<8; j++) {
      if (exp != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.us[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vs = vec_splats((signed short)(a));
    VERIFY_EQUAL(d, e);
  }

  /* word */
  for (i=0; i<samples; i++) {
    unsigned int a, exp;
    a = (unsigned int)rand_local();
    d.vui = vec_splats(a);
    exp = a;
    for (j=0; j<4; j++) {
      if (exp != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint: element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x%x\n", d.ui[j], exp);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vi = vec_splats((signed int)(a));
    VERIFY_EQUAL(d, e);
  }

  /* float */
  for (i=0; i<samples; i++) {
    float a, exp;
    a = rand_float(-1.0f, 1.0f);
    d.vf = vec_splats(a);
    exp = a;
    for (j=0; j<4; j++) {
      if (exp != d.f[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tfloat: element=%d ", j);
	DEBUG(2, "d=%f exp=%f\n", d.f[j], exp);
	return (1);
      }
    }
  }

  /* immediate forms */
  /* byte */
  {
    d.vuc = vec_splats((unsigned char)(240));
    for (j=0; j<16; j++) {
      if (240 != d.uc[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuchar (lit): element=%d ", j);
	DEBUG(2, "d=%d exp=240\n", d.uc[j]);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vc = vec_splats((signed char)(-16));
    VERIFY_EQUAL(d, e);
  }

  /* short */
  {
    d.vus = vec_splats((unsigned short)(0x8234));
    for (j=0; j<8; j++) {
      if (0x8234 != d.us[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tushort (lit): element=%d ", j);
	DEBUG(2, "d=0x%x exp=0x8234\n", d.us[j]);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vs = vec_splats((signed short)(-32204));
    VERIFY_EQUAL(d, e);
  }

  /* int */
  {
    d.vui = vec_splats((unsigned int)(0xFFFF1234));
    for (j=0; j<4; j++) {
      if (0xFFFF1234 != d.ui[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint (lit): element=%d ", j);
	DEBUG(2, "d=0x%x exp=0xFFFF1234\n", d.ui[j]);
	return (1);
      }
    }
    /* verify equivalent variations */
    e.vi = vec_splats((signed int)(-60876));
    VERIFY_EQUAL(d, e);
  }

  /* float */
  {
    d.vf = vec_splats(-1.0f);
    for (j=0; j<4; j++) {
      if (-1.0f != d.f[j]) {
	DEBUG(1, failed);
	DEBUG(2, "\tuint (lit): element=%d ", j);
	DEBUG(2, "d=%d exp=-1.0\n", d.ui[j]);
	return (1);
      }
    }
  }

  DEBUG(1, passed);
  return (0);
}

int test_vec_stvlx()
{
  int i;
  int base, offset;
  int index;

  DEBUG(1, "stvlx....");

  for (i=0; i<(int)(sizeof(data)); i++) data[i] = i;

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);
    index = (unsigned)base + (unsigned)offset;

    a.vui  = rand_vui();
    vec_stvlx(a.vuc, offset, (unsigned char *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vuc, offset, (vec_uchar16 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vc, offset, (signed char *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vc, offset, (vec_char16 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vbc, offset, (vec_bchar16 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);

    a.vui  = rand_vui();
    vec_stvlx(a.vus, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vus, offset, (vec_ushort8 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vs, offset, (signed short *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vs, offset, (vec_short8 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vbs, offset, (vec_bshort8 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);

    a.vui  = rand_vui();
    vec_stvlx(a.vp, offset, (vec_pixel8 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);

    a.vui  = rand_vui();
    vec_stvlx(a.vui, offset, (unsigned int *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vui, offset, (vec_uint4 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vi, offset, (signed int *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vi, offset, (vec_int4 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vbi, offset, (vec_bint4 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    
    a.vui  = rand_vui();
    vec_stvlx(a.vf, offset, (float *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlx(a.vf, offset, (vec_float4 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
  }
  DEBUG(1, passed);
  return (0);
}



int test_vec_stvlxl()
{
  int i;
  int base, offset;
  int index;

  DEBUG(1, "stvlxl...");

  for (i=0; i<(int)(sizeof(data)); i++) data[i] = i;

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);
    index = (unsigned)base + (unsigned)offset;

    a.vui  = rand_vui();
    vec_stvlxl(a.vuc, offset, (unsigned char *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vuc, offset, (vec_uchar16 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vc, offset, (signed char *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vc, offset, (vec_char16 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vbc, offset, (vec_bchar16 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);

    a.vui  = rand_vui();
    vec_stvlxl(a.vus, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vus, offset, (vec_ushort8 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vs, offset, (signed short *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vs, offset, (vec_short8 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vbs, offset, (vec_bshort8 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);

    a.vui  = rand_vui();
    vec_stvlxl(a.vp, offset, (vec_pixel8 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);

    a.vui  = rand_vui();
    vec_stvlxl(a.vui, offset, (unsigned int *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vui, offset, (vec_uint4 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vi, offset, (signed int *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vi, offset, (vec_int4 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vbi, offset, (vec_bint4 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    
    a.vui  = rand_vui();
    vec_stvlxl(a.vf, offset, (float *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
    a.vui  = rand_vui();
    vec_stvlxl(a.vf, offset, (vec_float4 *)(&data[base])); 
    VERIFY_DATA_LEFT(a, index);
  }
  DEBUG(1, passed);
  return (0);
}



int test_vec_stvrx()
{
  int i;
  int base, offset;
  int index;

  DEBUG(1, "stvrx....");

  for (i=0; i<(int)(sizeof(data)); i++) data[i] = i;

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);
    index = (unsigned)base + (unsigned)offset;

    a.vui  = rand_vui();
    vec_stvrx(a.vuc, offset, (unsigned char *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vuc, offset, (vec_uchar16 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vc, offset, (signed char *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vc, offset, (vec_char16 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vbc, offset, (vec_bchar16 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);

    a.vui  = rand_vui();
    vec_stvrx(a.vus, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vus, offset, (vec_ushort8 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vs, offset, (signed short *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vs, offset, (vec_short8 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vbs, offset, (vec_bshort8 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);

    a.vui  = rand_vui();
    vec_stvrx(a.vp, offset, (vec_pixel8 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);

    a.vui  = rand_vui();
    vec_stvrx(a.vui, offset, (unsigned int *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vui, offset, (vec_uint4 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vi, offset, (signed int *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vi, offset, (vec_int4 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vbi, offset, (vec_bint4 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    
    a.vui  = rand_vui();
    vec_stvrx(a.vf, offset, (float *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrx(a.vf, offset, (vec_float4 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
  }
  DEBUG(1, passed);
  return (0);
}



int test_vec_stvrxl()
{
  int i;
  int base, offset;
  int index;

  DEBUG(1, "stvrxl...");

  for (i=0; i<(int)(sizeof(data)); i++) data[i] = i;

  for (i=0; i<samples; i++) {
    base   = (int)(rand_local() & 0x7F);
    offset = (int)(rand_local() & 0x7F);
    index = (unsigned)base + (unsigned)offset;

    a.vui  = rand_vui();
    vec_stvrxl(a.vuc, offset, (unsigned char *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vuc, offset, (vec_uchar16 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vc, offset, (signed char *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vc, offset, (vec_char16 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vbc, offset, (vec_bchar16 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);

    a.vui  = rand_vui();
    vec_stvrxl(a.vus, offset, (unsigned short *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vus, offset, (vec_ushort8 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vs, offset, (signed short *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vs, offset, (vec_short8 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vbs, offset, (vec_bshort8 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);

    a.vui  = rand_vui();
    vec_stvrxl(a.vp, offset, (vec_pixel8 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);

    a.vui  = rand_vui();
    vec_stvrxl(a.vui, offset, (unsigned int *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vui, offset, (vec_uint4 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vi, offset, (signed int *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vi, offset, (vec_int4 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vbi, offset, (vec_bint4 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    
    a.vui  = rand_vui();
    vec_stvrxl(a.vf, offset, (float *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
    a.vui  = rand_vui();
    vec_stvrxl(a.vf, offset, (vec_float4 *)(&data[base])); 
    VERIFY_DATA_RIGHT(a, index);
  }
  DEBUG(1, passed);
  return (0);
}
#endif



DECLARE_MAIN()
{
  DECLARE_TEST_VARS();

  SCALE_SAMPLES(samples);

  INVOKE_TEST(test_vec_abs);
  INVOKE_TEST(test_vec_abss);
  INVOKE_TEST(test_vec_add);
  INVOKE_TEST(test_vec_addc);
  INVOKE_TEST(test_vec_adds);
  INVOKE_TEST(test_vec_and);
  INVOKE_TEST(test_vec_andc);
  INVOKE_TEST(test_vec_avg);
  INVOKE_TEST(test_vec_ceil);
  INVOKE_TEST(test_vec_cmpb);
  INVOKE_TEST(test_vec_cmpeq);
  INVOKE_TEST(test_vec_cmpge);
  INVOKE_TEST(test_vec_cmpgt);
  INVOKE_TEST(test_vec_cmple);
  INVOKE_TEST(test_vec_cmplt);
  INVOKE_TEST(test_vec_ctf);
  INVOKE_TEST(test_vec_cts);
  INVOKE_TEST(test_vec_ctu);
  INVOKE_TEST(test_vec_dss);
  INVOKE_TEST(test_vec_dssall);
  INVOKE_TEST(test_vec_dst);
  INVOKE_TEST(test_vec_dstst);
  INVOKE_TEST(test_vec_dststt);
  INVOKE_TEST(test_vec_dstt);
  INVOKE_TEST(test_vec_expte);
  INVOKE_TEST(test_vec_floor);
  INVOKE_TEST(test_vec_ld);
  INVOKE_TEST(test_vec_lde);
  INVOKE_TEST(test_vec_ldl);
  INVOKE_TEST(test_vec_loge);
  INVOKE_TEST(test_vec_lvsl);
  INVOKE_TEST(test_vec_lvsr);
  INVOKE_TEST(test_vec_madd);
  INVOKE_TEST(test_vec_madds);
  INVOKE_TEST(test_vec_max);
  INVOKE_TEST(test_vec_mergeh);
  INVOKE_TEST(test_vec_mergel);
  INVOKE_TEST(test_vec_mfvscr);
  INVOKE_TEST(test_vec_min);
  INVOKE_TEST(test_vec_mladd);
  INVOKE_TEST(test_vec_mradds);
  INVOKE_TEST(test_vec_msum);
  INVOKE_TEST(test_vec_msums);
  INVOKE_TEST(test_vec_mtvscr);
  INVOKE_TEST(test_vec_mule);
  INVOKE_TEST(test_vec_mulo);
  INVOKE_TEST(test_vec_nmsub);
  INVOKE_TEST(test_vec_nor);
  INVOKE_TEST(test_vec_or);
  INVOKE_TEST(test_vec_pack);
  INVOKE_TEST(test_vec_packpx);
  INVOKE_TEST(test_vec_packs);
  INVOKE_TEST(test_vec_packsu);
  INVOKE_TEST(test_vec_perm);
  INVOKE_TEST(test_vec_re);
  INVOKE_TEST(test_vec_rl);
  INVOKE_TEST(test_vec_round);
  INVOKE_TEST(test_vec_rsqrte);
  INVOKE_TEST(test_vec_sel);
  INVOKE_TEST(test_vec_sl);
  INVOKE_TEST(test_vec_sld);
  INVOKE_TEST(test_vec_sll);
  INVOKE_TEST(test_vec_slo);
  INVOKE_TEST(test_vec_splat);
  INVOKE_TEST(test_vec_splat_s8);
  INVOKE_TEST(test_vec_splat_s16);
  INVOKE_TEST(test_vec_splat_s32);
  INVOKE_TEST(test_vec_splat_u8);
  INVOKE_TEST(test_vec_splat_u16);
  INVOKE_TEST(test_vec_splat_u32);
  INVOKE_TEST(test_vec_sr);
  INVOKE_TEST(test_vec_sra);
  INVOKE_TEST(test_vec_srl);
  INVOKE_TEST(test_vec_sro);
  INVOKE_TEST(test_vec_st);
  INVOKE_TEST(test_vec_ste);
  INVOKE_TEST(test_vec_stl);
  INVOKE_TEST(test_vec_sub);
  INVOKE_TEST(test_vec_subc);
  INVOKE_TEST(test_vec_subs);
  INVOKE_TEST(test_vec_sum4s);
  INVOKE_TEST(test_vec_sum2s);
  INVOKE_TEST(test_vec_sums);
  INVOKE_TEST(test_vec_trunc);
  INVOKE_TEST(test_vec_unpackh);
  INVOKE_TEST(test_vec_unpackl);
  INVOKE_TEST(test_vec_xor);

  /* The CBEA specific VMX intrinsics.
   */
#if 0  /* Commented out until compilers support has been added */
  INVOKE_TEST(test_vec_extract);
  INVOKE_TEST(test_vec_insert);
  INVOKE_TEST(test_vec_lvlx);
  INVOKE_TEST(test_vec_lvlxl);
  INVOKE_TEST(test_vec_lvrx);
  INVOKE_TEST(test_vec_lvrxl);
  INVOKE_TEST(test_vec_promote);
  INVOKE_TEST(test_vec_splats);
  INVOKE_TEST(test_vec_stvlx);
  INVOKE_TEST(test_vec_stvlxl);
  INVOKE_TEST(test_vec_stvrx);
  INVOKE_TEST(test_vec_stvrxl);
#endif

  DISPLAY_RESULTS();
}
